/****************************************************************************
 *
 * File Name: Dispatcher.cpp
 *
 * General Description: General description of the file.
 *
 * Created By: Alex

 , Creation Date: 22/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Dispatcher.cpp
 *\brief Implements the Dispatcher Class.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "../Network_Driver/NetworkDriver.h"
#include "../LLRE/LLRE.h"
#include "../GFCP_FE_Client/src/GFCP_WS_Client.h"
#include "Dispatcher.h"

using namespace std;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
Dispatcher::Dispatcher()
{
  ndriver = NULL;
  llre = NULL;
  gfcpClient = NULL;
  startQueueProcessingThread();
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _ndriver - NetworkDriver pointer.
 */
/****************************************************************************/
Dispatcher::Dispatcher(NetworkDriver *_ndriver)
{
  ndriver = _ndriver;
  llre = NULL;
  gfcpClient = NULL;
  startQueueProcessingThread();
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
Dispatcher::~Dispatcher()
{
  Frame endThread(0, NULL, 0);
  dispatcherQ.push(&endThread);

  //wait for processing thread to stop
  pthread_join(threadPtr, NULL);
  Frame *frm = NULL;
  while (!dispatcherQ.empty())
  {
    dispatcherQ.try_pop(frm);
    if (frm != NULL && frm->frame != NULL)
    {
      delete[] frm->frame;
      delete frm;
    }
  }
}

/****************************************************************************/
//! Adds a new frame to the processing queue.
/*! Adds a new frame to the processing queue. This queue is used to pass
 * frames between GFEs modules, for thread safety.
 *
 * @param frame - new frame to process.
 */
/****************************************************************************/
void Dispatcher::pushFrameToQueue(Frame *frame)
{
  dispatcherQ.push(frame);
}

/****************************************************************************/
//! Forwards the received frame to Port/Virtual Port.
/*! Forwards the received frame to Port/Virtual Port.
 *
 * @param portID - the port to which the frame is to be sent.
 * @param frame - frame to forward.
 * @param length - frame length.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Dispatcher::sendFrame(Frame * frm)
{
  if (ndriver != NULL && llre != NULL)
  {

    if (frm->interface == CE_PORT)
    {
//      cout << "Dispatcher: Forwarding to: CE"<< endl;

      //TODO until CE is implemented
//      gfcpClient->sendFrameToCE(frm);
      delete[] frm->frame;
      delete frm;

      return "Forwarding to CE";
    }
    else if (frm->interface == LLRE_PORT)
    {
      //			cout << "Dispatcher: Forwarding to: LLRE"<< endl;

      llre->pushFrameToQueue(frm);
      return "Forwarding to LLRE";
    }
    else if (0 <= frm->interface && frm->interface < 100) ///temp!!!!!!!!!!!!!! TODO remove
    {
      //			cout << "Dispatcher: Forwarding to: 0x" << frm->interface << endl;

      ndriver->pushFrameToQueue(frm);
      return "Forwarding to Network Driver";
    }
    else
    {
      cout << "Dispatcher: Bad interface: 0x" << frm->interface << endl;

      return "Error";
    }
  }
  return "No NetworkDriver or LLRE assigned to this Dispatcher.";
}

/****************************************************************************/
//! Returns the NetworkDriver pointer.
/*! Returns the NetworkDriver pointer.
 *
 * @return Returns the NetworkDriver pointer.
 */
/****************************************************************************/
NetworkDriver *Dispatcher::getNdriver() const
{
  return ndriver;
}

/****************************************************************************/
//! Sets the NetworkDriver pointer.
/*! Sets the NetworkDriver pointer.
 *
 * @param ndriver - the NetworkDriver pointer.
 */
/****************************************************************************/
void Dispatcher::setNdriver(NetworkDriver *ndriver)
{
  this->ndriver = ndriver;
}

/****************************************************************************/
//! Returns the LLRE pointer.
/*! Returns the LLRE pointer.
 *
 * @return Returns the LLRE pointer.
 */
/****************************************************************************/
LLRE *Dispatcher::getLlre() const
{
  return llre;
}

/****************************************************************************/
//! Sets the LLRE pointer.
/*! Sets the LLRE pointer.
 *
 * @param llre - the LLRE pointer.
 */
/****************************************************************************/
void Dispatcher::setLlre(LLRE *llre)
{
  this->llre = llre;
}

/****************************************************************************/
//! Sets the gfcpClient pointer.
/*! Sets the gfcpClient pointer.
 *
 * @param gfcpClient - the gfcpClient pointer.
 */
/****************************************************************************/
void Dispatcher::setGfcp_client(GFCP_WS_Client *gfcpClient)
{
  this->gfcpClient = gfcpClient;
}

/****************************************************************************/
//! Starts a new queue processing thread.
/*! Starts a new queue processing thread, to serve the frame queue.
 */
/****************************************************************************/
void Dispatcher::startQueueProcessingThread()
{
  threadPtr = 0;
  if (pthread_create(&threadPtr, NULL, this->processQueue, (void *)this))
    throw GFEexception("Failed to start new Dispatcher thread.");
}

/****************************************************************************/
//! Processing thread function, to serve the frames queue.
/*! Processing thread function, to serve the frames queue. This thread is
 * terminated by adding a NULL frame to queue.
 *
 * @param _disp - a Dispatcher object.
 */
/****************************************************************************/
void *Dispatcher::processQueue(void *_disp)
{
  Dispatcher *disp = (Dispatcher *)_disp;
  Frame *frm = NULL;
  bool processFlag = true;
  while (processFlag)
  {
    disp->dispatcherQ.wait_and_pop(frm);

    if (frm != NULL && frm->frame != NULL)
      disp->sendFrame(frm);
    else
      processFlag = false;
  }
  pthread_exit(0);
  return 0;
}
