/****************************************************************************
 *
 * File Name: GFE.cpp
 *
 * General Description: Implements the Generic Forwarding Element Class.
 *
 * Created By: Alex

 , Creation Date: 10/05/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file GFE.cpp
 *\brief Implements the Generic Forwarding Element Class.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "GFE.h"

using namespace std;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
GFE::GFE()
{
  dispatcher = new Dispatcher();
  llre = new LLRE(dispatcher);
  netdriver = new NetworkDriver();

  //TODO temp loop

  gfcp_client = new GFCP_WS_Client("192.168.81.1", 8000,"");
  //	gfcp_server = new GFCP_Server(8000,gfcp_client);

  xmlHandler = new XML_Handler(llre, gfcp_client);
  gfcp_server = new GFCP_Server(8000, xmlHandler);

  dispatcher->setNdriver(netdriver);
  dispatcher->setLlre(llre);
  dispatcher->setGfcp_client(gfcp_client);
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
GFE::~GFE()
{
  delete gfcp_server;
  delete xmlHandler;
  delete gfcp_client;
  delete netdriver;
  delete llre;
  delete dispatcher;
}

/****************************************************************************/
//! Starts a packet capture the passed interface.
/*! Starts a packet capture on the passed interface, a callback function
 * is called each time a packet arrives.
 *
 * @param interface - the interface on which to capture packets.
 * @param num - number of packets to capture, -1 for infinity.
 */
/****************************************************************************/
string GFE::startCapture(unsigned int interface, int num)
{
  return netdriver->StartCapture(interface, num, my_callback);
}

/****************************************************************************/
//! Stops the packet capture on the passed interface.
/*! Stops the packet capture on the passed interface.
 *
 * @param interface - the interface on which to capture packets.
 */
/****************************************************************************/
string GFE::stopCapture(unsigned int interface)
{
  return netdriver->StopCapture(interface);
}

/****************************************************************************/
//! Stops the packet capture thread on all active interfaces.
/*! Stops the packet capture thread on all active interfaces.
 *
 */
/****************************************************************************/
string GFE::StopAllActiveCapture()
{
  return netdriver->StopAllActiveCapture();
}

/****************************************************************************/
//! Prints the Initial Lookup Table (Port->ID_KEYs).
/*! Prints the Initial Lookup Table (Port->ID_KEYs)
 */
/****************************************************************************/
void GFE::printInitialTable()
{
  if (llre != NULL)
    llre->printInitialTable();
  else
    cout << "LLRE not defined\n";
}

/****************************************************************************/
//! Prints the ID_to_FWD Table (Port,ID_KEYs->FWD_KEYs).
/*! Prints the ID_to_FWD Table (Port,ID_KEYs->FWD_KEYs)
 */
/****************************************************************************/
void GFE::printIDtoFWDTable()
{
  if (llre != NULL)
    llre->printIDtoFWDTable();
  else
    cout << "LLRE not defined\n";
}

/****************************************************************************/
//! Prints the Action Table (FWD_KEYs->Actions).
/*! Prints the Action Table (FWD_KEYs->Actions)
 */
/****************************************************************************/
void GFE::printActionTable()
{
  if (llre != NULL)
    llre->printActionTable();
  else
    cout << "LLRE not defined\n";
}

/****************************************************************************/
//! Disables entry in the ID_KEYs Table.
/*! Disables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string GFE::disableEntry(unsigned int entryId)
{
  if (llre != NULL)
    return llre->disableEntry(entryId);
  return "LLRE not defined\n";
}

/****************************************************************************/
//! Enables entry in the ID_KEYs Table.
/*! Enables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string GFE::enableEntry(unsigned int entryId)
{
  if (llre != NULL)
    return llre->enableEntry(entryId);
  return "LLRE not defined\n";
}

/****************************************************************************/
//! Prints the available network interfaces.
/*! Prints the available network interfaces.
 */
/****************************************************************************/
void GFE::printInterfaces()
{
  netdriver->printInterfaces();
}

/****************************************************************************/
//! Loads the classifier lfb file.
/*! Loads the classifier lfb file.
 *
 * @param LFB - file name of the LFB to load.
 * @return status message.
 */
/****************************************************************************/
string GFE::loadClassifier(const char *LFB)
{
  return xmlHandler->Handle_LFB_Classifier_File(LFB);
}

/****************************************************************************/
//! Loads the forwarder lfb file.
/*! Loads the forwarder lfb file.
 *
 * @param LFB - file name of the LFB to load.
 * @return status message.
 */
/****************************************************************************/
string GFE::loadForwarder(const char *LFB)
{
  return xmlHandler->Handle_LFB_Forwarder_File(LFB);
}

/****************************************************************************/
//! Loads the mapper lfb file.
/*! Loads the mapper lfb file.
 *
 * @param LFB - file name of the LFB to load.
 * @return status message.
 */
/****************************************************************************/
string GFE::loadMapper(const char *LFB)
{
  return xmlHandler->Handle_LFB_Mapper_File(LFB);
}

/****************************************************************************/
//! Callback function that handles received packets.
/*! Callback function that handles received packets.
 *
 * @param args - char array that stores the port number.
 * @param pkthdr - packet header from libpcap.
 * @param packet - pointer to the captured packet.
 */
/****************************************************************************/
void GFE::my_callback(u_char *args, const struct pcap_pkthdr* pkthdr, const u_char* packet)
{
  BYTE* frame = new BYTE[pkthdr->len];
  memmove(frame, packet, pkthdr->len);

  unsigned int port = NetworkDriver::charArryToInt(args);
  getInstance().llre->pushFrameToQueue(new Frame(port, frame, pkthdr->len));
}

/****************************************************************************/
//! Returns a pointer to the LLRE.
/*! Returns a pointer to the LLRE.
 *
 * @return a pointer to the LLRE.
 */
/****************************************************************************/
LLRE* GFE::getLlre()
{
  return llre;
}

/****************************************************************************/
//! Prints the xmls load status.
/*! Prints the xmls load status.
 */
/****************************************************************************/
void GFE::printXmlHandlerStatus()
{
  switch (xmlHandler->xmlLoadStatus())
  {
    case 0:
      cout << "║              │            │            ║\n";
      break;
    case 1:
      cout << "║  Classifier  │            │            ║\n";
      break;
    case 2:
      cout << "║  Classifier  │   Mapper   │            ║\n";
      break;
    case 3:
      cout << "║  Classifier  │   Mapper   │  Forwarder ║\n";
      break;
    default:
      cout << "║              │            │            ║\n";
      break;
  }
}
