/****************************************************************************
 *
 * File Name: Datastructures.h
 *
 * General Description: Defines the Data structures used in the GFE.
 *
 * Created By: Alex Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date Updated By Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/
#ifndef DATASTRUCTURES_H
#define DATASTRUCTURES_H
/****************************************************************************
 * File description
 ****************************************************************************/
/*! \file Datastructures.h
 *\brief Defines the Data structures used in the GFE.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "../Includes/Common.h"

/*
 * byte offset = 1
 * byte length = 3
 * first bit offset = 0
 * last bit offset = 7
 *
 *                           |---------------[3]---------------|
 *                           |                                 |
 *                           V                                 V
 *  |------(0)------||------(1)------||------(2)------||------(3)------|
 *
 *                  ||               ||               ||
 *  |0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7|
 *                  ||^              ||               ||              ^
 *                    ^                                               ^
 *                    |                                               |
 *
 *
 * byte offset = 1
 * byte length = 3
 * first bit offset = 3
 * last bit offset = 4
 *
 *                           |---------------[3]---------------|
 *                           |                                 |
 *                           V                                 V
 *  |------(0)------||------(1)------||------(2)------||------(3)------|
 *
 *                  ||               ||               ||
 *  |0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7|
 *                  ||      ^        ||               ||        ^
 *                          ^                                   ^
 *                          |                                   |
 *
 *
 * byte offset = 1
 * byte length = 2
 * first bit offset = 3
 * last bit offset = 3
 *
 *                           |------[2]-------|
 *                           |                |
 *                           V                V
 *  |------(0)------||------(1)------||------(2)------||------(3)------|
 *
 *                  ||               ||               ||
 *  |0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7|
 *                  ||      ^        ||      ^        ||
 *                          ^                ^
 *                          |                |
 *
 * byte offset = 1
 * byte length = 1
 * first bit offset = 2
 * last bit offset = 5
 *
 *                         |[1]|
 *                         |   |
 *                         V   V
 *  |------(0)------||------(1)------||------(2)------||------(3)------|
 *
 *                  ||               ||               ||
 *  |0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7||0|1|2|3|4|5|6|7|
 *                  ||    ^     ^    ||               ||
 *                        ^     ^
 *                        |     |
 */
class Position
{

  friend std::ostream& operator<<(std::ostream& output, const Position& pos)
  {
    output << "[Position: byteOffset=" << std::dec << pos.byteOffset << ", firstBitOffset=" << std::dec
        << (int)pos.firstBitOffset << ", byteLength=" << std::dec << pos.byteLength << ", lastBitOffset=" << std::dec
        << (int)pos.lastBitOffset << "]";
    return output;
  }

  friend std::ostream& operator<<(std::ostream& output, const Position *pos)
  {
    output << "[Position: byteOffset=" << std::dec << pos->byteOffset << ", firstBitOffset=" << std::dec
        << (int)pos->firstBitOffset << ", byteLength=" << std::dec << pos->byteLength << ", lastBitOffset=" << std::dec
        << (int)pos->lastBitOffset << "]";
    return output;
  }

public:

  Position(unsigned int byteOffsetVal, unsigned int byteLengthVal)
  {
    byteOffset = byteOffsetVal;
    byteLength = byteLengthVal;
    firstBitOffset = 0;
    lastBitOffset = 7;
  }

  Position(unsigned int byteOffsetVal, unsigned int byteLengthVal, unsigned char firstBitOffsetVal,
           unsigned char lastBitOffsetVal)
  {
    byteOffset = byteOffsetVal;
    byteLength = byteLengthVal;
    firstBitOffset = firstBitOffsetVal;
    lastBitOffset = lastBitOffsetVal;
  }

  Position()
  {
    byteOffset = 0;
    byteLength = 0;
    firstBitOffset = 0;
    lastBitOffset = 7;
  }

  Position(const Position &other) // copy constructor
  {
    *this = other;
  }

  unsigned int getOffset() const
  {
    return byteOffset;
  }
  unsigned int getLength() const
  {
    return byteLength;
  }
  void setOffset(unsigned int _byteOffset)
  {
    byteOffset = _byteOffset;
  }
  void setLength(unsigned int _byteLength)
  {
    byteLength = _byteLength;
  }

  unsigned char getFirstBitOffset() const
  {
    return firstBitOffset;
  }
  unsigned char getLastBitOffset() const
  {
    return lastBitOffset;
  }
  void setFirstBitOffset(unsigned char _firstBitOffset)
  {
    firstBitOffset = _firstBitOffset;
  }
  void setLastBitOffset(unsigned char _lastBitOffset)
  {
    lastBitOffset = _lastBitOffset;
  }
  void clearPosition()
  {
    byteOffset = 0;
    byteLength = 0;
    firstBitOffset = 0;
    lastBitOffset = 7;
  }

  unsigned int getNumberOfBits()
  {
    if (byteLength == 0)
      return 0;
    else if (byteLength == 1)
    {
      return lastBitOffset - firstBitOffset + 1;
    }
    else
    {
      return (8 - firstBitOffset) + (byteLength - 2) * 8 + (lastBitOffset + 1);
    }
  }

  const Position& operator=(const Position& pos)
  {
    this->byteOffset = pos.byteOffset;
    this->byteLength = pos.byteLength;
    this->firstBitOffset = pos.firstBitOffset;
    this->lastBitOffset = pos.lastBitOffset;
    return *this;
  }

  bool operator ==(const Position & other) const
  {
    return this->byteOffset == other.byteOffset && this->byteLength == other.byteLength && this->firstBitOffset
        == other.firstBitOffset && this->lastBitOffset == other.lastBitOffset;
  }

protected:
  unsigned int byteOffset;
  unsigned int byteLength;
  unsigned char firstBitOffset;
  unsigned char lastBitOffset;
};

class Frame_field : public Position
{
  friend std::ostream& operator<<(std::ostream& output, const Frame_field& Ff)
  {
    output << "[" << Ff.field_id << ", " << Ff.name << ", type=" << (int)(Ff.type) << ", byteOffset=" << std::dec
        << Ff.byteOffset << ", firstBitOffset=" << std::dec << (int)Ff.firstBitOffset << ", byteLength=" << std::dec
        << Ff.byteLength << ", lastBitOffset=" << std::dec << (int)Ff.lastBitOffset << "]";
    return output;
  }

  friend std::ostream& operator<<(std::ostream& output, const Frame_field *Ff)
  {
    output << "[" << Ff->field_id << ", " << Ff->name << ", type=" << (int)(Ff->type) << ", byteOffset=" << std::dec
        << Ff->byteOffset << ", firstBitOffset=" << std::dec << (int)Ff->firstBitOffset << ", byteLength=" << std::dec
        << Ff->byteLength << ", lastBitOffset=" << std::dec << (int)Ff->lastBitOffset << "]";
    return output;
  }

public:

  Frame_field(int idVal, unsigned int byteOffsetVal, unsigned int byteLengthVal, std::string &nameVal, BYTE typeVal) :
    Position(byteOffsetVal, byteLengthVal) // constructor
  {
    field_id = idVal;
    name = nameVal;
    type = typeVal;
  }

  Frame_field(int idVal, unsigned int byteOffsetVal, unsigned int byteLengthVal, std::string &nameVal, BYTE typeVal,
              unsigned char firstBitOffsetVal, unsigned char lastBitOffsetVal) :
    Position(byteOffsetVal, byteLengthVal, firstBitOffsetVal, lastBitOffsetVal) // constructor
  {
    field_id = idVal;
    name = nameVal;
    type = typeVal;
  }

  Frame_field()
  {
  }

  Frame_field(const Frame_field &other) // copy constructor
  {
    field_id = other.field_id;
    byteLength = other.byteLength;
    byteOffset = other.byteOffset;
    firstBitOffset = other.firstBitOffset;
    lastBitOffset = other.lastBitOffset;
    name = other.name;
    type = other.type;
  }

  void printField() const
  {
    std::cout << "[" << field_id << ", " << name << ", type=" << (int)type
    //					<<", byteOffset="<<std::dec<byteOffset
        //					<<", firstBitOffset="<<std::dec<<(int)firstBitOffset
        //					<<", byteLength="<<std::dec<byteLength
        //					<<", lastBitOffset="<<std::dec<<(int)lastBitOffset
        << "]";
  }
  std::string getName()
  {
    return name;
  }
  int getFieldId()
  {
    return field_id;
  }
  BYTE getType()
  {
    return type;
  }

protected:
  int field_id;
  std::string name;
  BYTE type;
};

class Frame
{
public:

  Frame(unsigned int _interface, BYTE *_frame, unsigned int _length)
  {
    interface = _interface;
    srcInterface = _interface;
    frame = _frame;
    length = _length;
  }

  Frame(const Frame &other) // copy constructor
  {
    this->interface = other.interface;
    this->srcInterface = other.srcInterface;
    this->frame = other.frame;
    this->length = other.length;
  }

  ~Frame()
  {
  } // destructor
  //		{
  //			if(this->frame != NULL)
  //				delete [] this->frame;
  //		}

  unsigned int interface;
  unsigned int srcInterface;
  BYTE *frame;
  unsigned int length;
};

#endif /* DATASTRUCTURES_H */
