/****************************************************************************
 *
 * File Name: concurrent_queue.h
 *
 * General Description: Implements the thread safe queue.
 *
 * Created By: Alex Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date Updated By Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/
#ifndef CONCURRENT_QUEUE_H_
#define CONCURRENT_QUEUE_H_
/****************************************************************************
 * File description
 ****************************************************************************/
/*! \file concurrent_queue.h
 *\brief Implements the thread safe queue.
 *
 * Taken from:
 * http://www.justsoftwaresolutions.co.uk/threading/implementing-a-thread-safe-queue-using-condition-variables.html
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include <queue>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/condition_variable.hpp>

/**
 * \brief Thread safe queue class.
 */
template<typename Data>
  class concurrent_queue
  {
  private:
    std::queue<Data> the_queue;
    mutable boost::mutex the_mutex;
    boost::condition_variable the_condition_variable;
  public:
    void push(Data const& data)
    {
      boost::mutex::scoped_lock lock(the_mutex);
      the_queue.push(data);
      lock.unlock();
      the_condition_variable.notify_one();
    }

    //TODO add push_front to stop threads

    bool empty() const
    {
      boost::mutex::scoped_lock lock(the_mutex);
      return the_queue.empty();
    }

    bool try_pop(Data& popped_value)
    {
      boost::mutex::scoped_lock lock(the_mutex);
      if (the_queue.empty())
      {
        return false;
      }

      popped_value = the_queue.front();
      the_queue.pop();
      return true;
    }

    void wait_and_pop(Data& popped_value)
    {
      boost::mutex::scoped_lock lock(the_mutex);
      while (the_queue.empty())
      {
        the_condition_variable.wait(lock);
      }

      popped_value = the_queue.front();
      the_queue.pop();
    }

  };

#endif /* CONCURRENT_QUEUE_H_ */
