/****************************************************************************
 *
 * File Name: Action_Table_Impl.cpp
 *
 * General Description: Implements the Actions Table.
 *
 * Created By: Alex

 , Creation Date: 09/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Action_Table_Impl.cpp
 *\brief Implements the Actions Table.
 *
 * This table stores the (FWD_KEYs, Actions) pairs.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "Action_Table_Impl.h"

using namespace boost;
using namespace std;
/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
Action_Table_Impl::Action_Table_Impl()
{
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
Action_Table_Impl::~Action_Table_Impl()
{
  clearTable();
}

/****************************************************************************/
//! Removes entry from the actionTable Table.
/*! Removes entry from the actionTable Table.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks to remove.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Action_Table_Impl::removeEntry(unsigned int interface, vector<fwdKeyMaskpair *> &fwdKeyMaskVect)
{
#ifdef PATRICIA_ACTIONS

#else
  vector<FWD_KEY *> fwdKeysWithMaskApplied;
  FWD_KEY::applyMask(fwdKeyMaskVect, fwdKeysWithMaskApplied);
  actionTable.erase(fwdKeysWithMaskApplied);
#endif
  return "OK";
}

/****************************************************************************/
//! Adds a new entry to the actionTable Table.
/*! Adds a new entry to the actionTable Table.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks to add.
 * @param actions - Actions to perform.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Action_Table_Impl::addNewEntry(unsigned int interface, vector<fwdKeyMaskpair *> &fwdKeyMaskVect, vector<Action*> &actions)
{
  if(interface < 65536)
  {
#ifdef PATRICIA_ACTIONS
    bytes b;
    b.byte = new unsigned char [MAX_KEY_SIZE];
    b.length = MAX_KEY_SIZE;
    memset(&b.byte[0], 0x00, b.length);

    bytes mask;
    mask.byte = new unsigned char [MAX_KEY_SIZE];
    mask.length = MAX_KEY_SIZE;
    memset(&mask.byte[0], 0xff, mask.length);

    b.byte[0] = 0x00; //needed for correct function of the tree (root value is always 0x00)
    //convert integer value to BYTEs
    b.byte[2] = (unsigned char)(interface);
    b.byte[1] = (unsigned char)(interface >> 8);

    unsigned int j = 3;

    for (vector<fwdKeyMaskpair *>::iterator it = fwdKeyMaskVect.begin();
        it != fwdKeyMaskVect.end();
        ++it)
    {
      //add offset to distinguish between fwdKeys with same values but at different offsets
      b.byte[j++] = (unsigned char)( (*it)->first.getPosition().getOffset() );
      b.byte[j++] = (unsigned char)( (*it)->first.getPosition().getOffset() >> 8);
      b.byte[j++] = (unsigned char)( (*it)->first.getPosition().getOffset() >> 8);
      b.byte[j++] = (unsigned char)( (*it)->first.getPosition().getOffset() >> 8);
      memmove(&b.byte[j], (*it)->first.getValue(), (*it)->first.getLength());
      j += (*it)->first.getLength();
    }

    vector<vector<Action *> > *toFindVect = actionTable.findReference(b);

    if(toFindVect->size()==0)
    {
      vector<vector<Action *> > toAddToVect;
      toAddToVect.push_back(actions);
      actionTable.insert(b,mask,toAddToVect);
    }
    else
    {
      toFindVect->push_back(actions);
    }
    delete [] b.byte;
    delete [] mask.byte;
#else
  vector<FWD_KEY *> fwdKeysWithMaskApplied;
  FWD_KEY::applyMask(fwdKeyMaskVect, fwdKeysWithMaskApplied);
  actionTable.insert(pair<vector<FWD_KEY *> , vector<Action*> > (fwdKeysWithMaskApplied, actions));
#endif
    return "Action table entry added successfully.";
  }
  else
    return "Port number is too high.";
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks to get.
 * @param actionsVect - a vector of vectors containing Actions corresponding to the fwdKey.
 */
/****************************************************************************/
void Action_Table_Impl::getActions(vector<fwdKeyMaskpair *> &fwdKeyMaskVect, vector<vector<Action*> > &actionsVect)
{
#ifdef PATRICIA_ACTIONS

#else
  vector<FWD_KEY *> fwdKeys;
  FWD_KEY::applyMask(fwdKeyMaskVect, fwdKeys);

  pair<umultimap_act::iterator, umultimap_act::iterator> range = actionTable.equal_range(fwdKeys);
  for (umultimap_act::iterator it = range.first; it != range.second; ++it)
  {
    actionsVect.push_back((it->second));
  }

  //TODO not optimal
  for (unsigned int i = 0; i < fwdKeys.size(); ++i)
  {
    if (fwdKeys[i] != NULL)
      delete fwdKeys[i];
  }
#endif
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks to get.
 * @param actionsVect - a vector of vectors containing Actions corresponding to the fwdKey.
 */
/****************************************************************************/
void Action_Table_Impl::getActions(bytes &keys, vector<vector<Action*> > &actionsVect)
{
#ifdef PATRICIA_ACTIONS
  actionsVect = actionTable.find(keys);
#endif
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKeys.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks to get.
 * @param actionsVect - a vector of vectors containing Actions corresponding to the fwdKey.
 */
/****************************************************************************/
vector<vector<Action*> > * Action_Table_Impl::returnActions(bytes &keys)
{
#ifdef PATRICIA_ACTIONS
//  return actionTable.findReference(keys);
  return &a;
#else
  return NULL;
#endif
}

/****************************************************************************/
//! Prints the Action Table.
/*! Prints the Action Table.
 */
/****************************************************************************/
void Action_Table_Impl::printTable()
{
  cout << "         Action Table:\n";
  cout << "------------------------------------------\n";
#ifdef PATRICIA_ACTIONS
  actionTable.printTree();
  cout << endl;
#else
  for (umultimap_act::const_iterator it = actionTable.begin(); it != actionTable.end(); ++it)
  {
    cout << it->first << it->second << "\n";
  }
#endif
}

/****************************************************************************/
//! Clear the Action Table.
/*! Clear the Action Table.
 */
/****************************************************************************/
void Action_Table_Impl::clearTable()
{
#ifdef PATRICIA_ACTIONS
  vector<maskedpatriciaNode< vector< vector<Action *> > > *> nodesVect;
  actionTable.fillVectorOfNodes(nodesVect);

  for (unsigned int i = 0; i < nodesVect.size(); ++i)
  {
    for (vector<vector<Action*> >::iterator actionVectItr = nodesVect[i]->value.begin(); actionVectItr != nodesVect[i]->value.end(); ++actionVectItr)
    {
      for (unsigned int i = 0; i < actionVectItr->size(); ++i)
        if ((actionVectItr->at(i)) != NULL)
          delete (actionVectItr->at(i));
    }
  }
  actionTable.clearTree();
#else
  for (umultimap_act::const_iterator it = actionTable.begin(); it != actionTable.end(); ++it)
  {
    for (vector<FWD_KEY*>::const_iterator fkVectItr = (it->first).begin(); fkVectItr != (it->first).end(); ++fkVectItr)
    {
      if ((*fkVectItr) != NULL)
        delete (*fkVectItr);
    }

    for (vector<Action*>::const_iterator actionVectItr = (it->second).begin(); actionVectItr != (it->second).end(); ++actionVectItr)
    {
      if ((*actionVectItr) != NULL)
        delete (*actionVectItr);
    }
  }
  actionTable.clear();
#endif
}
