/****************************************************************************
 *
 * File Name: Auxiliary.cpp
 *
 * General Description: General description of the file.
 *
 * Created By: Alex

 , Creation Date: 14/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Auxiliary.cpp
 *\brief Implements the Auxiliary namespace.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include <stdio_ext.h>
#include <termios.h>

#include "Auxiliary.h"
#include "../Includes/Datastructures.h"

namespace Auxiliary
{
/****************************************************************************/
//! Prints the received BYTE array.
/*! Prints the received BYTE array.
 *
 * @param value - pointer to the BYTE array.
 * @param length - array size.
 */
/****************************************************************************/
void printBYTEarray(unsigned char *value, unsigned int length)
{
  if (value != NULL)
  {
    std::cout << "|";
    for (unsigned int i = 0; i < length; ++i)
    {
      if (i != 0 && i % 8 == 0)
        std::cout << "\n|";

      std::cout << std::internal << std::hex << std::setw(4) << std::setfill('0') << (unsigned short *)value[i] << "|";
    }
  }
  else
    std::cout << "No value defined";
}

/****************************************************************************/
//! Prints the received Frame.
/*! Prints the received Frame array.
 *
 * @param frm - pointer to the Frame.
 * @param isBefore - true prints "Original", false prints "Processed".
 */
/****************************************************************************/
void printFrame(bool isBefore, Frame *frm)
{
  if (isBefore)
    std::cout << "\n	Original frame: [interface: " << frm->interface << "]\n";
  else
    std::cout << "\n	Processed frame: [interface: " << frm->interface << "]\n";
  std::cout << "-----------------------------------------\n";
  printBYTEarray(frm->frame, frm->length);
  std::cout << "\n\n";
}

/****************************************************************************/
//! Returns character from stdin, without having to press ENTER.
/*! Returns character from stdin, without having to press ENTER.
 *
 * @return int representing the pressed character.
 */
/****************************************************************************/
int getch()
{
  struct termios oldt, newt;
  int ch;

  __fpurge(stdin);
  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  ch = getchar();
  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);

  return ch;
}

/****************************************************************************/
//! Calculates checksum.
/*! Calculates checksum.
 *
 *	 ~C' = ~(C + (-m) + m') = ~C + (m - m') = ~C + m + ~m'
 *	  HC  - old checksum in header
 C   - one's complement sum of old header
 HC' - new checksum in header
 C'  - one's complement sum of new header
 m   - old value of a 16-bit field
 m'  - new value of a 16-bit field

 HC' = HC - ~m - m'

 HC'= = ~(~HC + ~m + m')
 *
 *
 *
 * @param oldChecksum - previous checksum
 * @param oldValue - pointer to previous value
 * @param newValue - pointer to new value
 * @param valueLength - value length
 * @return uint16_t - new checksum
 */
/****************************************************************************/
uint16_t calcUpdatedChecksum(uint16_t oldChecksum, const uint16_t *oldValue, const uint16_t *newValue,
                             unsigned int valueLength)
{
  uint32_t sum = oldChecksum;

  while (valueLength > 1)
  {

    sum = sum - ~(*oldValue++) - (*newValue++) - 1; //TODO -1 ?
    valueLength -= 2;
  }

  /*  Add left-over byte, if any */
  if (valueLength > 0)
  {
    sum = sum - ~(*(unsigned char *)oldValue) - (*(unsigned char *)newValue) - 1; //TODO -1 ?
  }

  /*  Fold 32-bit sum to 16 bits */
  while (sum >> 16)
    sum = (sum & 0xffff) + (sum >> 16);

  return sum;
}

/*
 * 	     |
 *	     V
 *  |0|1|2|3|4|5|6|7|
 *	     ^	   ^
 *       |--4--|
 *
 *  |0|0|1|1|1|1|0|0|
 *
 * 	     |
 *	     V
 *  |0|1|2|3|4|5|6|7|
 *	     ^ ^
 *       |2|
 *
 *  |0|0|1|1|0|0|0|0|
 *
 */
void zeroBits(unsigned char *byte, unsigned char fitstBitOffset, unsigned char lastBitOffset, bool isShiftBits,
              bool isZeroValue)
{
  //|1|1|1|1|1|1|1|1|
  //
  //|0|0|0|0|0|0|1|1|
  //	BYTE mask = 0xff >> (8-bitLength);
  //
  //	//|0|0|1|1|0|0|0|0|
  //	mask = mask << (8 - (bitOffset + bitLength));
  //
  //	if(isZeroValue)
  //		mask = ~mask;
  //
  //	//shift value back to lsb
  //	if(isShiftBits)
  //		byte[0] = ( byte[0] & mask ) >> (8 - (bitLength + bitOffset));
  //	else
  //		byte[0] = ( byte[0] & mask );

  //	BYTE mask = ( 0xff >> fitstBitOffset ) << (7 - lastBitOffset);
  BYTE mask = 0xff >> (7 - lastBitOffset + fitstBitOffset);
  mask = mask << (7 - lastBitOffset);

  //|0|0|1|1|0|0|0|0|
  //	mask = mask << lastBitOffset;

  if (isZeroValue)
    mask = ~mask;

  //shift value back to lsb
  if (isShiftBits)
    byte[0] = (byte[0] & mask) >> (7 - lastBitOffset);
  else
    byte[0] = (byte[0] & mask);

}

}
