/****************************************************************************
 *
 * File Name: FWD_KEY.cpp
 *
 * General Description: General description of the file.
 *
 * Created By: Alex

 , Creation Date: 05/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file FWD_KEY.cpp
 *\brief Implements the FWD_KEY Class.
 *
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
using namespace std;

#include "FWD_KEY.h"

using namespace boost;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
FWD_KEY::FWD_KEY()
{
  pos.setOffset(0);
  pos.setLength(1);
  pos.setFirstBitOffset(0);
  pos.setLastBitOffset(7);
  name = "Default FWD key";
  value = new BYTE[1];
  value[0] = (0x00);
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _length - key length
 * @param _name - key name
 */
/****************************************************************************/
FWD_KEY::FWD_KEY(unsigned int _offset, unsigned int _length, std::string _name)
{
  pos.setOffset(_offset);
  pos.setLength(_length);
  pos.setFirstBitOffset(0);
  pos.setLastBitOffset(7);
  name = _name;
  value = new BYTE[_length];
  memset(value, 0x00, _length);
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _pos - key position
 * @param _name - key name
 */
/****************************************************************************/
FWD_KEY::FWD_KEY(const Position &_pos, std::string _name)
{
  pos = _pos;
  name = _name;
  value = NULL;
  BYTE _value[] = {0x00};
  setFWD_KEYvalue(*this, _value, false);
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _length - key length
 * @param _name - key name
 * @param _value - key value
 */
/****************************************************************************/
FWD_KEY::FWD_KEY(const Position &_pos, std::string _name, BYTE *_value)
{
  pos = _pos;
  name = _name;
  value = NULL;
  setFWD_KEYvalue(*this, _value, false);
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _length - key length
 * @param _name - key name
 * @param _value - key value
 */
/****************************************************************************/
FWD_KEY::FWD_KEY(unsigned int _offset, unsigned int _length, std::string _name,
                 BYTE *_value)
{
  pos.setOffset(_offset);
  pos.setLength(_length);
  pos.setFirstBitOffset(0);
  pos.setLastBitOffset(7);
  name = _name;
  value = new BYTE[_length];
  memmove(value, _value, _length);
}

/****************************************************************************/
//! Copy Constructor.
/*! Copy Constructor.
 * @param other reference to the received FWD_KEY, that is going to be copied.
 */
/****************************************************************************/
FWD_KEY::FWD_KEY(const FWD_KEY &other)
{
  this->pos = other.pos;
  this->name = other.name;
  this->value = NULL;
  setFWD_KEYvalue(*this, other.value, false);
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
FWD_KEY::~FWD_KEY()
{
  if (this->value != NULL)
    delete[] this->value;
}

/****************************************************************************/
//! Implementation of the hash_value function.
/*! Implementation of the hash_value function needed for boosts unordered_map.
 * @param key reference to the received FWD_KEY, that is going to be hashed.
 * @return the hash value of key.
 */
/****************************************************************************/
size_t hash_value(const FWD_KEY& key)
{
  size_t hashValue = 0;

  hash_combine(hashValue, key.pos.getOffset());
  hash_combine(hashValue, key.pos.getLength());
  hash_combine(hashValue, key.pos.getFirstBitOffset());
  hash_combine(hashValue, key.pos.getLastBitOffset());
  hash_combine(hashValue, key.name);
  if (key.value != NULL)
  {
    for (unsigned int i = 0; i < key.pos.getLength(); ++i)
    {
      hash_combine(hashValue, key.value[i]);
    }
  }
  return hashValue;
}

/****************************************************************************/
//! Operator '=='.
/*! Implements the '==' operator.
 * @param other reference to the received FWD_KEY, that is going to be compared.
 * @return whether the Keys are equal or not.
 */
/****************************************************************************/
bool FWD_KEY::operator ==(const FWD_KEY& other) const
{
  return (this->pos == other.pos) && (this->name.compare(other.name) == 0)
      && (memcmp(this->value, other.value, this->pos.getLength() * sizeof(BYTE))
          == 0);
}

/****************************************************************************/
//! Operator '<<'.
/*! Implements the '<<' operator.
 * @param output reference to the output of operator.
 * @param key reference to the Key that is going to be outputted.
 * @return output of operator for cascaded output.
 */
/****************************************************************************/
ostream& operator<<(ostream& output, const FWD_KEY& key)
{
  output << "FWD_KEY: " << key.pos << ", name=" << key.name << ", value=";
  Auxiliary::printBYTEarray(key.value, key.pos.getLength());
  output << "]";
  return output;
}

/****************************************************************************/
//! Operator '<<'.
/*! Implements the '<<' operator.
 * @param output reference to the output of operator.
 * @param key pointer to the Key that is going to be outputted.
 * @return output of operator for cascaded output.
 */
/****************************************************************************/
ostream& operator<<(ostream& output, const FWD_KEY *key)
{
  output << "FWD_KEY: " << key->pos << ", name=" << key->name << ", value=";
  Auxiliary::printBYTEarray(key->getValue(), key->getLength());
  output << "]";
  return output;
}

/****************************************************************************/
//! Operator '<<'.
/*! Implements the '<<' operator.
 * @param output reference to the output of operator.
 * @param key pointer to the Key that is going to be outputted.
 * @return output of operator for cascaded output.
 */
/****************************************************************************/
ostream& operator<<(ostream& output, const fwdKeyMaskpair *key)
{
  output << key->first << ", " << key->second << "]";
  return output;
}

/****************************************************************************/
//! Operator '<<'.
/*! Implements the '<<' operator.
 * @param output reference to the output of operator.
 * @param key pointer to the Key that is going to be outputted.
 * @return output of operator for cascaded output.
 */
/****************************************************************************/
ostream& operator<<(ostream& output, const fwdKeyMaskpair &key)
{
  output << key.first << ", " << key.second << "]";
  return output;
}

/****************************************************************************/
//! Returns FWD_KEY's length.
/*! Returns FWD_KEY's length.
 * @return Returns FWD_KEY's length.
 */
/****************************************************************************/
unsigned int FWD_KEY::getLength() const
{
  return pos.getLength();
}

/****************************************************************************/
//! Returns FWD_KEY's position.
/*! Returns FWD_KEY's position.
 * @return Returns FWD_KEY's position.
 */
/****************************************************************************/
Position &FWD_KEY::getPosition()
{
  return pos;
}

/****************************************************************************/
//! Sets FWD_KEY's position.
/*! Sets FWD_KEY's value.
 * @param _pos - reference to the position to set.
 */
/****************************************************************************/
void FWD_KEY::setPosition(const Position &_pos)
{
  this->pos = _pos;
}

/****************************************************************************/
//! Returns FWD_KEY's name.
/*! Returns FWD_KEY's name.
 * @return Returns FWD_KEY's name.
 */
/****************************************************************************/
std::string FWD_KEY::getName() const
{
  return name;
}

/****************************************************************************/
//! Returns FWD_KEY's offset.
/*! Returns FWD_KEY's offset.
 * @return Returns FWD_KEY's offset.
 */
/****************************************************************************/
unsigned int FWD_KEY::getOffset() const
{
  return pos.getOffset();
}

/****************************************************************************/
//! Returns FWD_KEY's value.
/*! Returns FWD_KEY's value.
 * @return Returns FWD_KEY's value.
 */
/****************************************************************************/
unsigned char *FWD_KEY::getValue() const
{
  return value;
}

/****************************************************************************/
//! Sets FWD_KEY's value.
/*! Sets FWD_KEY's value.
 * @param _value - pointer to the value to set.
 */
/****************************************************************************/
void FWD_KEY::setValue(unsigned char *_value)
{
  setFWD_KEYvalue(*this, _value, false);
}

/****************************************************************************/
//! Applies mask to each FWD_KEY and stores them in fwdKeysWithMaskApplied.
/*! Applies mask to each FWD_KEY and stores them in fwdKeysWithMaskApplied.
 *
 * @param fwdKeyMaskVect - Forwarding key and mask pairs vector.
 * @param fwdKeysWithMaskApplied - Forwarding key with applied masks vector.
 */
/****************************************************************************/
void FWD_KEY::applyMask(vector<fwdKeyMaskpair *> &fwdKeyMaskVect, vector<
    FWD_KEY *> &fwdKeysWithMaskApplied)
{
  for (unsigned int i = 0; i < fwdKeyMaskVect.size(); i++)
  {
    FWD_KEY *toAdd = new FWD_KEY(fwdKeyMaskVect.at(i)->first);
    BYTE *buf = new BYTE[fwdKeyMaskVect.at(i)->first.getLength()];

    for (unsigned int j = 0; j < fwdKeyMaskVect.at(i)->first.getLength(); ++j)
    {
      buf[j] = (fwdKeyMaskVect.at(i)->first.getValue()[j])
          & (fwdKeyMaskVect.at(i)->second.getValue()[j]);
    }
    toAdd->setValue(buf);
    delete[] buf;
    fwdKeysWithMaskApplied.push_back(toAdd);
  }
}

/****************************************************************************/
//! Sets FWD_KEY value.
/*! Sets FWD_KEY value.
 * @param key - reference to the key to set.
 * @param _value - pointer to the value to set.
 * @param isShiftBits - if true the value in the last byte is shifted to lsb
 */
/****************************************************************************/
void FWD_KEY::setFWD_KEYvalue(FWD_KEY &key, unsigned char *_value,
                              bool isShiftBits)
{
  //free previous value
  if (key.value != NULL)
    delete[] key.value;

  if (key.pos.getLength() > 0)
  {
    if (key.pos.getFirstBitOffset() == 0 && key.pos.getLastBitOffset() == 7)
    { //size is in bytes only
      key.value = new BYTE[key.pos.getLength()];
      memmove(key.value, _value, key.pos.getLength());
    }
    else
    { //size is in bytes and in bits
      key.value = new BYTE[key.pos.getLength()];
      memmove(key.value, _value, key.pos.getLength());
      if (isShiftBits)
      {
        if (key.pos.getLength() == 1)
        {
          Auxiliary::zeroBits(&(key.value[0]), key.pos.getFirstBitOffset(),
                              key.pos.getLastBitOffset(), isShiftBits, false);
        }
        else
        {
          if (key.pos.getFirstBitOffset() != 0)
            Auxiliary::zeroBits(&(key.value[0]), key.pos.getFirstBitOffset(),
                                7, isShiftBits, false);

          if (key.pos.getLastBitOffset() != 7)
            Auxiliary::zeroBits(&(key.value[key.pos.getLength() - 1]), 0,
                                key.pos.getLastBitOffset(), isShiftBits, false);
        }
      }
    }
  }
  //	else
  //	{
  //		if(key.pos.getLastBitOffset() == 0)
  //		{	//size is zero
  //			key.value = NULL;
  //		}
  //		else
  //		{	//size is in bits only
  //			key.value = new BYTE[1];
  //			memmove(key.value,_value,1);
  //			if(isShiftBits)
  //				Auxiliary::zeroBits(&(key.value[0]),key.pos.getLastBitOffset(),key.pos.getFirstBitOffset(),isShiftBits,false);
  //		}
  //	}
}

/**
 * adjust mask to actual value size
 */
void FWD_KEY::adjustMaskToSize()
{
  if (pos.getLength() > 0)
  {
    if (pos.getFirstBitOffset() != 0 || pos.getLastBitOffset() != 7)
    { //size is in bytes and in bits
      if (pos.getLength() == 1)
      {
        Auxiliary::zeroBits(&(value[0]), pos.getFirstBitOffset(),
                            pos.getLastBitOffset(), false, false);
      }
      else
      {
        if (pos.getFirstBitOffset() != 0)
          Auxiliary::zeroBits(&(value[0]), pos.getFirstBitOffset(), 7, false,
                              false);

        if (pos.getLastBitOffset() != 7)
          Auxiliary::zeroBits(&(value[pos.getLength() - 1]), 0,
                              pos.getLastBitOffset(), false, false);
      }
    }
  }
}
