/****************************************************************************
 *
 * File Name: ID_to_FWD_Table_Impl.cpp
 *
 * General Description: Implements the Ident to Forward Table.
 *
 * Created By: Alex

 , Creation Date: 09/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file ID_to_FWD_Table_Impl.cpp
 *\brief Implements the Ident to Forward Table.
 *
 * Implements the Ident to Forward Table. This table stores the
 * (ID_KEYs,FWD_KEYs) pairs.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "ID_to_FWD_Table_Impl.h"
#include <string.h>

using namespace boost;
using namespace std;
/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
ID_to_FWD_Table_Impl::ID_to_FWD_Table_Impl()
{
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
ID_to_FWD_Table_Impl::~ID_to_FWD_Table_Impl()
{
  clearTable();
}

/****************************************************************************/
//! Removes entry from the FWD_KEYs Table.
/*! Removes entry from the FWD_KEYs Table.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Key Identifier.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string ID_to_FWD_Table_Impl::removeEntry(unsigned int PortID, vector<ID_KEY *> &idKeys)
{
#ifdef PATRICIA

#else
  portIDvectPair key(PortID, idKeys);
  FWD_KEYs.erase(key);
#endif
  return "OK";
}

/****************************************************************************/
//! Returns a vector of FWD_KEYs that match the (PortID,idKeys) pair.
/*! Returns a vector of FWD_KEYs that match the (PortID,idKeys) pair.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Identification key.
 * @param fwdKeysVect - a vector of FWD_KEYs that match the (PortID,idKeys) pair.
 */
/****************************************************************************/
void ID_to_FWD_Table_Impl::getFWD_KEYs(unsigned int PortID, vector<ID_KEY *> &idKeys,
                                       vector<fwdKeyMaskpair *> &fwdKeysVect)
{
#ifdef PATRICIA
  bytes b;
  b.byte = new unsigned char [MAX_KEY_SIZE];
  b.length = MAX_KEY_SIZE;
  memset(&b.byte[0], 0x00, b.length);

  //convert integer value to BYTEs
  b.byte[1] = (unsigned char)(PortID);
  b.byte[0] = (unsigned char)(PortID >> 8);

  unsigned int j = 2;
  for (vector<ID_KEY *>::iterator it = idKeys.begin();
      it != idKeys.end();
      ++it)
  {
    memmove(&b.byte[j], (*it)->getValue(), (*it)->getLength());
    j += (*it)->getLength();
  }

  //TODO check
  fwdKeysVect = FWD_KEYs.find(b);
#else
  portIDvectPair key(PortID, idKeys);

  pair<umultimap_idTofwdKeys::iterator, umultimap_idTofwdKeys::iterator> range = FWD_KEYs.equal_range(key);

  //  cout << key.second<<endl;
  //  printTable();
  //  for(umultimap_idTofwdKeys::iterator it = range.first;
  //          it != range.second;
  //          ++it)
  //  {
  //          fwdKeysVect.insert(fwdKeysVect.end(),(it->second).begin(),(it->second).end());
  //  }

  //return first match TODO change to regular map?
  if (range.first != range.second)
    fwdKeysVect.insert(fwdKeysVect.end(), range.first->second.begin(), range.first->second.end());
#endif

}


/****************************************************************************/
//! Returns a vector of FWD_KEYs that match the (PortID,idKeys) pair.
/*! Returns a vector of FWD_KEYs that match the (PortID,idKeys) pair.
 *
 * @param keys - Identification keys already stored in bytes.
 * @param fwdKeysVect - a vector of FWD_KEYs that match the (PortID,idKeys) pair.
 */
/****************************************************************************/
void ID_to_FWD_Table_Impl::getFWD_KEYs(bytes &keys,
                                       vector<fwdKeyMaskpair *> &fwdKeysVect)
{
#ifdef PATRICIA
//  cout << "Looking for: ";
//  maskedpatriciaNode<vector<fwdKeyMaskpair *> >::printBytes(keys);
  fwdKeysVect = FWD_KEYs.find(keys);
//  if(fwdKeysVect.size()>0)
//    cout << "Found!\n";
#endif
}


/****************************************************************************/
//! Adds a new entry to the FWD_KEYs Table.
/*! Adds a new entry to the FWD_KEYs Table. PortID is converted to two byte
 *  representation (ports higher than 65536 are unsupported).
 *
 * @param PortID - Port Identifier.
 * @param idKey - Key Identifier.
 * @param fwdKeysAndMasks - Forwarding Keys and masks vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string ID_to_FWD_Table_Impl::addNewEntry(unsigned int PortID, vector<ID_KEY *> &idKeys,
                                         vector<fwdKeyMaskpair *> &fwdKeysAndMasks)
{
  if(PortID < 65536)
  {
#ifdef PATRICIA
    bytes b;
    b.byte = new unsigned char [MAX_KEY_SIZE];
    b.length = MAX_KEY_SIZE;
    memset(&b.byte[0], 0x00, b.length);

    bytes mask;
    mask.byte = new unsigned char [MAX_KEY_SIZE];
    mask.length = MAX_KEY_SIZE;
    memset(&mask.byte[0], 0xff, mask.length);

    //convert integer value to BYTEs
    b.byte[1] = (unsigned char)(PortID);
    b.byte[0] = (unsigned char)(PortID >> 8);

    unsigned int j = 2;
    for (vector<ID_KEY *>::iterator it = idKeys.begin();
        it != idKeys.end();
        ++it)
    {
      memmove(&b.byte[j], (*it)->getValue(), (*it)->getLength());
      j += (*it)->getLength();
    }

    FWD_KEYs.insert(b,mask,fwdKeysAndMasks);
    delete [] b.byte;
    delete [] mask.byte;
#else
  FWD_KEYs.insert(pair<portIDvectPair, vector<fwdKeyMaskpair *> > (portIDvectPair(PortID, idKeys), fwdKeysAndMasks));
#endif
    return "ID to FWD table entry added successfully.";
  }
  else
    return "Port number is too high.";
}

/****************************************************************************/
//! Prints the ID_to_FWD Table.
/*! Prints the ID_to_FWD Table.
 */
/****************************************************************************/
void ID_to_FWD_Table_Impl::printTable()
{
  cout << "         ID_to_FWD Table:\n";
  cout << "------------------------------------------\n";

#ifdef PATRICIA
  FWD_KEYs.printTree();
  cout << endl;
#else
  for (umultimap_idTofwdKeys::const_iterator it = FWD_KEYs.begin(); it != FWD_KEYs.end(); ++it)
  {
    cout << "[Interface:" << dec << it->first.first << ",     ";

    for (unsigned int i = 0; i < it->first.second.size(); ++i)
    {
      cout << it->first.second.at(i) << ", ";
    }
    cout << "] ";
    for (unsigned int i = 0; i < it->second.size(); ++i)
    {
      cout << ",\n       key=" << it->second.at(i)->first << ",\n       mask=" << it->second.at(i)->second << "\n\n";
    }
    cout << "] ";
  }
#endif
}

/****************************************************************************/
//! Clears the ID_to_FWD Table.
/*! Clears the ID_to_FWD Table.
 */
/****************************************************************************/
void ID_to_FWD_Table_Impl::clearTable()
{

#ifdef PATRICIA
  vector<maskedpatriciaNode< vector<fwdKeyMaskpair *> > *> nodesVect;
  FWD_KEYs.fillVectorOfNodes(nodesVect);

  for (unsigned int i = 0; i < nodesVect.size(); ++i)
  {
    for (vector<fwdKeyMaskpair*>::const_iterator fkMaskItr = nodesVect[i]->value.begin(); fkMaskItr != nodesVect[i]->value.end(); ++fkMaskItr)
    {
      if ((*fkMaskItr) != NULL)
        delete (*fkMaskItr);
    }
  }
  FWD_KEYs.clearTree();
#else
  for (umultimap_idTofwdKeys::const_iterator it = FWD_KEYs.begin(); it != FWD_KEYs.end(); ++it)
  {
    for (vector<fwdKeyMaskpair*>::const_iterator fkMaskItr = (it->second).begin(); fkMaskItr != (it->second).end(); ++fkMaskItr)
    {
      if ((*fkMaskItr) != NULL)
        delete (*fkMaskItr);
    }
  }
  FWD_KEYs.clear();
#endif

}
