/****************************************************************************
 *
 * File Name: Initial_Lookup_Table_Impl.cpp
 *
 * General Description: Implements the Initial Lookup Table of the LLRE.
 *
 * Created By: Alex

 , Creation Date: 07/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Initial_Lookup_Table_Impl.cpp
 *\brief Implements the Initial Lookup Table of the LLRE.
 *
 * This class implements the Initial Lookup table, by using two hash tables,
 * one for active entries, and one for inactive entries. The results are
 * sorted in descending order so that entries with the most ID_KEYs are
 * matched first.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "Initial_Lookup_Table_Impl.h"

using namespace boost;
using namespace std;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
Initial_Lookup_Table_Impl::Initial_Lookup_Table_Impl()
{
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
Initial_Lookup_Table_Impl::~Initial_Lookup_Table_Impl()
{
  clearTable();
}

/****************************************************************************/
//! Adds a new entry to the ID_KEYs Table.
/*! Adds a new entry to the ID_KEYs Table.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Key Identifier in the frame.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Initial_Lookup_Table_Impl::addNewEntry(unsigned int PortID, bool isActive, unsigned int entryId, std::vector<
    ID_KEY*> &idKeys)
{
  if(PortID < 65536)
  {
    initialLookupResult *result = new initialLookupResult;
    result->entryId = entryId;
    result->idKeys = idKeys;

    if (isActive)
      insertKey(ID_KEYs,PortID,result);
    else
      insertKey(disabled_ID_KEYs,PortID,result);

    return "Initial lookup table entry added successfully.";
  }
  else
    return "Port number is too high.";
}

/****************************************************************************/
//! Returns a vector of ID_KEYs that match the PortID.
/*! Returns a vector of ID_KEYs that match the PortID, sorted in descending
 *	order.
 *
 * @param PortID - Port Identifier.
 * @param idKeysVect - a vector of ID_KEYs that match the PortID.
 */
/****************************************************************************/
void Initial_Lookup_Table_Impl::getID_KEYs(unsigned int PortID, vector<vector<ID_KEY*> > &idKeysVect)
{
  umap_id::const_iterator iterator = ID_KEYs.find(PortID);
  if (iterator != ID_KEYs.end())
  {
    for (unsigned int i = 0; i < iterator->second.size(); ++i) {
      idKeysVect.push_back(iterator->second[i]->idKeys);
    }
  }
}

/****************************************************************************/
//! Removes entry from the ID_KEYs Table.
/*! Removes entry from the ID_KEYs Table.
 *
 * @param PortID - Port Identifier.
 * @param ID_KEY_position - Key Identifier position in the frame.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Initial_Lookup_Table_Impl::removeEntry(unsigned int PortID)
{
  ID_KEYs.erase(PortID);
  return "OK";
}

/****************************************************************************/
//! Disables entry in the ID_KEYs Table.
/*! Disables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Initial_Lookup_Table_Impl::disableEntry(unsigned int entryId)
{
  bool isFound = false;
  stringstream strstr;
  umap_id::iterator it = ID_KEYs.begin();
  umap_id::iterator it_tmp;
  while (it != ID_KEYs.end() && !isFound)
  {
    it_tmp = it;
    ++it_tmp;
    vector<initialLookupResult *>::iterator vectIt = it->second.begin();
    vector<initialLookupResult *>::iterator vectItTmp;
    while (it != ID_KEYs.end() && vectIt != it->second.end() && !isFound)
    {
      if((*vectIt)->entryId == entryId)
      {
        isFound = true;
        insertKey(disabled_ID_KEYs,it->first,*vectIt);
        vectItTmp = vectIt;
        ++vectItTmp;
        it->second.erase(vectIt);
        vectIt = vectItTmp;
        if(it->second.size()==0)
        {
          ID_KEYs.erase(it);
          it = it_tmp;
        }
      }
      else
        ++vectIt;
    }
    if (!isFound)
      ++it;
  }
  if (isFound)
    strstr << "entry Id '" << entryId << "' disabled\n";
  else
    strstr << "entry Id '" << entryId << "' was not found or is already disabled.\n";


  return strstr.str();
}

/****************************************************************************/
//! Enables entry in the ID_KEYs Table.
/*! Enables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Initial_Lookup_Table_Impl::enableEntry(unsigned int entryId)
{
  bool isFound = false;
  stringstream strstr;
  umap_id::iterator it = disabled_ID_KEYs.begin();
  umap_id::iterator it_tmp;
  while (!isFound && it != disabled_ID_KEYs.end())
  {
    it_tmp = it;
    ++it_tmp;
    vector<initialLookupResult *>::iterator vectIt = it->second.begin();
    vector<initialLookupResult *>::iterator vectItTmp;
    while (it != disabled_ID_KEYs.end() && !isFound && vectIt != it->second.end())
    {
      if((*vectIt)->entryId == entryId)
      {
        isFound = true;
        insertKey(ID_KEYs,it->first,*vectIt);
        vectItTmp = vectIt;
        ++vectItTmp;
        it->second.erase(vectIt);
        vectIt = vectItTmp;
        if(it->second.size()==0)
        {
          disabled_ID_KEYs.erase(it);
          it = it_tmp;
        }
      }
      else
        ++vectIt;
    }
    if (!isFound)
      ++it;
  }
  if (isFound)
    strstr << "entry Id '" << entryId << "' enabled\n";
  else
    strstr << "entry Id '" << entryId << "' was not found or is already enabled.\n";
  return strstr.str();
}

/****************************************************************************/
//! Prints the Initial_Lookup Table.
/*! Prints the Initial_Lookup Table.
 */
/****************************************************************************/
void Initial_Lookup_Table_Impl::printTable()
{
  cout << "         Initial Lookup Table:\n";
  cout << "------------------------------------------\n";
  cout << "Active entries:\n";
  cout << "------------------------------------------\n";
  for (umap_id::const_iterator it = ID_KEYs.begin(); it != ID_KEYs.end(); ++it)
  {
    for (unsigned int i = 0; i < it->second.size(); ++i) {
      cout << "[PortID: " << dec << it->first << ", entryId: " << it->second[i]->entryId
          << ", " << it->second[i]->idKeys
          << "]\n";
    }
  }
  cout << endl;
  cout << "Inactive entries:\n";
  cout << "------------------------------------------\n";
  for (umap_id::const_iterator it = disabled_ID_KEYs.begin(); it != disabled_ID_KEYs.end(); ++it)
  {
    for (unsigned int i = 0; i < it->second.size(); ++i) {
      cout << "[PortID: " << dec << it->first << ", entryId: " << it->second[i]->entryId
          << ", " << it->second[i]->idKeys
          << "]\n";
    }
  }
  cout << endl;
}

/****************************************************************************/
//! Clears the Initial_Lookup Table.
/*! Clears the Initial_Lookup Table.
 */
/****************************************************************************/
void Initial_Lookup_Table_Impl::clearTable()
{
  for (umap_id::const_iterator it = ID_KEYs.begin(); it != ID_KEYs.end(); ++it)
  {
    for (unsigned int i = 0; i < it->second.size(); ++i)
      if (it->second[i] != NULL)
        delete it->second[i];
  }
  ID_KEYs.clear();

  for (umap_id::const_iterator it = disabled_ID_KEYs.begin(); it != disabled_ID_KEYs.end(); ++it)
  {
    for (unsigned int i = 0; i < it->second.size(); ++i)
          if (it->second[i] != NULL)
            delete it->second[i];
  }
  disabled_ID_KEYs.clear();
}


bool Initial_Lookup_Table_Impl::resultCompare(initialLookupResult const *lhs, initialLookupResult const *rhs )
{
    return lhs->idKeys.size() >= rhs->idKeys.size();
}

void Initial_Lookup_Table_Impl::insertKey(umap_id &table, unsigned int PortID, initialLookupResult *result)
{
  umap_id::iterator iterator = table.find(PortID);
  if (iterator == table.end())
  {//key doesn't exist, add new vector
    vector<initialLookupResult *> newVect;
    newVect.push_back(result);
    table[PortID] = newVect;
  }
  else
  {
    iterator->second.push_back(result);
    sort(iterator->second.begin(), iterator->second.end(), resultCompare );
  }
}
