/****************************************************************************
 *
 * File Name: LLRE.cpp
 *
 * General Description: General description of the file.
 *
 * Created By: Alex

 , Creation Date: 05/04/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file LLRE.cpp
 *\brief Implements the Low Level Rule Enforcer Class.
 *
 * This module processes all received frames, either from the Network
 * Driver or from the Dispatcher. Processing is performed by parsing the
 * received frame, searching the parsed fields in the appropriate lookup
 * tables and executing an action according to the search results.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "LLRE.h"
#include "../Dispatcher/Dispatcher.h"

using namespace std;
using namespace boost;

//#define _DEBUG

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
LLRE::LLRE()
{
  dispatcher = NULL;
  pthread_mutex_init(&initTable_lock, NULL);
  pthread_mutex_init(&idToFwdTable_lock, NULL);
  pthread_mutex_init(&actTable_lock, NULL);
  keyByte = new bytes(new unsigned char[MAX_KEY_SIZE], MAX_KEY_SIZE);
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param _dispatcher - Dispatcher pointer.
 */
/****************************************************************************/
LLRE::LLRE(Dispatcher *_dispatcher)
{
  dispatcher = _dispatcher;
  pthread_mutex_init(&initTable_lock, NULL);
  pthread_mutex_init(&idToFwdTable_lock, NULL);
  pthread_mutex_init(&actTable_lock, NULL);
  keyByte = new bytes(new unsigned char[MAX_KEY_SIZE], MAX_KEY_SIZE);

  startQueueProcessingThread();
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
LLRE::~LLRE()
{
  Frame endThread(0, NULL, 0);
  llreQ.push(&endThread);

  //wait for processing thread to stop
  pthread_join(threadPtr, NULL);

  pthread_mutex_destroy(&initTable_lock);
  pthread_mutex_destroy(&idToFwdTable_lock);
  pthread_mutex_destroy(&actTable_lock);

  Frame *frm = NULL;
  while (!llreQ.empty())
  {
    llreQ.try_pop(frm);
    if (frm != NULL && frm->frame != NULL)
    {
      delete[] frm->frame;
      delete frm;
    }
  }
  if(keyByte->byte != NULL)
  {
    delete [] keyByte->byte;
    delete keyByte;
  }
}

/****************************************************************************/
//! Adds a new frame to the processing queue.
/*! Adds a new frame to the processing queue. This queue is used to pass
 * frames between GFEs modules, for thread safety.
 *
 * @param frame - new frame to process.
 */
/****************************************************************************/
void LLRE::pushFrameToQueue(Frame *frame)
{
  llreQ.push(frame);
}

/****************************************************************************/
//! Processes the received frame.
/*! Processes the received frame:
 * 	(1) Port_ID-->ID_KEY position
 * 	(2) (Port_ID,ID_KEY)-->FWD_KEY position
 * 	(3) FWD_KEY-->Actions
 * 	(4) Perform actions
 * 	(5) Forward to Dispatcher
 *
 * @param frame - frame structure to forward.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::processFrame(Frame *frm)
{
  vector<vector<ID_KEY *> > idKeys;
  vector<fwdKeyMaskpair *> fwdKeys;
  vector<bytes*> keyBytes;
#ifdef PATRICIA_ACTIONS
  vector<vector<Action*> > *actions=NULL;
#else
  vector<vector<Action*> > actions;
#endif

#ifdef _DEBUG
  //cout << "LLRE Received frame on interface: "<<frm->interface<<endl;
//  Auxiliary::printFrame(true,frm);
#endif

  try
  {
    //port -> ID_KEYs
    get_Initial_Table_ID_KEYs(frm->interface, idKeys);

    if (idKeys.size() > 0)
    { // this port has ID_KEYs
      //Retrieve the values of the ID_KEYs from current frame
#ifdef PATRICIA
      getID_KEYvaluesFromFrame(idKeys, frm, keyBytes);
#else
      getID_KEYvaluesFromFrame(idKeys, frm);
#endif


#ifdef _DEBUG
      std::cout<< "	ID_KEYs with values: \n";
      for (unsigned int i = 0; i < idKeys.size(); ++i)
      {
        cout<<"		[-----------------------------]\n";
        for (std::vector<ID_KEY *>::const_iterator it = idKeys.at(i).begin();
            it != idKeys.at(i).end();
            ++it)
        {
          std::cout<<"	"<< *it <<std::endl;
        }
      }
#endif

      // (port,ID_KEYs) -> FWD_KEYs
#ifdef PATRICIA
      get_ID_to_FWD_Table_FWD_KEYs(keyBytes, fwdKeys);
#else
      get_ID_to_FWD_Table_FWD_KEYs(frm->interface, idKeys, fwdKeys);
#endif

      if (fwdKeys.size() > 0)
      { //FWD_KEYs found for these (port,ID_KEYs)
        //Retrieve the values of the FWD_KEYs from current frame
#ifdef PATRICIA_ACTIONS
//        bytes *keyByte = getFWD_KEYvaluesFromFrameAsByte(fwdKeys, frm);
        getFWD_KEYvaluesFromFrameAsByte(fwdKeys, frm);
#else
        getFWD_KEYvaluesFromFrame(fwdKeys, frm);
#endif

#ifdef _DEBUG
        std::cout<< "\nInterface: "<<frm->interface<<std::endl;
        std::cout<< "FWD_KEYs with values: \n";
        for (std::vector<fwdKeyMaskpair *>::const_iterator it = fwdKeys.begin();
            it != fwdKeys.end();
            ++it)
        {
          std::cout<<"	"<< (*it)->first<<" , "<< (*it)->second <<std::endl;
        }
        std::cout<<endl;
#ifdef PATRICIA_ACTIONS
        std::cout<< "   keyByte: \n";
        maskedpatriciaNode<int>::printBytes(*keyByte);
        std::cout<< "\n";
#endif
#endif

        // FWD_KEYs -> Actions
#ifdef PATRICIA_ACTIONS
//        get_Action_Table_Actions(keyByte, actions);
        actions = get_Action_Table_Actions(keyByte);
#else
        get_Action_Table_Actions(fwdKeys, actions);
#endif
        frm->srcInterface = frm->interface;//save src interface
        frm->interface = CE_PORT;//default port
#ifdef PATRICIA_ACTIONS
        if (actions->size() > 0)
        { //Actions exist for these FWD_KEYs
          performAllActions(*actions, frm);
//          actions->clear();
        }
#else
        if (actions.size() > 0)
        { //Actions exist for these FWD_KEYs
          performAllActions(actions, frm);
          actions.clear();
        }
#endif
        else
        {
#ifdef _DEBUG
          cout << "No actions for these FWD_KEYs\n";
#endif
          frm->interface = CE_PORT;
          //cout << forwardFrame(frm) << endl;
          forwardFrame(frm);
        }
      }
      else
      {
        //cout << "No matching FWD_KEYs\n";
        frm->interface = CE_PORT;
        //cout << forwardFrame(frm) << endl;
        forwardFrame(frm);
      }
    }
    else
    {
      //cout << "No matching ID_KEYs\n";
      frm->interface = CE_PORT;
      //cout << forwardFrame(frm) << endl;
      forwardFrame(frm);
    }
  }
  catch (GFEexception& e)
  {
    cerr << "LLRE exception: " << e.getMessage() << endl;
  }

  return "DONE.\n";
}

/****************************************************************************/
//! Adds a new entry to the Initial_Lookup Table.
/*! Adds a new entry to the Initial_Lookup Table.
 *
 * @param PortID - Port Identifier.
 * @param isActive - true if the entry is active, false otherwise
 * @param entryId - the ID number of the entry.
 * @param idKeys - Identifier Keys in the frame.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::add_Initial_Lookup_Table_Entry(unsigned int PortID, bool isActive,
                                            unsigned int entryId, std::vector<
                                                ID_KEY*> &idKeys)
{
  pthread_mutex_lock(&initTable_lock);
  initTable.addNewEntry(PortID, isActive, entryId, idKeys);
  pthread_mutex_unlock(&initTable_lock);
  return "OK";
}

/****************************************************************************/
//! Adds a new entry to the ID_to_FWD Table.
/*! Adds a new entry to the ID_to_FWD Table.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Key Identifier.
 * @param fwdKeysAndMasks - Forwarding keys and masks in the frame.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::add_ID_to_FWD_Entry(unsigned int PortID, vector<ID_KEY *> &idKeys,
                                 vector<fwdKeyMaskpair *> &fwdKeysAndMasks)
{
  pthread_mutex_lock(&idToFwdTable_lock);
  idToFwdTable.addNewEntry(PortID, idKeys, fwdKeysAndMasks);
  pthread_mutex_unlock(&idToFwdTable_lock);
  return "OK";
}

/****************************************************************************/
//! Adds a new entry to the Initial_Lookup Table.
/*! Adds a new entry to the Initial_Lookup Table.
 *
 * @param fwdKeysAndMasks - Forwarding keys and masks in the frame.
 * @param actions - Actions to perform.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::add_Action_Table_Entry(unsigned int interface, vector<fwdKeyMaskpair *> &fwdKeyMaskVect,
                                    vector<Action*> &actions)
{
  pthread_mutex_lock(&actTable_lock);
  actTable.addNewEntry(interface,fwdKeyMaskVect,actions);
  pthread_mutex_unlock(&actTable_lock);
  return "OK";
}

/****************************************************************************/
//! Disables entry in the ID_KEYs Table.
/*! Disables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::disableEntry(unsigned int entryId)
{
  pthread_mutex_lock(&initTable_lock);
  string res(initTable.disableEntry(entryId));
  pthread_mutex_unlock(&initTable_lock);
  return res;
}

/****************************************************************************/
//! Enables entry in the ID_KEYs Table.
/*! Enables entry in the ID_KEYs Table..
 *
 * @param entryId - entry id number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::enableEntry(unsigned int entryId)
{
  pthread_mutex_lock(&initTable_lock);
  string res(initTable.enableEntry(entryId));
  pthread_mutex_unlock(&initTable_lock);
  return res;
}

/****************************************************************************/
//! Prints the Initial_Lookup Table.
/*! Prints the Initial_Lookup Table.
 */
/****************************************************************************/
void LLRE::printInitialTable()
{
  pthread_mutex_lock(&initTable_lock);
  initTable.printTable();
  pthread_mutex_unlock(&initTable_lock);
}

/****************************************************************************/
//! Clears the Initial_Lookup Table.
/*! Clears the Initial_Lookup Table.
 */
/****************************************************************************/
void LLRE::clearInitialTable()
{
  pthread_mutex_lock(&initTable_lock);
  initTable.clearTable();
  pthread_mutex_unlock(&initTable_lock);
}

/****************************************************************************/
//! Prints the ID_to_FWD Table.
/*! Prints the ID_to_FWD Table.
 */
/****************************************************************************/
void LLRE::printIDtoFWDTable()
{
  pthread_mutex_lock(&idToFwdTable_lock);
  idToFwdTable.printTable();
  pthread_mutex_unlock(&idToFwdTable_lock);
}

/****************************************************************************/
//! Clears the ID_to_FWD Table.
/*! Clears the ID_to_FWD Table.
 */
/****************************************************************************/
void LLRE::clearIDtoFWDTable()
{
  pthread_mutex_lock(&idToFwdTable_lock);
  idToFwdTable.clearTable();
  pthread_mutex_unlock(&idToFwdTable_lock);
}

/****************************************************************************/
//! Prints the Action Table.
/*! Prints the Action Table.
 */
/****************************************************************************/
void LLRE::printActionTable()
{
  pthread_mutex_lock(&actTable_lock);
  actTable.printTable();
  pthread_mutex_unlock(&actTable_lock);
}

/****************************************************************************/
//! Clears the Action Table.
/*! Clears the Action Table.
 */
/****************************************************************************/
void LLRE::clearActionTable()
{
  pthread_mutex_lock(&actTable_lock);
  actTable.clearTable();
  pthread_mutex_unlock(&actTable_lock);
}

/****************************************************************************/
//! Returns a vector of ID_KEY_positions that match the PortID.
/*! Returns a vector of ID_KEY_positions that match the PortID.
 *
 * @param PortID - Port Identifier.
 * @param idKeysVect - a vector of ID_KEY_positions that match the PortID.
 */
/****************************************************************************/
void LLRE::get_Initial_Table_ID_KEYs(unsigned int PortID,
                                     vector<vector<ID_KEY*> > &idKeysVect)
{
  pthread_mutex_lock(&initTable_lock);
  initTable.getID_KEYs(PortID, idKeysVect);
  pthread_mutex_unlock(&initTable_lock);
}

/****************************************************************************/
//! Returns a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
/*! Returns a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Identification key.
 * @param fwdKeysVect - a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
 */
/****************************************************************************/
void LLRE::get_ID_to_FWD_Table_FWD_KEYs(unsigned int PortID, vector<vector<
    ID_KEY *> > &idKeys, vector<fwdKeyMaskpair *> &fwdKeysVect)
{
  //TODO change vect<vect<id>> to [port|keys|port|keys|...|port|keys]
  bool found = false;
  for (unsigned int i = 0; i < idKeys.size() && !found; ++i)
  {
    pthread_mutex_lock(&idToFwdTable_lock);
    idToFwdTable.getFWD_KEYs(PortID, idKeys.at(i), fwdKeysVect);
    pthread_mutex_unlock(&idToFwdTable_lock);
    if (fwdKeysVect.size() > 0)
      found = true;
  }
}

/****************************************************************************/
//! Returns a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
/*! Returns a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
 *
 * @param PortID - Port Identifier.
 * @param idKey - Identification key.
 * @param fwdKeysVect - a vector of FWD_KEY_positions that match the (PortID,idKey) pair.
 */
/****************************************************************************/
void LLRE::get_ID_to_FWD_Table_FWD_KEYs(vector<bytes*> &keyBytes, vector<
    fwdKeyMaskpair *> &fwdKeysVect)
{
  //TODO change vect<vect<id>> to [port|keys|port|keys|...|port|keys]
  bool found = false;
  unsigned int i = 0;
  for (; i < keyBytes.size() && !found; ++i)
  {
    pthread_mutex_lock(&idToFwdTable_lock);
    idToFwdTable.getFWD_KEYs(*(keyBytes.at(i)), fwdKeysVect);
    pthread_mutex_unlock(&idToFwdTable_lock);

    delete [] keyBytes[i]->byte;
    delete keyBytes[i];

    if (fwdKeysVect.size() > 0)
      found = true;
  }

  //free the rest of the keyBytes
  for (; i < keyBytes.size(); ++i)
  {
    delete [] keyBytes[i]->byte;
    delete keyBytes[i];
  }
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKey.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKey.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks.
 * @param actionsVect - a vector of vectors containing Actions corresponding to
 * 					   the fwdKey.
 */
/****************************************************************************/
void LLRE::get_Action_Table_Actions(vector<fwdKeyMaskpair *> &fwdKeyMaskVect,
                                    vector<vector<Action*> > &actionsVect)
{
  pthread_mutex_lock(&actTable_lock);
  actTable.getActions(fwdKeyMaskVect, actionsVect);
  pthread_mutex_unlock(&actTable_lock);
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKey.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKey.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks.
 * @param actionsVect - a vector of vectors containing Actions corresponding to
 *                                         the fwdKey.
 */
/****************************************************************************/
void LLRE::get_Action_Table_Actions(bytes* keyByte,
                                    vector<vector<Action*> > &actionsVect)
{
  pthread_mutex_lock(&actTable_lock);
  actTable.getActions(*keyByte, actionsVect);
  pthread_mutex_unlock(&actTable_lock);
//  if(keyByte->byte != NULL)
//  {
//    delete [] keyByte->byte;
//    delete keyByte;
//  }
}

/****************************************************************************/
//! Returns a vector of vectors containing Actions corresponding to the fwdKey.
/*! Returns a vector of vectors containing Actions corresponding to the fwdKey.
 *
 * @param fwdKeyMaskVect - Forwarding keys and masks.
 * @param actionsVect - a vector of vectors containing Actions corresponding to
 *                                         the fwdKey.
 */
/****************************************************************************/
vector<vector<Action*> > *LLRE::get_Action_Table_Actions(bytes* _keyByte)
{
  pthread_mutex_lock(&actTable_lock);
  vector<vector<Action*> > *result = actTable.returnActions(*keyByte);
  pthread_mutex_unlock(&actTable_lock);
  return result;
//  if(keyByte->byte != NULL)
//  {
//    delete [] keyByte->byte;
//    delete keyByte;
//  }
}

/****************************************************************************/
//! Returns the Dispatcher pointer.
/*! Returns the Dispatcher pointer.
 *
 * @return Returns the Dispatcher pointer.
 */
/****************************************************************************/
Dispatcher *LLRE::getDispatcher() const
{
  return dispatcher;
}

/****************************************************************************/
//! Sets the Dispatcher pointer.
/*! Sets the Dispatcher pointer.
 *
 * @param ndriver - the Dispatcher pointer.
 */
/****************************************************************************/
void LLRE::setDispatcher(Dispatcher *dispatcher)
{
  this->dispatcher = dispatcher;
}

/****************************************************************************/
//! Forwards the received frame to the Dispatcher.
/*! Forwards the received frame to the Dispatcher.
 *
 * @param interface - the interface on which the frame is to be sent.
 * @param frame - frame to forward.
 * @param length - frame length.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string LLRE::forwardFrame(Frame *frm)
{
  if (dispatcher != NULL)
  {
#ifdef _DEBUG
    std::cout << "LLRE: Forwarding to Dispatcher(interface: 0x" << frm->interface <<") " <<endl;
#endif
    dispatcher->pushFrameToQueue(frm);
    return "Sending to Dispatcher queue.";
  }
  else
    return "No Dispatcher assigned to this LLRE.";
}

/****************************************************************************/
//! Performs all the actions in the 'actions' vector.
/*! Performs all the actions in the 'actions' vector.
 *
 * @param actions - actions to perform.
 * @param frame - frame to forward.
 */
/****************************************************************************/
void LLRE::performAllActions(vector<vector<Action*> >& actions, Frame *frm)
{
  //cout<< "Performing actions:\n";
  for (vector<vector<Action*> >::const_iterator ii = actions.begin(); ii
      != actions.end(); ++ii)
  {
    //multicast
    if (*ii != actions.back() && actions.size() > 1)
    {
      Frame *multicastFrame = new Frame(*frm);
      multicastFrame->frame = new u_char[frm->length];
      memcpy(multicastFrame->frame, frm->frame, frm->length);

      for (vector<Action*>::const_iterator jj = (*ii).begin(); jj != (*ii).end(); ++jj)
      {
        //cout<<(**jj)<<endl; //prints action
        (*jj)->performAction(multicastFrame);
      }

#ifdef _DEBUG
//      Auxiliary::printFrame(false,multicastFrame);
#endif

      //forward frame to Dispatcher
      //cout << forwardFrame(multicastFrame) << endl;
      forwardFrame(multicastFrame);
    }
    else
    {
      for (vector<Action*>::const_iterator jj = (*ii).begin(); jj != (*ii).end(); ++jj)
      {
        //cout<<(**jj)<<endl; //prints action
        (*jj)->performAction(frm);
      }

#ifdef _DEBUG
//      Auxiliary::printFrame(false,frm);
#endif

      //forward frame to Dispatcher
      //cout << forwardFrame(frm) << endl;
      forwardFrame(frm);
    }
  }
}

/****************************************************************************/
//! Inserts values into keys from the frame.
/*! Inserts values into keys from the frame.
 *
 * @param idKeys - ID_KEYs vector.
 * @param frame - frame.
 */
/****************************************************************************/
void LLRE::getID_KEYvaluesFromFrame(vector<vector<ID_KEY*> > &idKeys, Frame *frm)
{
  for (unsigned int i = 0; i < idKeys.size(); ++i)
  {

    for (vector<ID_KEY*>::const_iterator ii = idKeys.at(i).begin(); ii
        != idKeys.at(i).end(); ++ii)
    {
      if ((*ii)->getOffset() + (*ii)->getLength() > frm->length)
        throw GFEexception("invalid ID_KEY offset");
      else
        (*ii)->setValue(&frm->frame[(*ii)->getOffset()]);
    }
  }
}

/****************************************************************************/
//! Inserts values into keys from the frame.
/*! Inserts values into keys from the frame.
 *
 * @param idKeys - ID_KEYs vector.
 * @param frame - frame.
 */
/****************************************************************************/
void LLRE::getID_KEYvaluesFromFrame(vector<vector<ID_KEY*> > &idKeys, Frame *frm,
                                    vector<bytes *> &keysAsBytes)
{
  for (unsigned int i = 0; i < idKeys.size(); ++i)
  {
    bytes *keyByte = new bytes(new unsigned char[MAX_KEY_SIZE], MAX_KEY_SIZE);

    keyByte->byte[1] = (unsigned char)(frm->interface);
    keyByte->byte[0] = (unsigned char)(frm->interface >> 8);
    unsigned int index = 2;
    for (vector<ID_KEY*>::const_iterator ii = idKeys.at(i).begin(); ii
        != idKeys.at(i).end(); ++ii)
    {
      if ((*ii)->getOffset() + (*ii)->getLength() > frm->length)
        throw GFEexception("invalid ID_KEY offset");
      else
      {
        memmove(&keyByte->byte[index], &frm->frame[(*ii)->getOffset()],
                (*ii)->getLength());
        index += (*ii)->getLength();
      }
    }
    //TODO is this necessary ?
    memset(&keyByte->byte[index], 0x00, MAX_KEY_SIZE - index);
    keysAsBytes.push_back(keyByte);
  }
}

/****************************************************************************/
//! Inserts values into keys from the frame.
/*! Inserts values into keys from the frame.
 *
 * @param idKeys - ID_KEYs vector.
 * @param frame - frame.
 */
/****************************************************************************/
void LLRE::getFWD_KEYvaluesFromFrame(vector<fwdKeyMaskpair*>& fwdKeys, Frame *frm)
{
  for (vector<fwdKeyMaskpair*>::const_iterator ii = fwdKeys.begin(); ii
      != fwdKeys.end(); ++ii)
  {
    if ((*ii)->first.getOffset() + (*ii)->first.getLength() > frm->length)
      throw GFEexception("invalid FWD_KEY offset");
    else
    {
      (*ii)->first.setValue(&frm->frame[(*ii)->first.getOffset()]);
      BYTE *buf = new BYTE[(*ii)->first.getLength()];
      for (unsigned int i = 0; i < (*ii)->first.getLength(); ++i)
      {
        buf[i] = ((*ii)->first.getValue()[i]) & ((*ii)->second.getValue()[i]);
      }
      (*ii)->first.setValue(buf);
      delete[] buf;
    }
  }
}

/****************************************************************************/
//! Inserts values into keys from the frame.
/*! Inserts values into keys from the frame.
 *
 * @param idKeys - ID_KEYs vector.
 * @param frame - frame.
 */
/****************************************************************************/
bytes *LLRE::getFWD_KEYvaluesFromFrameAsByte(vector<fwdKeyMaskpair*>& fwdKeys, Frame *frm)
{
//  bytes *keyByte = new bytes(new unsigned char[MAX_KEY_SIZE], MAX_KEY_SIZE);
  memset(keyByte->byte,0x00,MAX_KEY_SIZE);
  unsigned int index = 3;
  for (vector<fwdKeyMaskpair*>::const_iterator ii = fwdKeys.begin(); ii
      != fwdKeys.end(); ++ii)
  {
    if ((*ii)->first.getOffset() + (*ii)->first.getLength() > frm->length)
      throw GFEexception("invalid FWD_KEY offset");
    else
    {
	keyByte->byte[0] = 0x00; //needed for correct function of the tree (root value is always 0x00)
      //convert integer value to BYTEs
      keyByte->byte[2] = (unsigned char)(frm->interface);
      keyByte->byte[1] = (unsigned char)(frm->interface >> 8);
      //add offset to distinguish between fwdKeys with same values but at different offsets
      keyByte->byte[index++] = (unsigned char)( (*ii)->first.getPosition().getOffset() );
      keyByte->byte[index++] = (unsigned char)( (*ii)->first.getPosition().getOffset() >> 8);
      keyByte->byte[index++] = (unsigned char)( (*ii)->first.getPosition().getOffset() >> 8);
      keyByte->byte[index++] = (unsigned char)( (*ii)->first.getPosition().getOffset() >> 8);

//      (*ii)->first.setValue(&frm->frame[(*ii)->first.getOffset()]);
//      BYTE *buf = new BYTE[(*ii)->first.getLength()];
      for (unsigned int i = 0; i < (*ii)->first.getLength(); ++i)
      {
//        buf[i] = ((*ii)->first.getValue()[i]) & ((*ii)->second.getValue()[i]);
        keyByte->byte[index] = frm->frame[(*ii)->first.getOffset() + i] & ((*ii)->second.getValue()[i]);
        ++index;
      }
//      (*ii)->first.setValue(buf);
//      delete[] buf;
    }
  }
  return keyByte;
}

/****************************************************************************/
//! Starts a new queue processing thread.
/*! Starts a new queue processing thread, to serve the frame queue.
 */
/****************************************************************************/
void LLRE::startQueueProcessingThread()
{
  threadPtr = 0;
  if (pthread_create(&threadPtr, NULL, processQueue, (void *)this))
    throw GFEexception("Failed to start new LLRE thread.");
}

/****************************************************************************/
//! Processing thread function, to serve the frames queue.
/*! Processing thread function, to serve the frames queue. This thread is
 * terminated by adding a NULL frame to queue.
 *
 * @param _llre - an LLRE object.
 */
/****************************************************************************/
void *LLRE::processQueue(void *_llre)
{
  LLRE *llre = (LLRE *)_llre;
  Frame *frm = NULL;
  bool processFlag = true;
  while (processFlag)
  {
    llre->llreQ.wait_and_pop(frm);

    if (frm != NULL && frm->frame != NULL)
      llre->processFrame(frm);
    else
      processFlag = false;
  }
  pthread_exit(0);
  return 0;
}
