/****************************************************************************
 *
 * File Name: NetworkDriver.cpp
 *
 * General Description: Implements the Network Driver class.
 *
 * Created By: Alex, Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file NetworkDriver.cpp
 *\brief Implements the Network Driver class.
 *
 * This module interacts directly with the network interfaces. Its role is to
 * capture frames on each interface, send them to the LLRE, and to forward
 * frames from the Dispatcher to the network.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "NetworkDriver.h"
#include <fcntl.h>

using namespace std;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
NetworkDriver::NetworkDriver()
{
  try
  {
    if (pcap_findalldevs(&pcap_interfaces, errbuf) == -1)
    {
      throw NetworkDriverException(errbuf);
    }

    if (lookForInterfaces() == 0)
    {
      throw NetworkDriverException("No interfaces found, consider running in sudo mode.");
    }
  }
  catch (NetworkDriverException& e)
  {
    cerr << "NetworkDriver exception: " << e.getMessage() << endl;
  }

  for (unsigned int i = 0; i < pcap_c_descriptors.size(); ++i)
  {
    pthread_mutex_init(pcap_send_desc_lock.at(i), NULL);
  }

  pthread_mutex_init(&interface_names_lock, NULL);
  pthread_mutex_init(&activeSendInterfaces_lock, NULL);
  pthread_mutex_init(&activeCaptInterfaces_lock, NULL);

  startQueueProcessingThread();
  printInterfaces();
  received = 0;
  dropped = 0;
  if_dropped = 0;
  sent = 0;
}

/****************************************************************************/
//! Default Destructor.
/*! Default Destructor.
 */
/****************************************************************************/
NetworkDriver::~NetworkDriver()
{
  for (unsigned int i = 0; i < isCapture.size(); ++i)
  {
    isCapture[i] = false;
  }

  //sending null frame to stop queue processing thread
  Frame endThread(0, NULL, 0);
  ndriverQ.push(&endThread);

  pthread_mutex_lock(&activeCaptInterfaces_lock);

  for (unsigned int i = 0; i < pcap_c_descriptors.size(); ++i)
  {
    if (activeCaptInterfaces.at(i) == 1)
    {
      //wakes up select() in capture thread
      if (write(captureThreadFds.at(i), "1", 1) == -1)
        throw NetworkDriverException("Error writing to comm. pipe.");
      //			pcap_breakloop(pcap_descriptors.at(i));

      //wait for capture thread to stop
      pthread_join(thread_ids.at(i), NULL);
      close(captureThreadFds.at(i));
    }
  }

  pcap_freealldevs(pcap_interfaces);

  //wait for processing thread to stop
  pthread_join(threadPtr, NULL);

  for (unsigned int i = 0; i < pcap_c_descriptors.size(); ++i)
  {
    if (activeCaptInterfaces.at(i) > 0)
    {
      if (pcap_stats(pcap_c_descriptors.at(i), &ps) < 0)
      {
        cerr << "\nError setting the mode.\n";
      }
      //received includes packets sent by me
      received += ps.ps_recv;
      dropped += ps.ps_drop;
      if_dropped += ps.ps_ifdrop;
      //			sent += sent_vect_stats.at(i);

      pcap_close(pcap_c_descriptors.at(i));
    }
  }
  pthread_mutex_unlock(&activeCaptInterfaces_lock);

  for (unsigned int i = 0; i < pcap_s_descriptors.size(); ++i)
  {
    if (activeSendInterfaces.at(i) > 0)
    {
      sent += sent_vect_stats.at(i);
      pcap_close(pcap_s_descriptors.at(i));
    }
  }
  //	received -= sent;

  //send NULL frame to stop queue processing
  Frame *frm = NULL;
  while (!ndriverQ.empty())
  {
    ndriverQ.try_pop(frm);
    if (frm != NULL && frm->frame != NULL)
    {
      delete[] frm->frame;
      delete frm;
    }
  }

  for (unsigned int i = 0; i < pcap_c_descriptors.size(); ++i)
  {
    pthread_mutex_destroy(pcap_send_desc_lock.at(i));
    delete pcap_send_desc_lock.at(i);
  }

  pthread_mutex_destroy(&interface_names_lock);
  pthread_mutex_destroy(&activeSendInterfaces_lock);
  pthread_mutex_destroy(&activeCaptInterfaces_lock);

  cout << "Received " << dec << received << " packets.\n";
  cout << "Sent " << dec << sent << " packets.\n";
  cout << "Dropped " << dec << dropped << " packets.\n";
  cout << "Interface Dropped " << dec << if_dropped << " packets.\n";
}

/****************************************************************************/
//! Stops the packet capture thread on the passed interface.
/*! Stops the packet capture thread on the passed interface.
 *
 * @param interface - the interface on which to capture packets.
 */
/****************************************************************************/
string NetworkDriver::StopCapture(unsigned int interface)
{
  stringstream strstr;
  if (interface > interface_names.size())
  {
    strstr << "Interface number: " << interface << " is invalid.";
    throw NetworkDriverException(strstr.str());
  }

  pthread_mutex_lock(&activeCaptInterfaces_lock);

  if (activeCaptInterfaces.at(interface) == 0)
  {
    pthread_mutex_unlock(&activeCaptInterfaces_lock);
    throw NetworkDriverException("Attempting to stop capture on disabled interface.");
  }
  isCapture[interface] = false;
  activeCaptInterfaces[interface] = 0;

  pthread_mutex_unlock(&activeCaptInterfaces_lock);

  if (write(captureThreadFds.at(interface), "1", 1) == -1)
    throw NetworkDriverException("Error writing to comm. pipe.");

  //wait for capture thread to stop
  pthread_join(thread_ids.at(interface), NULL);

  //update stats
  if (pcap_stats(pcap_c_descriptors.at(interface), &ps) < 0)
  {
    cerr << "\nError setting the mode.\n";
  }
  //received includes packets sent by me
  received += ps.ps_recv;
  dropped += ps.ps_drop;
  if_dropped += ps.ps_ifdrop;

  close(captureThreadFds.at(interface));
  pcap_close(pcap_c_descriptors.at(interface));
  strstr << "Capture on interface number: " << interface << " stopped.\n";
  return strstr.str();
}

/****************************************************************************/
//! Stops the packet capture thread on all active interfaces.
/*! Stops the packet capture thread on all active interfaces.
 *
 */
/****************************************************************************/
string NetworkDriver::StopAllActiveCapture()
{
  stringstream strstr;
  int zzz = 0;
  for (unsigned int i = 0; i < pcap_c_descriptors.size(); ++i)
  {
    try
    {
      strstr << StopCapture(i);
    }
    catch (NetworkDriverException e)
    {
      //			strstr << "Network driver error: " << e.getMessage() << endl;
      //doesnt work without something here
      zzz++;
    }
  }
  return strstr.str();
}

/****************************************************************************/
//! Adds a new frame to the processing queue.
/*! Adds a new frame to the processing queue. This queue is used to pass
 * frames between GFEs modules, for thread safety.
 *
 * @param frame - new frame to process.
 */
/****************************************************************************/
void NetworkDriver::pushFrameToQueue(Frame *frame)
{
  ndriverQ.push(frame);
}

/****************************************************************************/
//! Starts a packet capture thread on the passed interface.
/*! Starts a packet capture thread on the passed interface, a callback function
 * is called each time a packet arrives.
 *
 * @param interface - the interface on which to capture packets.
 * @param num - number of packets to capture, -1 for infinity.
 * @param callback - callback function pointer to pass to the thread.
 */
/****************************************************************************/
string NetworkDriver::StartCapture(unsigned int interface, int num, pcap_handler callback)
{
  stringstream strstr;
  if ((int)interface < 0 || (int)interface > (int)(interface_names.size() - 1))
  {
    strstr << "Interface number: " << interface << " is invalid.";
    throw NetworkDriverException(strstr.str());
  }

  pthread_mutex_lock(&activeCaptInterfaces_lock);

  if (activeCaptInterfaces.at(interface) == 1)
  {
    pthread_mutex_unlock(&activeCaptInterfaces_lock);
    throw NetworkDriverException("Attempting to start 2 capture threads on 1 interface.");
  }

  if (activeCaptInterfaces.at(interface) == 0)
  {
    pcap_c_descriptors.at(interface) = pcap_open_live(interface_names.at(interface), BUFSIZ, 1, -1, errbuf);
    if (pcap_c_descriptors.at(interface) == NULL)
    {
      pthread_mutex_unlock(&activeCaptInterfaces_lock);
      throw NetworkDriverException(errbuf);
    }

    if (pcap_setdirection(pcap_c_descriptors.at(interface), PCAP_D_IN) == -1)
    {
      pthread_mutex_unlock(&activeCaptInterfaces_lock);
      throw NetworkDriverException("Capture direction not supported.");
    }

    activeCaptInterfaces.at(interface) = 1;
  }

  pthread_mutex_unlock(&activeCaptInterfaces_lock);

  thread_args *strc = new thread_args;
  strc->handlerFunction = callback;
  strc->numberOfPackets = num;
  strc->descriptor = pcap_c_descriptors.at(interface);
  strc->userOptions = intToCharArry(interface);
  strc->ndriver = this;
  strc->interfaceNum = interface;

  //thread communication pipe
  int pipes[2];
  if (pipe(pipes))
    throw NetworkDriverException("Failed to create pipe.");

  strc->listenPipeFd = pipes[0];
  captureThreadFds.at(interface) = pipes[1];
  fcntl(captureThreadFds.at(interface), F_SETFL, O_NONBLOCK);

  isCapture[interface] = true;

  if (pthread_create(&thread_ids.at(interface), NULL, captureThreadWrapper, (void *)strc))
    throw NetworkDriverException("Failed to start new capture thread.");

  strstr << "Started capture on interface: " << interface_names.at(interface) << endl;
  return strstr.str();
}

/****************************************************************************/
//! Sends the passed packet to the passed interface.
/*! Sends the passed packet to the passed interface.
 *
 * @param interface - the interface on which the frame is to be sent.
 * @param packet - packet to send.
 * @param size - packet size.
 * @return 0 on success and -1 on failure.
 */
/****************************************************************************/
int NetworkDriver::SendPacket(unsigned int interface, u_char *packet, unsigned int size)
{
  pthread_mutex_lock(&interface_names_lock);
  pthread_mutex_lock(&activeSendInterfaces_lock);
  //check if interface number is in range
  if (interface > interface_names.size())
  {
    cerr << "Interface number is invalid: " << interface << "\n";
    pthread_mutex_unlock(&interface_names_lock);
    pthread_mutex_unlock(&activeSendInterfaces_lock);
    throw NetworkDriverException("Interface number is invalid.");
  }//check if descriptor exists
  else if (activeSendInterfaces.at(interface) == 0)
  {
    pcap_s_descriptors.at(interface) = pcap_open_live(interface_names.at(interface), BUFSIZ, 1, -1, errbuf);
    if (pcap_s_descriptors.at(interface) == NULL)
    {
      pthread_mutex_unlock(&interface_names_lock);
      pthread_mutex_unlock(&activeSendInterfaces_lock);
      throw NetworkDriverException(errbuf);
    }

    activeSendInterfaces.at(interface) = 1;
  }

  pthread_mutex_unlock(&interface_names_lock);
  pthread_mutex_unlock(&activeSendInterfaces_lock);

  int res = 1;
  pthread_mutex_lock(pcap_send_desc_lock.at(interface));
  res = pcap_sendpacket(pcap_s_descriptors.at(interface), packet, size);
  sent_vect_stats.at(interface) += 1;
  pthread_mutex_unlock(pcap_send_desc_lock.at(interface));

  if (res)
    cout << pcap_geterr(pcap_s_descriptors.at(interface)) << endl;
  //	else
  //		cout << "driver sent to: "<< interface<< endl;

  return res;
}

/****************************************************************************/
//! Displays the network interfaces of this computer.
/*! Displays the network interfaces of this computer.
 */
/****************************************************************************/
void NetworkDriver::printInterfaces()
{
  pthread_mutex_lock(&interface_names_lock);
  for (unsigned int i = 0; i < interface_names.size(); ++i)
  {
    cout << "(" << i << ")" << interface_names.at(i) << "\n";
  }
  pthread_mutex_unlock(&interface_names_lock);
}

/****************************************************************************/
//! Returns the interface name at position i.
/*! Returns the interface name at position i.
 *
 * @param i - the interface position.
 * @return interface name at position i.
 */
/****************************************************************************/
string NetworkDriver::getInterfaceNameAt(unsigned int i)
{
  pthread_mutex_lock(&interface_names_lock);
  if (i <= interface_names.size() && i > 0)
  {
    pthread_mutex_unlock(&interface_names_lock);
    return interface_names.at(i);
  }
  else
  {
    pthread_mutex_unlock(&interface_names_lock);
    return "";
  }
}

/****************************************************************************/
//! Thread wrapper function.
/*! Thread wrapper function, performs the packet capture using pcap_loop.
 * The thread sleeps on select() until either a new frame arrives, or a write
 * signal is sent from the destructor through pipe().
 *
 * @param ptr - points to a struct with thread arguments.
 * @return 0.
 */
/****************************************************************************/
void *NetworkDriver::captureThreadWrapper(void *ptr)
{
  thread_args *args = (thread_args *)ptr;

  //80% CPU load when idle
  //	pcap_loop(args->descriptor,args->numberOfPackets,args->handlerFunction,args->userOptions);

  //0% CPU load when idle
  fd_set writefds;
  FD_ZERO(&writefds);

  int capturefd = pcap_get_selectable_fd(args->descriptor);
  fd_set capturefds;
  FD_ZERO(&capturefds);

  //TODO num of packets != -1 doesnt free all memory
  int packet_num = (args->numberOfPackets == -1) ? 1 : args->numberOfPackets;
  while (args->ndriver->isCapture[args->interfaceNum] && packet_num > 0)//possible race condition
  {
    FD_SET(capturefd,&capturefds);
    FD_SET(args->listenPipeFd,&capturefds);
    select(args->listenPipeFd + 1, &capturefds, &writefds, NULL, NULL);
    pcap_dispatch(args->descriptor, packet_num, args->handlerFunction, args->userOptions);

    if (args->numberOfPackets != -1)
      packet_num--;
  }
  close(args->listenPipeFd);

  //	cout << "Stopped capture on interface: "<< args->userOptions<<endl;
  delete[] args->userOptions;
  delete args;
  pthread_exit(0);
  return 0;
}

/****************************************************************************/
//! Starts a new queue processing thread.
/*! Starts a new queue processing thread, to serve the frame queue.
 */
/****************************************************************************/
void NetworkDriver::startQueueProcessingThread()
{
  threadPtr = 0;
  if (pthread_create(&threadPtr, NULL, processQueue, (void *)this))
    throw NetworkDriverException("Failed to start new NetworkDriver thread.");
}

/****************************************************************************/
//! Processing thread function, to serve the frames queue.
/*! Processing thread function, to serve the frames queue. This thread is
 * terminated by adding a NULL frame to queue.
 *
 * @param _driver - a Network Driver object.
 */
/****************************************************************************/
void *NetworkDriver::processQueue(void *_driver)
{
  NetworkDriver *driver = (NetworkDriver *)_driver;
  Frame *frm;
  bool processFlag = true;

  while (processFlag)
  {
    driver->ndriverQ.wait_and_pop(frm);

    if (frm->frame != NULL)
    {
      try
      {
        driver->SendPacket(frm->interface, frm->frame, frm->length);
        //				driver->startSendingThread(frm);
      }
      catch (NetworkDriverException& e)
      {
        cerr << "NetworkDriver exception: " << e.getMessage() << endl;
      }

      delete[] frm->frame;
      delete frm;
    }
    else
    {
      processFlag = false;
    }
  }
  pthread_exit(0);
  return 0;
}

/****************************************************************************/
//! Looks for network interfaces, and adds their details to the class members.
/*! Looks for network interfaces, and adds their details to the class members.
 *
 * @return the number of interfaces found.
 */
/****************************************************************************/
int NetworkDriver::lookForInterfaces()
{
  int num = 0;
  pcap_if *ptr = pcap_interfaces;
  while (ptr != NULL && ptr->next != NULL)
  {
    interface_names.push_back(ptr->name);
    activeSendInterfaces.push_back(0);
    activeCaptInterfaces.push_back(0);
    sent_vect_stats.push_back(0);
    pcap_c_descriptors.push_back(NULL);
    pcap_s_descriptors.push_back(NULL);
    pcap_send_desc_lock.push_back(new pthread_mutex_t());
    captureThreadFds.push_back(0);
    thread_ids.push_back(0);
    isCapture.push_back(false);
    ptr = ptr->next;
    num++;
  }
  return num;
}

/****************************************************************************/
//! Starts a new packet sending thread.
/*! Starts a new packet sending.
 */
/****************************************************************************/
void NetworkDriver::startSendingThread(Frame *_frm)
{
  send_thread_args *strc = new send_thread_args;
  strc->ndriver = this;
  strc->frm = _frm;

  int pts = 0;
  pthread_t sendThreadPtr = 0;
  if ((pts = pthread_create(&sendThreadPtr, NULL, sendWrapper, (void *)strc)))
  {
    cout << "Pthread failed with error code: " << pts << endl;
    throw NetworkDriverException("Failed to start new NetworkDriver Sending thread.");
  }
}

/****************************************************************************/
//! Sending thread function, to send the frames.
/*! Sending thread function, to send the frames.
 *
 * @param ptr - a Network Driver sending thread struct.
 */
/****************************************************************************/
void *NetworkDriver::sendWrapper(void *ptr)
{
  send_thread_args *args = (send_thread_args *)ptr;

  args->ndriver->SendPacket(args->frm->interface, args->frm->frame, args->frm->length);

  delete[] args->frm->frame;
  delete args->frm;
  delete args;
  pthread_exit(0);
  return 0;
}

u_char *NetworkDriver::intToCharArry(unsigned int integer)
{
  u_char *result = new u_char[sizeof(unsigned int)];

  for (unsigned int i = 0; i < sizeof(unsigned int); ++i)
  {
    result[sizeof(unsigned int) - 1 - i] = (integer >> (8 * i));
  }
  return result;
}

unsigned int NetworkDriver::charArryToInt(u_char *bytes)
{
  unsigned int arrToInt = 0;

  for (unsigned int i = 0; i < sizeof(unsigned int); ++i)
  {
    arrToInt = (arrToInt << 8) | bytes[i];
  }
  return arrToInt;
}
