/****************************************************************************
 *
 * File Name: NetworkDriver.h
 *
 * General Description: Implements the Network Driver class.
 *
 * Created By: Alex, Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

#ifndef NETWORKDRIVER_H_
#define NETWORKDRIVER_H_

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file NetworkDriver.h
 *\brief Implements the Network Driver class.
 *
 * This module interacts directly with the network interfaces. Its role is to
 * capture frames on each interface, send them to the LLRE, and to forward
 * frames from the Dispatcher to the network.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/

#include <pcap.h> /* if this gives you an error try pcap/pcap.h */
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include "NetworkDriverException.h"
#include "../Includes/Datastructures.h"
#include "../Includes/concurrent_queue.h"

class NetworkDriver;

/**
 * \brief Holds the arguments of the pcap_loop function that are passed to the thread.
 */
typedef struct thread_args_t
{
  pcap_t * descriptor; //pcap descriptor
  int numberOfPackets; //number of packets to capture
  pcap_handler handlerFunction; //pcap handler function pointer
  u_char * userOptions; //arguments passed to the capture thread
  NetworkDriver *ndriver; //network driver pointer
  int listenPipeFd; //file descriptor to wake up select() in capture thread
  unsigned int interfaceNum; //network interface number
} thread_args;

typedef struct send_thread_args_t
{
  NetworkDriver *ndriver; //network driver pointer
  Frame *frm; //Frame pointer
} send_thread_args;

/**
 * \brief Network Driver class which encapsulates libpcap.
 */
class NetworkDriver
{
public:
  NetworkDriver();
  virtual ~NetworkDriver();
  std::string StartCapture(unsigned int interface, int num, pcap_handler callback);
  std::string StopCapture(unsigned int interface);
  std::string StopAllActiveCapture();
  int SendPacket(unsigned int interface, u_char *packet, unsigned int size);
  void printInterfaces();
  std::string getInterfaceNameAt(unsigned int i);
  void pushFrameToQueue(Frame *frame);
  static u_char *intToCharArry(unsigned int integer);
  static unsigned int charArryToInt(u_char *bytes);

private:
  int lookForInterfaces();
  static void *captureThreadWrapper(void *ptr);
  void startQueueProcessingThread();
  static void *processQueue(void *_driver);
  void startSendingThread(Frame *_frm);
  static void *sendWrapper(void *ptr);

  pthread_t threadPtr; //queue processing thread id
  std::vector<pcap_t*> pcap_c_descriptors; //pcap capture descriptors
  std::vector<pcap_t*> pcap_s_descriptors; //pcap send descriptors
  std::vector<pthread_mutex_t *> pcap_send_desc_lock; //send descriptors mutexes

  char errbuf[PCAP_ERRBUF_SIZE]; //pcap error buffer
  pcap_if_t* pcap_interfaces; //pcap interfaces

  std::vector<pthread_t> thread_ids; //capture thread ids

  std::vector<char *> interface_names; //network interface names
  pthread_mutex_t interface_names_lock; //interface names lock
  std::vector<int> activeSendInterfaces; //0 not active, 1 descriptor exists
  pthread_mutex_t activeSendInterfaces_lock; //active Send Interfaces lock
  std::vector<int> activeCaptInterfaces; //0 not active, 1 descriptor exists
  pthread_mutex_t activeCaptInterfaces_lock; //active Capture Interfaces lock

  std::vector<bool> isCapture; //flags to continue capture
  std::vector<int> captureThreadFds; //capture threads file descriptors piped to destructor
  concurrent_queue<Frame *> ndriverQ; //thread safe frame queue

  pcap_stat ps;
  std::vector<u_int> sent_vect_stats;
  u_int received;
  u_int dropped;
  u_int if_dropped;
  u_int sent;
};

#endif /* NETWORKDRIVER_H_ */
