/****************************************************************************
 *
 * File Name: Classifier_Handler.cpp
 *
 * General Description: Implements the Classifier XML Handler class.
 *
 * Created By: Alex

 , Creation Date: 11/09/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Classifier_Handler.cpp
 *\brief Implements the Classifier_Handler class.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "Classifier_Handler.h"
#include "XML_Handler.h"

using namespace std;
using namespace rapidxml;

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param XML_Handler - the XML_Handler pointer.
 */
/****************************************************************************/
Classifier_Handler::Classifier_Handler(XML_Handler *_xmlHandler)
{
  xmlHandler = _xmlHandler;
}

/****************************************************************************/
//! Destructor.
/*! Destructor.
 */
/****************************************************************************/
Classifier_Handler::~Classifier_Handler()
{
}

/****************************************************************************/
//! Handles the converted Classifier LFB XML.
/*! Handles the converted Classifier LFB XML
 *
 *	<?xml version="1.0" encoding="UTF-8" standalone="no"?>
 <LFBs>
 <LFB_Classifier>
 <synopsis> LFB used to classify frames. It includes all frame definitions </synopsis>
 <frameDefs>
 <frame>
 <frameTypeId>0x05</frameTypeId>
 <name>IPv4 over ethernet</name>
 <synopsis>IPv4 over ethernet</synopsis>
 <fields>
 <field type="regular"	 	length="48" name="Dst Mac"/>
 <field type="regular" 		length="48" name="Src Mac"/>
 <field type="identifier" 	length="16" name="Ether Type">
 <identifier_type type="strict">0x0800</identifier_type>
 </field>
 <field type="identifier" 	length="4" 	name="Version">
 <identifier_type type="strict">0x004</identifier_type>
 </field>
 <field type="regular"	 	length="4" 	name="IP Header Length"/>
 <field type="QoS" 			length="6" 	name="DiffServ"/>
 <field type="forwarding" 	length="2" 	name="ECN"/>
 <field type="regular"	 	length="8" 	name="Total IP Length"/>
 <field type="regular" 		length="32" name="Identification"/>
 <field type="regular" 		length="3" 	name="Flags"/>
 <field type="regular" 		length="13" name="Fragment Offset"/>
 <field type="identifier" 	length="8" 	name="TTL">
 <identifier_type type="arithmetic_bigger_than">0x0</identifier_type>
 </field>
 <field type="regular" 		length="8" 	name="Protocol"/>
 <field type="checksum" 		length="16" name="IP Header Checksum">
 <checksum_fields>
 <frameField name="Version"/>
 <frameField name="Header Length"/>
 <frameField name="DiffServ"/>
 <frameField name="ECN"/>
 <frameField name="Length"/>
 <frameField name="Identification"/>
 <frameField name="Flags"/>
 <frameField name="Fragment Offset"/>
 <frameField name="TTL"/>
 <frameField name="Protocol"/>
 <frameField name="SIP"/>
 <frameField name="DIP"/>
 </checsum_fields>
 </field>
 <field type="regular" 		length="32" name="SIP"/>
 <field type="regular" 		length="32" name="DIP"/>
 </fields>
 </frame>
 </frameDefs>
 </LFB_Classifier>
 </LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Classifier_Handler::parse(vector<char> &XML)
{
  xml_document<> doc;
  try
  {
    doc.parse<parse_declaration_node | parse_no_data_nodes> (&XML[0]);
  }
  catch (parse_error e)
  {

    string error("Parsing error in Classifier: ");
    error.append(e.what());
    return error.append("\n");;
  }

  xml_node<>* cur_node = doc.first_node();
  if (!boost::iequals(cur_node->name(),"LFBs"))
  {
    cur_node = cur_node->next_sibling();
    if (cur_node == NULL || !boost::iequals(cur_node->name(),"LFBs") )
      return "Invalid XML file: Not LFB type XML.\n";
  }

  cur_node = cur_node->first_node();
  if (!boost::iequals(cur_node->name(),"LFB_Classifier"))
    return "Invalid XML file: Not a Classifier type LFB.\n";

  cur_node = cur_node->first_node();
  if (!boost::iequals(cur_node->name(),"synopsis"))
    return "Invalid XML file: No 'synopsis' field in XML.\n";
  string synopsis = cur_node->value();

  xml_node<>* frame_node = cur_node->next_sibling();
  if (!boost::iequals(frame_node->name(),"frameDefs"))
    return "Invalid XML file: No 'frameDefs' field.\n";

  frame_node = frame_node->first_node("frame");

  string frame_type_id;
  string name;
  string frame_synopsis;

//  xmlHandler->llre->clearInitialTable();
//  xmlHandler->llre->clearIDtoFWDTable();
//  xmlHandler->llre->clearActionTable();
//
//  xmlHandler->idKeysMap.clear();
//  xmlHandler->idKeysWithValuesMap.clear();
//  xmlHandler->fwdKeysMap.clear();
//  xmlHandler->classifierKeys.clear();
//  xmlHandler->isClassifierLoaded = false;

  while (frame_node != NULL)
  {
    cur_node = frame_node;

    if (cur_node->first_node("frameTypeId") == NULL)
      return "Invalid XML file: No 'frameTypeId' field.\n";
    frame_type_id = cur_node->first_node("frameTypeId")->value();

    if (cur_node->first_node("name") == NULL)
      return "Invalid XML file: No 'name' field.\n";
    name = cur_node->first_node("name")->value();

    if (cur_node->first_node("synopsis") == NULL)
      return "Invalid XML file: No 'synopsis' field.\n";
    frame_synopsis = cur_node->first_node("synopsis")->value();

    cur_node = cur_node->first_node("fields");
    if (cur_node == NULL)
      return "Invalid XML file: No 'fields' field.\n";

    xml_node<>* fields_node = cur_node->first_node();

    unsigned int length = 0;

    Position currentPosition;

    string lenStr;
    while (fields_node != NULL)
    {
      if (!field_node_has_all_attr(fields_node))
        return "Invalid XML file: 'field' doesn't have all the required attributes.\n";

      lenStr = fields_node->first_attribute("length")->value();
      length = StringToNumber<int> (lenStr);

      if (length > 0)
      {
        //	|0|1|2|3|4|5|6|7|
        currentPosition.setLastBitOffset(currentPosition.getFirstBitOffset() + ((length % 8) ? (length % 8 - 1) : 7));
        currentPosition.setLength(((length % 8) ? (length / 8 + 1) : (length / 8)));

        try
        {
          handleClassifierFrameField(frame_type_id, fields_node, currentPosition);
        }
        catch (GFEexception &e)
        {
          return e.getMessage();
        }

        currentPosition.setOffset(currentPosition.getOffset() + (currentPosition.getFirstBitOffset() + length) / 8);
        currentPosition.setFirstBitOffset((currentPosition.getLastBitOffset() + 1) % 8);
      }

      fields_node = fields_node->next_sibling();
    }
    frame_node = frame_node->next_sibling();
  }

  xmlHandler->isClassifierLoaded = true;

  return "Classifier XML loaded successfully.\n";
}

/****************************************************************************/
//! Handles the converted Classifier LFB XML.
/*! Handles the converted Classifier LFB XML
 *
 *      <?xml version="1.0" encoding="UTF-8" standalone="no"?>
 <LFBs>
 <LFB_Classifier>
 <synopsis> LFB used to classify frames. It includes all frame definitions </synopsis>
 <frameDefs>
 <frame>
 <frameTypeId>0x05</frameTypeId>
 <name>IPv4 over ethernet</name>
 <synopsis>IPv4 over ethernet</synopsis>
 <fields>
 <field type="regular"          length="48" name="Dst Mac"/>
 <field type="regular"          length="48" name="Src Mac"/>
 <field type="identifier"       length="16" name="Ether Type">
 <identifier_type type="strict">0x0800</identifier_type>
 </field>
 <field type="identifier"       length="4"      name="Version">
 <identifier_type type="strict">0x004</identifier_type>
 </field>
 <field type="regular"          length="4"      name="IP Header Length"/>
 <field type="QoS"                      length="6"      name="DiffServ"/>
 <field type="forwarding"       length="2"      name="ECN"/>
 <field type="regular"          length="8"      name="Total IP Length"/>
 <field type="regular"          length="32" name="Identification"/>
 <field type="regular"          length="3"      name="Flags"/>
 <field type="regular"          length="13" name="Fragment Offset"/>
 <field type="identifier"       length="8"      name="TTL">
 <identifier_type type="arithmetic_bigger_than">0x0</identifier_type>
 </field>
 <field type="regular"          length="8"      name="Protocol"/>
 <field type="checksum"                 length="16" name="IP Header Checksum">
 <checksum_fields>
 <frameField name="Version"/>
 <frameField name="Header Length"/>
 <frameField name="DiffServ"/>
 <frameField name="ECN"/>
 <frameField name="Length"/>
 <frameField name="Identification"/>
 <frameField name="Flags"/>
 <frameField name="Fragment Offset"/>
 <frameField name="TTL"/>
 <frameField name="Protocol"/>
 <frameField name="SIP"/>
 <frameField name="DIP"/>
 </checsum_fields>
 </field>
 <field type="regular"          length="32" name="SIP"/>
 <field type="regular"          length="32" name="DIP"/>
 </fields>
 </frame>
 </frameDefs>
 </LFB_Classifier>
 </LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Classifier_Handler::parseFromNode(rapidxml::xml_node<>* cur_node)
{
  if (!boost::iequals(cur_node->name(),"LFB_Classifier"))
    return "Invalid XML file: Not a Classifier type LFB.\n";

  cur_node = cur_node->first_node();
  if (!boost::iequals(cur_node->name(),"synopsis"))
    return "Invalid XML file: No 'synopsis' field in XML.\n";
  string synopsis = cur_node->value();

  xml_node<>* frame_node = cur_node->next_sibling();
  if (!boost::iequals(frame_node->name(),"frameDefs"))
    return "Invalid XML file: No 'frameDefs' field.\n";

  frame_node = frame_node->first_node("frame");

  string frame_type_id;
  string name;
  string frame_synopsis;

//  xmlHandler->llre->clearInitialTable();
//  xmlHandler->llre->clearIDtoFWDTable();
//  xmlHandler->llre->clearActionTable();
//
//  xmlHandler->idKeysMap.clear();
//  xmlHandler->idKeysWithValuesMap.clear();
//  xmlHandler->fwdKeysMap.clear();
//  xmlHandler->classifierKeys.clear();
//  xmlHandler->isClassifierLoaded = false;

  while (frame_node != NULL)
  {
    cur_node = frame_node;

    if (cur_node->first_node("frameTypeId") == NULL)
      return "Invalid XML file: No 'frameTypeId' field.\n";
    frame_type_id = cur_node->first_node("frameTypeId")->value();

    if (cur_node->first_node("name") == NULL)
      return "Invalid XML file: No 'name' field.\n";
    name = cur_node->first_node("name")->value();

    if (cur_node->first_node("synopsis") == NULL)
      return "Invalid XML file: No 'synopsis' field.\n";
    frame_synopsis = cur_node->first_node("synopsis")->value();

    cur_node = cur_node->first_node("fields");
    if (cur_node == NULL)
      return "Invalid XML file: No 'fields' field.\n";

    xml_node<>* fields_node = cur_node->first_node();

    unsigned int length = 0;

    Position currentPosition;

    string lenStr;
    while (fields_node != NULL)
    {
      if (!field_node_has_all_attr(fields_node))
        return "Invalid XML file: 'field' doesn't have all the required attributes.\n";

      lenStr = fields_node->first_attribute("length")->value();
      length = StringToNumber<int> (lenStr);

      if (length > 0)
      {
        //      |0|1|2|3|4|5|6|7|
        currentPosition.setLastBitOffset(currentPosition.getFirstBitOffset() + ((length % 8) ? (length % 8 - 1) : 7));
        currentPosition.setLength(((length % 8) ? (length / 8 + 1) : (length / 8)));

        try
        {
          handleClassifierFrameField(frame_type_id, fields_node, currentPosition);
        }
        catch (GFEexception &e)
        {
          return e.getMessage();
        }

        currentPosition.setOffset(currentPosition.getOffset() + (currentPosition.getFirstBitOffset() + length) / 8);
        currentPosition.setFirstBitOffset((currentPosition.getLastBitOffset() + 1) % 8);
      }

      fields_node = fields_node->next_sibling();
    }
    frame_node = frame_node->next_sibling();
  }

  xmlHandler->isClassifierLoaded = true;

  return "Classifier XML loaded successfully.\n";
}


/****************************************************************************/
//! Checks that the field node has all the required attributes.
/*!	Checks that the field node has all the following attributes:
 *
 *	<field type="forwarding-key" length="48" name="Src Mac(arp)"/>
 *
 * @param fields_node - reference to the field node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Classifier_Handler::field_node_has_all_attr(xml_node<>* fields_node)
{
  return (fields_node->first_attribute("type") != NULL) && (fields_node->first_attribute("length") != NULL)
      && (fields_node->first_attribute("name") != NULL);
}

/****************************************************************************/
//! Handles the identifier-key fields.
/*! Handles the identifier-key fields.
 *
 * @param frame_type_id - type ID of the frame.
 * @param fields_node - xml node containing the field.
 * @param length - field length
 * @param offset - field offset
 */
/****************************************************************************/
void Classifier_Handler::handleClassifierFrameField(string frame_type_id, xml_node<>* fields_node, Position &pos)
{
  //store all fields in classifier
  string name = processClassifierName(frame_type_id, fields_node, pos);
  //TODO check empty name?

  string type(fields_node->first_attribute("type")->value());

  std::vector<string> strs;
  boost::split(strs, type, boost::is_any_of(","));

  for (unsigned int i = 0; i < strs.size(); ++i)
  {
    if( boost::iequals(strs.at(i),IDENTIFIER) )
    { //store id-keys
      handleIdentifierKeyType(frame_type_id, fields_node, pos, name);
    }
    else if ( boost::iequals(strs.at(i),FORWARDING) )
    {
      handleForwardingKeyType(frame_type_id, fields_node, pos, name);
    }
    else if ( boost::iequals(strs.at(i),CHECKSUM) )
    {
      handleCheckSumKeyType(frame_type_id, fields_node, pos, name);
    }
    else if ( boost::iequals(strs.at(i),QOS) )
    {
      //TODO implement
    }
    else if ( boost::iequals(strs[i],METERING) )
    {
      //TODO implement
    }
  }
}

/****************************************************************************/
//! Handles the 'name' field in 'field.
/*! Handles the 'name' field in 'field.
 *
 * @param frame_type_id - type ID of the frame.
 * @param fields_node - xml node containing the field.
 * @param length - field length
 * @param offset - field offset
 */
/****************************************************************************/
string Classifier_Handler::processClassifierName(string frame_type_id, xml_node<>* fields_node, Position &pos)
{
  string name(fields_node->first_attribute("name")->value());
  FWD_KEY fwdK(pos, name);

  //check if the FWD_KEY already exists
  umultimap_mapOFfwdKeysMaps::iterator iter = xmlHandler->classifierKeys.find(frame_type_id);
  if (iter != xmlHandler->classifierKeys.end())
  { //add the FWD_KEY to the hash_map at frame_type_id
    iter->second.insert(pair<string, FWD_KEY> (name, fwdK));
  }
  else
  { //create a new hash_map with the new FWD_KEY at frame_type_id
    umultimap_fwdKeys fwdKeysM;
    fwdKeysM.insert(pair<string, FWD_KEY> (name, fwdK));
    xmlHandler->classifierKeys.insert(pair<string, umultimap_fwdKeys> (frame_type_id, fwdKeysM));
  }
  return name;
}

/****************************************************************************/
//! Handles the 'type' field in 'field.
/*! Handles the 'type' field in 'field.
 *
 * @param frame_type_id - type ID of the frame.
 * @param fields_node - xml node containing the field.
 * @param length - field length
 * @param offset - field offset
 * @param name - field name
 */
/****************************************************************************/
void Classifier_Handler::handleIdentifierKeyType(string frame_type_id, xml_node<>* fields_node, Position &pos,
                                                 string name)
{
  if (fields_node->first_node("identifier_type") != NULL
      && fields_node->first_node("identifier_type")->first_attribute("type") != NULL)
  {
    string type(fields_node->first_node("identifier_type")->first_attribute("type")->value());
    if (boost::iequals(type,"strict"))
    {
      string valueStr(fields_node->first_node("identifier_type")->value());
      BYTE *value = XML_Handler::stringToBYTEs(valueStr, pos);
      if (value != NULL)
      {
        ID_KEY idkey(pos, name, value);
        xmlHandler->idKeysMap.insert(pair<string, ID_KEY> (frame_type_id, idkey));

        ID_KEY idkey_with_val(pos, name, value);
        xmlHandler->idKeysWithValuesMap.insert(pair<string, ID_KEY> (frame_type_id, idkey_with_val));
        delete[] value;
      }
      else if (boost::iequals(type,"arithmetic_bigger_than"))
      {
        //TODO implement
      }
    }
  }
}

/****************************************************************************/
//! Handles the forwarder-key fields.
/*! Handles the forwarder-key fields.
 *
 * @param frame_type_id - type ID of the frame.
 * @param fields_node - xml node containing the field.
 * @param length - field length
 * @param offset - field offset
 * @param name - field name
 */
/****************************************************************************/
void Classifier_Handler::handleForwardingKeyType(string frame_type_id, xml_node<>* fields_node, Position &pos,
                                                 string name)
{
  string valueStr(fields_node->value());
  stringstream strstr;
  if (valueStr.length() == 0)
  {
    FWD_KEY fwdkey(pos, name);
    xmlHandler->fwdKeysMap.insert(pair<string, FWD_KEY> (frame_type_id, fwdkey));
  }
  else
  {
    BYTE *value = XML_Handler::stringToBYTEs(valueStr, pos);
    if (value != NULL)
    {
      FWD_KEY fwdkey_with_val(pos, name, value);
      xmlHandler->fwdKeysMap.insert(pair<string, FWD_KEY> (frame_type_id, fwdkey_with_val));
      delete[] value;
    }
    else
    {
      strstr << "The value: '" << valueStr << "' is invalid.\n";
      throw GFEexception(strstr.str());
    }
  }
}

/****************************************************************************/
//! Handles the CheckSumKey type fields.
/*! Handles the CheckSumKey type fields.
 *
 * @param frame_type_id - type ID of the frame.
 * @param fields_node - xml node containing the field.
 * @param length - field length
 * @param offset - field offset
 * @param name - field name
 */
/****************************************************************************/
void Classifier_Handler::handleCheckSumKeyType(string frame_type_id, xml_node<>* fields_node, Position &pos,
                                               string name)
{
  if (fields_node->first_node("checksum_fields") != NULL && fields_node->first_node("checksum_fields")->first_node()
      != NULL)
  { //checksum fields exist

    xml_node<>* curr_node = fields_node->first_node("checksum_fields")->first_node();
    Frame_field cheksumField(0, pos.getOffset(), pos.getLength(), name, 0, pos.getFirstBitOffset(),
                             pos.getLastBitOffset());

    //check if the frame_type_id already exists
    umultimap_mapOFchksumFieldsNamesMaps::iterator iter = xmlHandler->fieldsToChecksumsMaps.find(frame_type_id);

    if (iter != xmlHandler->fieldsToChecksumsMaps.end())
    { //this frame_type_id already exists
      do
      {
        if (curr_node->first_attribute("name") != NULL)
          iter->second.insert(pair<string, Frame_field> (curr_node->first_attribute("name")->value(), cheksumField));
        curr_node = curr_node->next_sibling();
      } while (curr_node != NULL);

    }
    else
    { //create a new hash_map with the new fields at frame_type_id
      umultimap_chksumFieldsNames chksumFieldsM;
      do
      {
        if (curr_node->first_attribute("name") != NULL)
          chksumFieldsM.insert(pair<string, Frame_field> (curr_node->first_attribute("name")->value(), cheksumField));
        curr_node = curr_node->next_sibling();
      } while (curr_node != NULL);
      xmlHandler->fieldsToChecksumsMaps.insert(pair<string, umultimap_chksumFieldsNames> (frame_type_id, chksumFieldsM));
    }
  }
}

/****************************************************************************/
//! Returns true if the ID_KEY is in the idKeysMap.
/*! Returns true if the ID_KEY is in the idKeysMap.
 *
 * @param key - id_key.
 * @return - true if the ID_KEY is in the idKeysMap, false otherwise.
 */
/****************************************************************************/
bool Classifier_Handler::slowFindIdKey(ID_KEY &key)
{
  bool found = false;
  for (umultimap_idKeys::iterator it = xmlHandler->idKeysMap.begin(); it != xmlHandler->idKeysMap.end() && !found; ++it)
  {
    if (it->second == key)
      found = true;
  }
  return found;
}
