/****************************************************************************
 *
 * File Name: Forwarder_Handler.cpp
 *
 * General Description: Implements the Forwarder_Handler class.
 *
 * Created By: Alex

 , Creation Date: 11/09/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Forwarder_Handler.cpp
 *\brief Implements the Forwarder_Handler class.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "boost/lexical_cast.hpp"
#include <boost/algorithm/string.hpp>
#include "Forwarder_Handler.h"
#include "XML_Handler.h"

using namespace std;
using namespace rapidxml;

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param XML_Handler - the XML_Handler pointer.
 */
/****************************************************************************/
Forwarder_Handler::Forwarder_Handler(XML_Handler *_xmlHandler)
{
  xmlHandler = _xmlHandler;
}

/****************************************************************************/
//! Destructor.
/*! Destructor.
 */
/****************************************************************************/
Forwarder_Handler::~Forwarder_Handler()
{
}

/****************************************************************************/
//! Handles the converted Forwarder LFB XML.
/*! Handles the converted Forwarder LFB XML, all fields in the following
 *	example must be present:
 *
 *	<?xml version="1.0" encoding="UTF-8" standalone="no"?>
 *	<LFBs>
 *		<LFB_Forwarder>
 *			<synopsis> LFB holding the forwarding database </synopsis>
 *			<entries>
 *				 <fdb_entry active="yes" id="0x1" frameTypeId ="0x01" tableID="0x01" static="yes" noMatchEntry="no">
 *					<incoming_frame>
 *						<key>
 *							<ports>
 *								<portId>2</portId>
 *							</ports>
 *							<frameFields>
 *								<frameField name="Target IP Address">0x0a0000fe</frameField>
 *							</frameFields>
 *						</key>
 *					</incoming_frame>
 *					<outgoing_frame>
 *						<isMulticast value="no"/>
 *						<results>
 *							<result id="0x1" dstPort="2" QoS="0" mappingId="1">
 *								<frameFields>
 *									<frameField name="Src Address(arp)">0x0007e9a51ade</frameField>
 *									<frameField name="Sender MAC Address">0x0007e9a51ade</frameField>
 *									<frameField name="Sender IP Address">0x0a0000fe</frameField>
 *								</frameFields>
 *							</result>
 *						</results>
 *					</outgoing_frame>
 *				</fdb_entry>
 *			</entries>
 *		</LFB_Forwarder>
 *	</LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Forwarder_Handler::parse(vector<char> &XML)
{
  if (xmlHandler->isClassifierLoaded && xmlHandler->isMapperLoaded)
  {
    xml_document<> doc;

    try
    {
      doc.parse<parse_declaration_node | parse_no_data_nodes> (&XML[0]);
    }
    catch (parse_error e)
    {

      string error("Parsing error in Forwarder: ");
      error.append(e.what());
      return error.append("\n");
    }

    xml_node<>* cur_node = doc.first_node();
    if (!boost::iequals(cur_node->name(),"LFBs"))
    {
      cur_node = cur_node->next_sibling();
      if (cur_node == NULL || !boost::iequals(cur_node->name(),"LFBs") )
        return "Invalid XML file: Not LFB type XML.\n";
    }

    cur_node = cur_node->first_node();
    if (!boost::iequals(cur_node->name(),"LFB_Forwarder"))
      return "Invalid XML file: Not a Forwarder type LFB.\n";

    cur_node = cur_node->first_node();
    if (!boost::iequals(cur_node->name(),"synopsis"))
      return "Invalid XML file: No 'synopsis' field in XML.\n";
    string synopsis = cur_node->value();

    cur_node = cur_node->next_sibling();
    if (!boost::iequals(cur_node->name(),"entries"))
      return "Invalid XML file: No 'Classifier entries' field.\n";

    cur_node = cur_node->first_node("fdb_entry");
    xml_node<>* fdb_entry_node = cur_node;

//    CLEAR_TABLES();

//    xmlHandler->isForwarderLoaded = false;
    string fdb_entry_result;

    while (fdb_entry_node != NULL)
    {
      fdb_entry_result = Handle_LFB_Forwarder_fdb_entry_node(fdb_entry_node,
                                                             cur_node);
      if (fdb_entry_result.compare("OK") != 0)
        return fdb_entry_result;

      fdb_entry_node = fdb_entry_node->next_sibling();
    }

    xmlHandler->isForwarderLoaded = true;
    return "Forwarder XML loaded successfully.\n";
  }
  else
    return "Classifier and/or Mapper not loaded yet.\n";
}


/****************************************************************************/
//! Handles the converted Forwarder LFB XML.
/*! Handles the converted Forwarder LFB XML, all fields in the following
 *      example must be present:
 *
 *      <?xml version="1.0" encoding="UTF-8" standalone="no"?>
 *      <LFBs>
 *              <LFB_Forwarder>
 *                      <synopsis> LFB holding the forwarding database </synopsis>
 *                      <entries>
 *                               <fdb_entry active="yes" id="0x1" frameTypeId ="0x01" tableID="0x01" static="yes" noMatchEntry="no">
 *                                      <incoming_frame>
 *                                              <key>
 *                                                      <ports>
 *                                                              <portId>2</portId>
 *                                                      </ports>
 *                                                      <frameFields>
 *                                                              <frameField name="Target IP Address">0x0a0000fe</frameField>
 *                                                      </frameFields>
 *                                              </key>
 *                                      </incoming_frame>
 *                                      <outgoing_frame>
 *                                              <isMulticast value="no"/>
 *                                              <results>
 *                                                      <result id="0x1" dstPort="2" QoS="0" mappingId="1">
 *                                                              <frameFields>
 *                                                                      <frameField name="Src Address(arp)">0x0007e9a51ade</frameField>
 *                                                                      <frameField name="Sender MAC Address">0x0007e9a51ade</frameField>
 *                                                                      <frameField name="Sender IP Address">0x0a0000fe</frameField>
 *                                                              </frameFields>
 *                                                      </result>
 *                                              </results>
 *                                      </outgoing_frame>
 *                              </fdb_entry>
 *                      </entries>
 *              </LFB_Forwarder>
 *      </LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Forwarder_Handler::parseFromNode(rapidxml::xml_node<>* cur_node)
{
  if (xmlHandler->isClassifierLoaded && xmlHandler->isMapperLoaded)
  {
    if (!boost::iequals(cur_node->name(),"LFB_Forwarder"))
      return "Invalid XML file: Not a Forwarder type LFB.\n";

    cur_node = cur_node->first_node();
    if (!boost::iequals(cur_node->name(),"synopsis"))
      return "Invalid XML file: No 'synopsis' field in XML.\n";
    string synopsis = cur_node->value();

    cur_node = cur_node->next_sibling();
    if (!boost::iequals(cur_node->name(),"entries"))
      return "Invalid XML file: No 'Classifier entries' field.\n";

    cur_node = cur_node->first_node("fdb_entry");
    xml_node<>* fdb_entry_node = cur_node;

//    CLEAR_TABLES();
//
//    xmlHandler->isForwarderLoaded = false;
    string fdb_entry_result;

    while (fdb_entry_node != NULL)
    {
      fdb_entry_result = Handle_LFB_Forwarder_fdb_entry_node(fdb_entry_node,
                                                             cur_node);
      if (fdb_entry_result.compare("OK") != 0)
        return fdb_entry_result;

      fdb_entry_node = fdb_entry_node->next_sibling();
    }

    xmlHandler->isForwarderLoaded = true;
    return "Forwarder XML loaded successfully.\n";
  }
  else
    return "Classifier and/or Mapper not loaded yet.\n";
}


/****************************************************************************/
//! Handles the fdb_entry field.
/*!	Handles the fdb_entry field:
 *
 *				 <fdb_entry active="yes" id="0x1" frameTypeId ="0x01" tableID="0x01" static="yes" noMatchEntry="no">
 *					<incoming_frame>
 *						<key>
 *							<ports>
 *								<portId>2</portId>
 *							</ports>
 *							<frameFields>
 *								<frameField name="Target IP Address">0x0a0000fe</frameField>
 *							</frameFields>
 *						</key>
 *					</incoming_frame>
 *					<outgoing_frame>
 *						<isMulticast value="no"/>
 *						<results>
 *							<result id="0x1" dstPort="2" QoS="0" mappingId="1">
 *								<frameFields>
 *									<frameField name="Src Address(arp)">0x0007e9a51ade</frameField>
 *									<frameField name="Sender MAC Address">0x0007e9a51ade</frameField>
 *									<frameField name="Sender IP Address">0x0a0000fe</frameField>
 *								</frameFields>
 *							</result>
 *						</results>
 *					</outgoing_frame>
 *				</fdb_entry>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Forwarder_Handler::Handle_LFB_Forwarder_fdb_entry_node(
                                                              xml_node<>* fdb_entry_node,
                                                              xml_node<>* cur_node)
{
  if (!fdb_entry_has_all_attr(fdb_entry_node))
  {
    CLEAR_TABLES();
    return "Invalid XML file: 'fdb_entry' doesn't have all the required attributes.\n";
  }
  string activeStr = fdb_entry_node->first_attribute("active")->value();
  bool isActive = boost::iequals(activeStr,"yes") ? true : false;
  string entryIdStr = fdb_entry_node->first_attribute("id")->value();
  unsigned int entryId = StringToHexNumber<unsigned int> (entryIdStr);
  string frameTypeId = fdb_entry_node->first_attribute("frameTypeId")->value();

  vector<ID_KEY *> idKeys;
  vector<ID_KEY *> idKeysWithValues;
  vector<FWD_KEY *> fwdKeys;
  vector<fwdKeyMaskpair *> fwdKeysAndMasks;
  vector<string> frameFields;

  xmlHandler->getID_KEYs(frameTypeId, idKeys, xmlHandler->idKeysMap);
  xmlHandler->getID_KEYs(frameTypeId, idKeysWithValues, xmlHandler->idKeysWithValuesMap);
  xmlHandler->getFWD_KEYs(frameTypeId, fwdKeys);

  if (idKeys.size() > 0)
  {
    if (fdb_entry_node->first_node("incoming_frame") == NULL
        || fdb_entry_node->first_node("incoming_frame")->first_node("key")
            == NULL)
    {
      CLEAR_TABLES();
      return "Invalid XML file: No 'incoming_frame or key' field.\n";
    }
    cur_node = fdb_entry_node->first_node("incoming_frame")->first_node("key");

    if (cur_node->first_node("ports") == NULL
        || cur_node->first_node("ports")->first_node("portId") == NULL)
    {
      CLEAR_TABLES();
      return "Invalid XML file: No 'ports or portId' field.\n";
    }
    xml_node<>* port_node = cur_node->first_node("ports")->first_node("portId");

    unsigned int interface = 0;
    FWD_KEY *fwdkv = NULL;
    FWD_KEY *fwdkeyMask = NULL;

    while (port_node != NULL)//handle incoming port
    {
      interface = StringToNumber<unsigned int> (port_node->value());

      if (cur_node->first_node("frameFields") == NULL
          || cur_node->first_node("frameFields")->first_node("frameField")
              == NULL)
      {
        CLEAR_TABLES();
        return "Invalid XML file: No 'frameFields or frameField' field.\n";
      }
      xml_node<>* cur_frame_node =
          cur_node->first_node("frameFields")->first_node("frameField");

      try{
      //fwdKeysAndMasks
      handleIncomingFrameField(cur_frame_node, fwdKeys, interface, &fwdkv,
                               &fwdkeyMask, fwdKeysAndMasks);
      }catch (GFEexception& e) {
        CLEAR_TABLES();
        return e.getMessage();
      }

      if (fdb_entry_node->first_node("outgoing_frame") == NULL
          || fdb_entry_node->first_node("outgoing_frame")->first_node(
                                                                      "isMulticast")
              == NULL)
      {
        CLEAR_TABLES();
        return "Invalid XML file: No 'outgoing_frame or isMulticast' field.\n";
      }
      cur_frame_node
          = fdb_entry_node->first_node("outgoing_frame")->first_node(
                                                                     "isMulticast");
      string isMulticast(cur_frame_node->first_attribute()->value());

      if (cur_frame_node->next_sibling("results") == NULL
          || cur_frame_node->next_sibling("results")->first_node("result")
              == NULL)
      {
        CLEAR_TABLES();
        return "Invalid XML file: No 'results or result' field.\n";
      }
      cur_frame_node
          = cur_frame_node->next_sibling("results")->first_node("result");

      //handles multiple results
      if (fwdkv != NULL && fwdkeyMask != NULL)
      {
        try
        {
          //fwdKeysAndMasks
          handleForwarderResults(cur_frame_node, fwdKeysAndMasks, frameTypeId, interface);
        }
        catch (GFEexception& e)
        {
          CLEAR_TABLES();
          return e.getMessage();
        }
      }

      port_node = port_node->next_sibling();

      if (fwdkv != NULL)
        delete fwdkv;
      if (fwdkeyMask != NULL)
        delete fwdkeyMask;

    }

    xmlHandler->llre->add_Initial_Lookup_Table_Entry(interface, isActive,
                                                     entryId, idKeys);

    xmlHandler->llre->add_ID_to_FWD_Entry(interface, idKeysWithValues,
                                          fwdKeysAndMasks);
  }
  return "OK";
}

/****************************************************************************/
//! Checks that the fdb_entry field has all the required attributes.
/*!	Checks that the fdb_entry field has all the following attributes:
 *
 *	 <fdb_entry active="yes" id="0x1" frameTypeId ="0x01" tableID="0x01" static="yes" noMatchEntry="no">
 *
 * @param fdb_entry_node - reference to the fdb_entry node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Forwarder_Handler::fdb_entry_has_all_attr(xml_node<>* fdb_entry_node)
{
  return (fdb_entry_node->first_attribute("active") != NULL)
      && (fdb_entry_node->first_attribute("id") != NULL)
      && (fdb_entry_node->first_attribute("frameTypeId") != NULL)
      && (fdb_entry_node->first_attribute("tableID") != NULL)
      && (fdb_entry_node->first_attribute("static") != NULL)
      && (fdb_entry_node->first_attribute("noMatchEntry") != NULL);
}

/****************************************************************************/
//! Handles the incoming_frame fields.
/*! Handles the incoming_frame fields.
 *
 * @param cur_node - xml node containing the current field.
 * @param fwdKeys - vector with forwarding keys.
 * @param interface - NIC number.
 * @param fwdkv - pointer to pointer of the fwd key.
 * @param fwdkeyMask - pointer to pointer of the fwd key mask.
 */
/****************************************************************************/
void Forwarder_Handler::handleIncomingFrameField(
                                                 xml_node<>* cur_node,
                                                 vector<FWD_KEY *> &fwdKeys,
                                                 unsigned int interface,
                                                 FWD_KEY **fwdkv,
                                                 FWD_KEY **fwdkeyMask,
                                                 vector<fwdKeyMaskpair *> &fwdKeysAndMasks)
{
  while (cur_node != NULL && cur_node->first_attribute("name") != NULL)
  {
    string frameFieldName(cur_node->first_attribute("name")->value());
    string frameFieldVal(cur_node->value());

    int pos = findNameInFwdKeysVect(fwdKeys, frameFieldName);
    if (pos >= 0)
    {
      BYTE *value = XML_Handler::stringToBYTEs(frameFieldVal,
                                               fwdKeys.at(pos)->getPosition());
      if (value != NULL)
      {
        *fwdkv = new FWD_KEY(*fwdKeys.at(pos));
        (*fwdkv)->setValue(value);
        delete[] value;

        size_t delim_pos = frameFieldVal.find('/');
        if (delim_pos != string::npos)
        {
          string maskStr = frameFieldVal.substr(delim_pos + 1,
                                                frameFieldVal.size());
          BYTE *MACmask =
              XML_Handler::stringToBYTEs(maskStr,
                                         fwdKeys.at(pos)->getPosition());
          *fwdkeyMask = new FWD_KEY(fwdKeys.at(pos)->getPosition(),
                                    fwdKeys.at(pos)->getName(), MACmask);
          delete[] MACmask;
        }
        else
        {
          BYTE *MACmask = new BYTE[fwdKeys.at(pos)->getLength()];
          for (unsigned int i = 0; i < fwdKeys.at(pos)->getLength(); ++i)
          {
            MACmask[i] = 0xff;
          }

          *fwdkeyMask = new FWD_KEY(fwdKeys.at(pos)->getPosition(),
                                    fwdKeys.at(pos)->getName(), MACmask);
          delete[] MACmask;
        }

        //adjust mask to actual value size
        (*fwdkeyMask)->adjustMaskToSize();

        fwdKeysAndMasks.push_back(new fwdKeyMaskpair(**fwdkv, **fwdkeyMask));
      }
    }
    else
    {
      stringstream strstr;
      strstr << "The forwarding key field '"<< frameFieldName <<"' not found in Classifier.\n";
      throw GFEexception(strstr.str());
    }
    cur_node = cur_node->next_sibling();
  }
}

/****************************************************************************/
//! Handles the results fields.
/*! Handles the results fields.
 *
 * @param cur_node - xml node containing the current field.
 * @param fwdkv - pointer of the fwd key.
 * @param fwdkeyMask - pointer of the fwd key mask.
 * @param frame_type_id - type ID of the frame.
 */
/****************************************************************************/
void Forwarder_Handler::handleForwarderResults(xml_node<>* cur_node, vector<
    fwdKeyMaskpair *> &fwdKeysAndMasks, string frameTypeId, unsigned int interface)
{
  stringstream strstr;
  bool correct = false;
  vector<Action*> actionVect;
  while (cur_node != NULL)
  {
    //TODO check multiple ports -> same result
    xml_node<>* destPortNode = cur_node->first_node("destPortId");

    string dstIntStr;
    string mappingID;
    string QoS;
    double qosNum = 0;
    if (destPortNode != NULL)//legacy xml handler
    {
      dstIntStr = destPortNode->value();
      correct = true;
    }
    else
    {
      correct = result_entry_has_all_attr(cur_node);
      if (correct)
      {
        destPortNode = cur_node;
        dstIntStr = destPortNode->first_attribute("dstPort")->value();
        QoS = destPortNode->first_attribute("QoS")->value();

        try
        {
          qosNum = boost::lexical_cast<double>(QoS);
        }
        catch (const std::exception&)
        {
          qosNum = 0;
        }

        mappingID = destPortNode->first_attribute("mappingId")->value();
      }
      else
      {
        stringstream strstr;
        strstr << "The forwarding result of frameTypeId"<< frameTypeId
              <<" doesn't have all the required attributes.\n";
        throw GFEexception(strstr.str());
      }
    }

    if (correct)
    {
      string frameID(frameTypeId);
      if (mappingID.compare("0") != 0)
      {//mapping exists
        vector<Action*> mapActionsVect;
        xmlHandler->getMapperActions(mappingID, mapActionsVect);
        if (mapActionsVect.size() > 0)
          actionVect.insert(actionVect.end(), mapActionsVect.begin(),
                            mapActionsVect.end());

        umap_mappingIdToOutFrameID::const_iterator iterator =
            xmlHandler->mappingIdToOutFrameMap.find(mappingID);
        if (iterator == xmlHandler->mappingIdToOutFrameMap.end())
        {
          strstr << "Mapping ID: " << mappingID << " not found.\n";
          throw GFEexception(strstr.str());
        }
        else
        {
          frameID = iterator->second;
        }
      }

      xml_node<>* frameFieldNode = cur_node->first_node("frameFields");
      handle_forwarder_result_frameFields(frameFieldNode, actionVect, frameID);

      xml_node<>* arithmeticsNode = cur_node->first_node("arithmetics");
      handle_forwarder_result_arithmetics(arithmeticsNode, actionVect, frameID);

      unsigned int destInterface = StringToNumber<unsigned int> (dstIntStr);
      actionVect.push_back(new Action_Forward(destInterface, qosNum));
      xmlHandler->llre->add_Action_Table_Entry(interface,fwdKeysAndMasks, actionVect);
      actionVect.clear();
    }

    cur_node = cur_node->next_sibling("result");
  }
}

/****************************************************************************/
//! Handles the forwarder result field.
/*! Handles the forwarder result field.
 *
 * @param frameFieldNode - xml node containing the frameField.
 * @param actionVect - reference to the actions vector.
 * @param frame_type_id - type ID of the frame.
 */
/****************************************************************************/
void Forwarder_Handler::handle_forwarder_result_frameFields(
                                                            xml_node<> *frameFieldNode,
                                                            vector<Action*> &actionVect,
                                                            string frameTypeId)
{
  stringstream strstr;
  if (frameFieldNode != NULL)
  {
    frameFieldNode = frameFieldNode->first_node("frameField");
    while (frameFieldNode != NULL)
    {
      if (frameFieldNode->first_attribute("name") != NULL)
      {
        string fieldNameStr = frameFieldNode->first_attribute("name")->value();
        string fieldValueStr(frameFieldNode->value());
        BYTE *fieldValue;
        //				vector<FWD_KEY *> resFwdKeys;
        //				getResFWD_KEYs(fieldNameStr,resFwdKeys);
        //				if(resFwdKeys.size()>0)
        //				{
        //					fieldValue = stringToBYTEs(fieldValueStr,resFwdKeys.at(0)->getPosition());
        //
        //					addAction(actionVect,frameTypeId,fieldNameStr,resFwdKeys.at(0),NULL,fieldValue,NULL,Assign,"");
        //					delete [] fieldValue;
        //				}
        //				else
        //				{
        FWD_KEY * fk = xmlHandler->getClassifierFields(frameTypeId,
                                                       fieldNameStr);
        if (fk != NULL)
        {
          fieldValue = XML_Handler::stringToBYTEs(fieldValueStr,
                                                  fk->getPosition());
          xmlHandler->addAction(actionVect, frameTypeId, fieldNameStr, fk,
                                NULL, fieldValue, NULL, Assign, "");
          delete[] fieldValue;
        }
        else
        {
          strstr << "Forwarding key: " << fieldNameStr << " not found.\n";
          throw GFEexception(strstr.str());
        }
        //				}
      }

      frameFieldNode = frameFieldNode->next_sibling();
    }
  }
  else
  {
    //result can have no additional nodes
    //		strstr << "'frameFields' field not found in 'results'.\n";
    //		throw GFEexception(strstr.str());
  }
}

/****************************************************************************/
//! Handles the forwarder result arithmetics field.
/*! Handles the forwarder result arithmetics field.
 *
 * @param frameFieldNode - xml node containing the frameField.
 * @param actionVect - reference to the actions vector.
 * @param frame_type_id - type ID of the frame.
 */
/****************************************************************************/
void Forwarder_Handler::handle_forwarder_result_arithmetics(
                                                            xml_node<> *arithmeticsNode,
                                                            vector<Action*> &actionVect,
                                                            string frameTypeId)
{
  stringstream strstr;
  if (arithmeticsNode != NULL)
  {
    arithmeticsNode = arithmeticsNode->first_node("frameField");
    while (arithmeticsNode != NULL)
    {
      if (arithmeticsNode->first_attribute("name") != NULL)
      {
        string fieldNameStr = arithmeticsNode->first_attribute("name")->value();

        FWD_KEY * fk = xmlHandler->getClassifierFields(frameTypeId,
                                                       fieldNameStr);
        if (fk != NULL)
        {
          xml_node<> *operationNode = arithmeticsNode->first_node("operation");
          if (operationNode != NULL)
          {
            if (operationNode->first_attribute("operator") != NULL)
            {
              string operatorStr =
                  operationNode->first_attribute("operator")->value();
              xml_node<> *operandNode = operationNode->first_node("operand");
              if (operandNode != NULL)
              {
                string fieldValueStr(operandNode->value());
                BYTE
                     *operandFieldValue =
                         XML_Handler::stringToBYTEs(fieldValueStr,
                                                    fk->getPosition());

                xml_node<> *maskNode = operationNode->first_node("mask");
                if (maskNode != NULL)
                {
                  string maskFieldValueStr(maskNode->value());
                  BYTE *maskFieldValue =
                      XML_Handler::stringToBYTEs(maskFieldValueStr,
                                                 fk->getPosition());
                  xmlHandler->addAction(actionVect, frameTypeId, fieldNameStr,
                                        fk, NULL, operandFieldValue,
                                        maskFieldValue, Arithmetic, operatorStr);
                  delete[] maskFieldValue;
                }
                else
                {
                  xmlHandler->addAction(actionVect, frameTypeId, fieldNameStr,
                                        fk, NULL, operandFieldValue, NULL,
                                        Arithmetic, operatorStr);
                }
                delete[] operandFieldValue;
              }
              else
              {
                strstr << "Arithmetic: no operand found\n";
                throw GFEexception(strstr.str());
              }
            }
            else
            {
              strstr << "Arithmetic: no operator attribute found\n";
              throw GFEexception(strstr.str());
            }
          }
          else
          {
            strstr << "Arithmetic: no operation found\n";
            throw GFEexception(strstr.str());
          }
        }
        else
        {
          strstr << "Arithmetic: frameField '" << fieldNameStr
              << "' not found\n";
          throw GFEexception(strstr.str());
        }
      }
      arithmeticsNode = arithmeticsNode->next_sibling();
    }
  }
}

/****************************************************************************/
//! Checks that the result_node has all the required attributes.
/*!	Checks that the result_node has all the following attributes:
 *
 *	 <result id="0x1" dstPort="8" QoS="0" mappingId="0">
 *
 * @param result_node - reference to the result node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Forwarder_Handler::result_entry_has_all_attr(xml_node<>* result_node)
{
  return (result_node->first_attribute("id") != NULL)
      && (result_node->first_attribute("dstPort") != NULL)
      && (result_node->first_attribute("QoS") != NULL)
      && (result_node->first_attribute("mappingId") != NULL);
}

/****************************************************************************/
//! Returns the index of the FWD_KEY with the name 'name' in the passed vector.
/*! Returns the index of the FWD_KEY with the name 'name' in the passed vector.
 *
 * @param fwdKeysVect - keys vector to search in.
 * @param name - name of the key to search for.
 * @return the index of the FWD_KEY if found, -1 otherwise.
 */
/****************************************************************************/
int Forwarder_Handler::findNameInFwdKeysVect(vector<FWD_KEY*> &fwdKeysVect,
                                             string &name)
{
  bool found = false;
  unsigned int i = 0;
  for (; i < fwdKeysVect.size() && !found; ++i)
  {
    if (fwdKeysVect.at(i)->getName().compare(name) == 0)
      found = true;
  }
  if (!found)
    i = -1;

  return i - 1;
}
