/****************************************************************************
 *
 * File Name: Mapper_Handler.cpp
 *
 * General Description: Implements the Mapper_Handler class.
 *
 * Created By: Alex , Creation Date: 11/09/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file Mapper_Handler.cpp
 *\brief Implements the Mapper_Handler class.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include "Mapper_Handler.h"
#include "XML_Handler.h"

using namespace std;
using namespace rapidxml;

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param XML_Handler - the XML_Handler pointer.
 */
/****************************************************************************/
Mapper_Handler::Mapper_Handler(XML_Handler *_xmlHandler)
{
  xmlHandler = _xmlHandler;
}

/****************************************************************************/
//! Destructor.
/*! Destructor.
 */
/****************************************************************************/
Mapper_Handler::~Mapper_Handler()
{
}

/****************************************************************************/
//! Handles the converted Mapper LFB XML.
/*! Handles the converted Mapper LFB XML
 *
 *
 <LFBs>
 <LFB_mapper>
 <synopsis>LFB mapper for ARP example</synopsis>
 <mappings>
 <frame_mapping mappingId="1" inFrame="0x01" outFrame="0x02">
 <mapped_field>
 <incoming_field name="Src Address(arp)" discard="no"/>
 <outgoing_field name="Dest Address" fixed="no"/>
 </mapped_field>
 <mapped_field>
 <outgoing_field name="Src Address(arp)" fixed="no"/>
 </mapped_field>
 </frame_mapping>
 </mappings>
 </LFB_mapper>
 </LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Mapper_Handler::parse(vector<char> &XML)
{
  if (xmlHandler->isClassifierLoaded)
  {
    xml_document<> doc;
    try
    {
      doc.parse<parse_declaration_node | parse_no_data_nodes> (&XML[0]);
    }
    catch (parse_error e)
    {
      string error("Parsing error in Mapper: ");
      error.append(e.what());
      return error.append("\n");;
    }

    xml_node<>* cur_node = doc.first_node();
    if (!boost::iequals(cur_node->name(),"LFBs"))
    {
      cur_node = cur_node->next_sibling();
      if (cur_node == NULL || !boost::iequals(cur_node->name(),"LFBs") )
        return "Invalid XML file: Not LFB type XML.\n";
    }

    cur_node = cur_node->first_node();
    //
    if (!boost::iequals(cur_node->name(),"LFB_Mapper"))
      return "Invalid XML file: Not Mapper type LFB.\n";

    cur_node = cur_node->first_node();
    if (!boost::iequals(cur_node->name(),"synopsis"))
      return "Invalid XML file: No 'synopsis' field in XML.\n";
    string synopsis = cur_node->value();

    xml_node<>* mapped_field_node = cur_node->next_sibling();
    if (!boost::iequals(mapped_field_node->name(),"mappings"))
      return "Invalid XML file: No 'mappings' field in XML.\n";

    mapped_field_node = mapped_field_node->first_node("frame_mapping");

    string mappingId;
    string inFrame;
    string outFrame;
    string name;
    string incoming_field_name;
    string outgoing_field_name;
    string outgoing_field;
    string no("no");
    string type;
    xml_node<>* fields_node;
    FWD_KEY *fk_incoming;
    FWD_KEY *fk_outgoing;
    bool discard;
    bool fixed;
//    xmlHandler->clearMapperActions();
//    xmlHandler->isMapperLoaded = false;
    vector<Action*> mapper_actions;
    while (mapped_field_node != NULL)
    {
      if (!frame_mapping_entry_has_all_attr(mapped_field_node))
      {
        xmlHandler->clearMapperActions();
        return "Invalid XML file: 'frame_mapping' doesn't have all the required attributes.\n";
      }

      cur_node = mapped_field_node;
      mappingId = cur_node->first_attribute("mappingId")->value();
      inFrame = cur_node->first_attribute("inFrame")->value();
      outFrame = cur_node->first_attribute("outFrame")->value();
      cur_node = cur_node->first_node("mapped_field");

      //add mapping ID to out frame mapping
      xmlHandler->mappingIdToOutFrameMap[mappingId] = outFrame;
      xmlHandler->addPosition.clearPosition();
      xmlHandler->cutPosition.clearPosition();

      while (cur_node != NULL)
      {
        incoming_field_name = "";
        outgoing_field_name = "";
        discard = false;
        fixed = false;

        if (cur_node->first_attribute("type") != NULL)
        {
          type = cur_node->first_attribute("type")->value();
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: no mapped_field type found.\n";
        }

        fields_node = cur_node->first_node("incoming_field");
        if (fields_node != NULL)
        {
          if (!incoming_field_entry_has_all_attr(fields_node))
          {
            xmlHandler->clearMapperActions();
            return "Invalid XML file: 'incoming_field' doesn't have all the required attributes.\n";
          }

          if (fields_node->first_attribute("discard")->value() != NULL
              && !boost::iequals(fields_node->first_attribute("discard")->value(),no) )
          {
            discard = true;
          }

          incoming_field_name = fields_node->first_attribute("name")->value();
        }

        fields_node = cur_node->first_node("outgoing_field");
        if (fields_node != NULL)
        {
          if (!outgoing_field_entry_has_all_attr(fields_node))
          {
            xmlHandler->clearMapperActions();
            return "Invalid XML file: 'outgoing_field' doesn't have all the required attributes.\n";
          }

          if (fields_node->first_attribute("fixed")->value() != NULL
              && !boost::iequals(fields_node->first_attribute("fixed")->value(),no) )
          {
            fixed = true;
            outgoing_field = fields_node->value();
          }

          outgoing_field_name = fields_node->first_attribute("name")->value();
        }

        if (incoming_field_name.size() > 0 && outgoing_field_name.size() > 0)
        {
          if (!boost::iequals(incoming_field_name,outgoing_field_name))
          {
            fk_incoming = xmlHandler->getClassifierFields(inFrame, incoming_field_name);
            fk_outgoing = xmlHandler->getClassifierFields(outFrame, outgoing_field_name);
            if (fk_incoming != NULL && fk_outgoing != NULL)
            {
              xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, fk_incoming, NULL, NULL,
                                    Copy, "");
            }
          }
        }
        else if (incoming_field_name.size() == 0 && outgoing_field_name.size() > 0)
        {
          fk_outgoing = xmlHandler->getClassifierFields(outFrame, outgoing_field_name);
          if (fk_outgoing != NULL)
          {
            xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, NULL, NULL, NULL, Add, "");
            if (fixed)
            {
              BYTE *fieldValue = XML_Handler::stringToBYTEs(outgoing_field, fk_outgoing->getPosition());
              xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, NULL, fieldValue, NULL,
                                    Assign, "");
              delete[] fieldValue;

            }
          }
        }
        else if (incoming_field_name.size() > 0 && outgoing_field_name.size() == 0)
        {
          fk_incoming = xmlHandler->getClassifierFields(inFrame, incoming_field_name);

          if (fk_incoming != NULL)
          {
            xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, NULL, fk_incoming, NULL, NULL, Cut, "");
          }
        }

        cur_node = cur_node->next_sibling();
      }

      if (xmlHandler->addPosition.getLength() > 0)
      {
        if (xmlHandler->addPosition.getFirstBitOffset() == 0)
        {
          mapper_actions.insert(mapper_actions.begin(), new Action_Add(xmlHandler->addPosition.getOffset(),
                                                                       xmlHandler->addPosition.getLength()));
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: Total added fields must be in byte resolution.\n";
        }
      }

      if (xmlHandler->cutPosition.getLength() > 0)
      {
        if (xmlHandler->cutPosition.getFirstBitOffset() == 0)
        {
          mapper_actions.insert(mapper_actions.begin(), new Action_Cut(xmlHandler->cutPosition.getOffset(),
                                                                       xmlHandler->cutPosition.getLength()));
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: Total cut fields must be in byte resolution.\n";
        }
      }

      xmlHandler->mapperActions.insert(pair<string, vector<Action*> > (mappingId, mapper_actions));
      mapper_actions.clear();
      mapped_field_node = mapped_field_node->next_sibling();
    }

    xmlHandler->isMapperLoaded = true;
    return "Mapper XML loaded successfully.\n";
  }
  else
    return "Classifier not loaded yet.\n";
}

/****************************************************************************/
//! Handles the converted Mapper LFB XML.
/*! Handles the converted Mapper LFB XML
 *
 *
 <LFBs>
 <LFB_mapper>
 <synopsis>LFB mapper for ARP example</synopsis>
 <mappings>
 <frame_mapping mappingId="1" inFrame="0x01" outFrame="0x02">
 <mapped_field>
 <incoming_field name="Src Address(arp)" discard="no"/>
 <outgoing_field name="Dest Address" fixed="no"/>
 </mapped_field>
 <mapped_field>
 <outgoing_field name="Src Address(arp)" fixed="no"/>
 </mapped_field>
 </frame_mapping>
 </mappings>
 </LFB_mapper>
 </LFBs>
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string Mapper_Handler::parseFromNode(xml_node<>* cur_node)
{
  if (xmlHandler->isClassifierLoaded)
  {
    if (!boost::iequals(cur_node->name(),"LFB_Mapper"))
      return "Invalid XML file: Not Mapper type LFB.\n";

    cur_node = cur_node->first_node();
    if (!boost::iequals(cur_node->name(),"synopsis"))
      return "Invalid XML file: No 'synopsis' field in XML.\n";
    string synopsis = cur_node->value();

    xml_node<>* mapped_field_node = cur_node->next_sibling();
    if (!boost::iequals(mapped_field_node->name(),"mappings"))
      return "Invalid XML file: No 'mappings' field in XML.\n";

    mapped_field_node = mapped_field_node->first_node("frame_mapping");

    string mappingId;
    string inFrame;
    string outFrame;
    string name;
    string incoming_field_name;
    string outgoing_field_name;
    string outgoing_field;
    string no("no");
    string type;
    xml_node<>* fields_node;
    FWD_KEY *fk_incoming;
    FWD_KEY *fk_outgoing;
    bool discard;
    bool fixed;
//    xmlHandler->clearMapperActions();
//    xmlHandler->isMapperLoaded = false;
    vector<Action*> mapper_actions;
    while (mapped_field_node != NULL)
    {
      if (!frame_mapping_entry_has_all_attr(mapped_field_node))
      {
        xmlHandler->clearMapperActions();
        return "Invalid XML file: 'frame_mapping' doesn't have all the required attributes.\n";
      }

      cur_node = mapped_field_node;
      mappingId = cur_node->first_attribute("mappingId")->value();
      inFrame = cur_node->first_attribute("inFrame")->value();
      outFrame = cur_node->first_attribute("outFrame")->value();
      cur_node = cur_node->first_node("mapped_field");

      //add mapping ID to out frame mapping
      xmlHandler->mappingIdToOutFrameMap[mappingId] = outFrame;
      xmlHandler->addPosition.clearPosition();
      xmlHandler->cutPosition.clearPosition();

      while (cur_node != NULL)
      {
        incoming_field_name = "";
        outgoing_field_name = "";
        discard = false;
        fixed = false;

        if (cur_node->first_attribute("type") != NULL)
        {
          type = cur_node->first_attribute("type")->value();
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: no mapped_field type found.\n";
        }

        fields_node = cur_node->first_node("incoming_field");
        if (fields_node != NULL)
        {
          if (!incoming_field_entry_has_all_attr(fields_node))
          {
            xmlHandler->clearMapperActions();
            return "Invalid XML file: 'incoming_field' doesn't have all the required attributes.\n";
          }

          if (fields_node->first_attribute("discard")->value() != NULL
              && !boost::iequals(fields_node->first_attribute("discard")->value(),no) )
          {
            discard = true;
          }

          incoming_field_name = fields_node->first_attribute("name")->value();
        }

        fields_node = cur_node->first_node("outgoing_field");
        if (fields_node != NULL)
        {
          if (!outgoing_field_entry_has_all_attr(fields_node))
          {
            xmlHandler->clearMapperActions();
            return "Invalid XML file: 'outgoing_field' doesn't have all the required attributes.\n";
          }

          if (fields_node->first_attribute("fixed")->value() != NULL
              && !boost::iequals(fields_node->first_attribute("fixed")->value(),no) )
          {
            fixed = true;
            outgoing_field = fields_node->value();
          }

          outgoing_field_name = fields_node->first_attribute("name")->value();
        }

        if (incoming_field_name.size() > 0 && outgoing_field_name.size() > 0)
        {
          if (!boost::iequals(incoming_field_name,outgoing_field_name))
          {
            fk_incoming = xmlHandler->getClassifierFields(inFrame, incoming_field_name);
            fk_outgoing = xmlHandler->getClassifierFields(outFrame, outgoing_field_name);
            if (fk_incoming != NULL && fk_outgoing != NULL)
            {
              xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, fk_incoming, NULL, NULL,
                                    Copy, "");
            }
          }
        }
        else if (incoming_field_name.size() == 0 && outgoing_field_name.size() > 0)
        {
          fk_outgoing = xmlHandler->getClassifierFields(outFrame, outgoing_field_name);
          if (fk_outgoing != NULL)
          {
            xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, NULL, NULL, NULL, Add, "");
            if (fixed)
            {
              BYTE *fieldValue = XML_Handler::stringToBYTEs(outgoing_field, fk_outgoing->getPosition());
              xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, fk_outgoing, NULL, fieldValue, NULL,
                                    Assign, "");
              delete[] fieldValue;

            }
          }
        }
        else if (incoming_field_name.size() > 0 && outgoing_field_name.size() == 0)
        {
          fk_incoming = xmlHandler->getClassifierFields(inFrame, incoming_field_name);

          if (fk_incoming != NULL)
          {
            xmlHandler->addAction(mapper_actions, outFrame, outgoing_field_name, NULL, fk_incoming, NULL, NULL, Cut, "");
          }
        }

        cur_node = cur_node->next_sibling();
      }

      if (xmlHandler->addPosition.getLength() > 0)
      {
        if (xmlHandler->addPosition.getFirstBitOffset() == 0)
        {
          mapper_actions.insert(mapper_actions.begin(), new Action_Add(xmlHandler->addPosition.getOffset(),
                                                                       xmlHandler->addPosition.getLength()));
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: Total added fields must be in byte resolution.\n";
        }
      }

      if (xmlHandler->cutPosition.getLength() > 0)
      {
        if (xmlHandler->cutPosition.getFirstBitOffset() == 0)
        {
          mapper_actions.insert(mapper_actions.begin(), new Action_Cut(xmlHandler->cutPosition.getOffset(),
                                                                       xmlHandler->cutPosition.getLength()));
        }
        else
        {
          xmlHandler->clearMapperActions();
          return "Invalid XML file: Total cut fields must be in byte resolution.\n";
        }
      }

      xmlHandler->mapperActions.insert(pair<string, vector<Action*> > (mappingId, mapper_actions));
      mapper_actions.clear();
      mapped_field_node = mapped_field_node->next_sibling();
    }

    xmlHandler->isMapperLoaded = true;
    return "Mapper XML loaded successfully.\n";
  }
  else
    return "Classifier not loaded yet.\n";
}

/****************************************************************************/
//! Checks that the frame_mapping has all the required attributes.
/*!	Checks that the frame_mapping has all the following attributes:
 *
 *	 <frame_mapping mappingId="1" inFrame="0x01" outFrame="0x02">
 *
 * @param frame_mapping - reference to the frame_mapping node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Mapper_Handler::frame_mapping_entry_has_all_attr(xml_node<>* frame_mapping)
{
  return (frame_mapping->first_attribute("mappingId") != NULL) && (frame_mapping->first_attribute("inFrame") != NULL)
      && (frame_mapping->first_attribute("outFrame") != NULL);
}

/****************************************************************************/
//! Checks that the incoming_field has all the required attributes.
/*!	Checks that the incoming_field has all the following attributes:
 *
 *	 <incoming_field name="Src Address(arp)" discard="no"/>
 *
 * @param fields_node - reference to the incoming_field node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Mapper_Handler::incoming_field_entry_has_all_attr(xml_node<>* fields_node)
{
  return (fields_node->first_attribute("name") != NULL) && (fields_node->first_attribute("discard") != NULL);
}

/****************************************************************************/
//! Checks that the outgoing_field has all the required attributes.
/*!	Checks that the outgoing_field has all the following attributes:
 *
 *	 <outgoing_field name="Dest Address" fixed="no"/>
 *
 * @param fields_node - reference to the outgoing_field node.
 * @return true if the node has all the attributes, false otherwise.
 */
/****************************************************************************/
bool Mapper_Handler::outgoing_field_entry_has_all_attr(xml_node<>* fields_node)
{
  return (fields_node->first_attribute("name") != NULL) && (fields_node->first_attribute("fixed") != NULL);
}
