/****************************************************************************
 *
 * File Name: XML_Handler.cpp
 *
 * General Description: Implements the XML Handler class.
 *
 * Created By: Alex, Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/

/****************************************************************************
 * File description
 ****************************************************************************/

/*! \file XML_Handler.cpp
 *\brief Implements the XML Handler class.
 *
 * This module is responsible for converting received LFBs, in a form of
 * XMLs, into the data base entries which it then stores in the lookup tables
 * of the LLRE.
 *
 *\author Alex
 */

/****************************************************************************
 * Include header files
 ****************************************************************************/
#include <fstream>
#include <exception>
#include "boost/lexical_cast.hpp"
#include "XML_Handler.h"
#include "Classifier_Handler.h"
#include "Mapper_Handler.h"
#include "Forwarder_Handler.h"
#include "../GFCP_FE_Client/src/GFCP_WS_Client.h"

using namespace std;
using namespace rapidxml;

/****************************************************************************
 * functions implementations
 ****************************************************************************/

//TODO save loaded xml

/****************************************************************************/
//! Default Constructor.
/*! Default Constructor.
 */
/****************************************************************************/
XML_Handler::XML_Handler()
{
  llre = NULL;
  isClassifierLoaded = false;
  isForwarderLoaded = false;
  isMapperLoaded = false;
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param llre - the LLRE pointer.
 */
/****************************************************************************/
XML_Handler::XML_Handler(LLRE *_llre, GFCP_WS_Client *_gfcpClient)
{
  llre = _llre;
  gfcpClient = _gfcpClient;
  isClassifierLoaded = false;
  isForwarderLoaded = false;
  isMapperLoaded = false;
}

/****************************************************************************/
//! Constructor.
/*! Constructor.
 * @param llre - the LLRE pointer.
 */
/****************************************************************************/
XML_Handler::XML_Handler(LLRE *_llre)
{
  llre = _llre;
  isClassifierLoaded = false;
  isForwarderLoaded = false;
  isMapperLoaded = false;
}

/****************************************************************************/
//! Destructor.
/*! Destructor.
 */
/****************************************************************************/
XML_Handler::~XML_Handler()
{
  clearMapperActions();
}

/****************************************************************************/
//! Handles an LFB XML.
/*! Handles an LFB XML.
 *
 * @param lfb - pointer to the LFB structure.
 * @return an lfb struct pointer that describes the result of the operation.
 */
/****************************************************************************/
void XML_Handler::Handle_LFB(LFB_structure &lfb)
{
  lfb.description = "";
  lfb.rsltType = Not_Set;

  vector<char> xmlVect(lfb.xml.begin(), lfb.xml.end());
  xmlVect.push_back('\0');

  xml_document<> doc;
  try
  {
    doc.parse<parse_declaration_node | parse_no_data_nodes> (&xmlVect[0]);
  }
  catch (parse_error e)
  {
    string error("Parsing error in LFB: ");
    error.append(e.what());
    error.append("\n");

    lfb.description = error;
    lfb.rsltType = Fail;
  }

  xml_node<>* cur_node = doc.first_node();
  if (!boost::iequals(cur_node->name(),"LFBs"))
  {
    cur_node = cur_node->next_sibling();
    if (cur_node == NULL || !boost::iequals(cur_node->name(),"LFBs") )
    {
      lfb.description = "Invalid XML file: Not LFB type XML.\n";
      lfb.rsltType = Fail;
    }
  }
  else
  {
    cur_node = cur_node->first_node();
    do
    {
      dispatch_LFB(lfb,cur_node);
      cur_node = cur_node->next_sibling();
    }while(cur_node!=NULL);

//    cur_node = cur_node->first_node();
//    string desc;
//    string nameStr(cur_node->name());
//
//    //rapidXML changes the passed vector<char>
//    vector<char>  xmlVect2(lfb.xml.begin(), lfb.xml.end());
//    xmlVect2.push_back('\0');
//
//    if(boost::iequals(nameStr,"LFB_Classifier"))
//    {
//      lfb.description = Handle_LFB_Classifier(xmlVect2);
//      if (lfb.description.compare("Classifier XML loaded successfully.\n") == 0)
//        lfb.rsltType = Success;
//      else
//        lfb.rsltType = Fail;
//    }
//    else if(boost::iequals(nameStr,"LFB_Mapper"))
//    {
//      lfb.description = Handle_LFB_Mapper(xmlVect2);
//      if (lfb.description.compare("Mapper XML loaded successfully.\n") == 0)
//        lfb.rsltType = Success;
//      else
//        lfb.rsltType = Fail;
//    }
//    else if(boost::iequals(nameStr,"LFB_Forwarder"))
//    {
//      lfb.description = Handle_LFB_Forwarder(xmlVect2);
//      if (lfb.description.compare("Forwarder XML loaded successfully.\n") == 0)
//        lfb.rsltType = Success;
//      else
//        lfb.rsltType = Fail;
//    }
//    else if(boost::iequals(nameStr,"portControl"))
//    {
//      lfb.description = "portControl is currently unsupported.\n";
//      lfb.rsltType = Fail;
//    }
//    else if(boost::iequals(nameStr,"fdbControl"))
//    {
//      cur_node = cur_node->first_node("type");
//      if(cur_node != NULL)
//      {
//        unsigned int entryID = 0;
//        int type = StringToNumber<int> (cur_node->value());
//        if(type == 0)
//        {
//          cur_node = cur_node->next_sibling("entryID");
//          if(cur_node != NULL)
//          {
//            entryID = StringToNumber<int> (cur_node->value());
//            lfb.description = llre->enableEntry(entryID);
//            lfb.rsltType = Success;
//          }
//          else
//          {
//            lfb.description = "Invalid XML, no entryID field in fdbControl.\n";
//            lfb.rsltType = Fail;
//          }
//        }
//        else
//        {
//          cur_node = cur_node->next_sibling("entryID");
//          if(cur_node != NULL)
//          {
//            entryID = StringToNumber<int> (cur_node->value());
//            lfb.description = llre->disableEntry(entryID);
//            lfb.rsltType = Success;
//          }
//          else
//          {
//            lfb.description = "Invalid XML, no entryID field in fdbControl.\n";
//            lfb.rsltType = Fail;
//          }
//        }
//      }
//      else
//      {
//        lfb.description = "Invalid XML, no type field in fdbControl.\n";
//        lfb.rsltType = Fail;
//      }
//    }
//    else if(boost::iequals(nameStr,"query"))
//    {
//      lfb.description = "query is currently unsupported.\n";
//      lfb.rsltType = Fail;
//    }
//    else
//    {
//      lfb.description = "Unknown LFB type";
//      lfb.rsltType = Fail;
//    }
  }
}

/****************************************************************************/
//! Handles an LFB XML.
/*! Handles an LFB XML.
 *
 * @param lfb - pointer to the LFB structure.
 * @return an lfb struct pointer that describes the result of the operation.
 */
/****************************************************************************/
void XML_Handler::dispatch_LFB(LFB_structure &lfb, xml_node<>* cur_node)
{
  string nameStr(cur_node->name());

  //rapidXML changes the passed vector<char>
  vector<char>  xmlVect2(lfb.xml.begin(), lfb.xml.end());
  xmlVect2.push_back('\0');

  if(boost::iequals(nameStr,"LFB_Classifier"))
  {
    lfb.description = Handle_LFB_Classifier_FromNode(cur_node);
    if (lfb.description.compare("Classifier XML loaded successfully.\n") == 0)
      lfb.rsltType = Success;
    else
      lfb.rsltType = Fail;
  }
  else if(boost::iequals(nameStr,"LFB_Mapper"))
  {
    lfb.description = Handle_LFB_Mapper_FromNode(cur_node);
    if (lfb.description.compare("Mapper XML loaded successfully.\n") == 0)
      lfb.rsltType = Success;
    else
      lfb.rsltType = Fail;
  }
  else if(boost::iequals(nameStr,"LFB_Forwarder"))
  {
    lfb.description = Handle_LFB_Forwarder_FromNode(cur_node);
    if (lfb.description.compare("Forwarder XML loaded successfully.\n") == 0)
      lfb.rsltType = Success;
    else
      lfb.rsltType = Fail;
  }
  else if(boost::iequals(nameStr,"portControl"))
  {
    lfb.description = "portControl is currently unsupported.\n";
    lfb.rsltType = Fail;
  }
  else if(boost::iequals(nameStr,"fdbControl"))
  {
    cur_node = cur_node->first_node("type");
    if(cur_node != NULL)
    {
      unsigned int entryID = 0;
      int type = StringToNumber<int> (cur_node->value());
      if(type == 0)
      {
        cur_node = cur_node->next_sibling("entryID");
        if(cur_node != NULL)
        {
          entryID = StringToNumber<int> (cur_node->value());
          lfb.description = llre->enableEntry(entryID);
          lfb.rsltType = Success;
        }
        else
        {
          lfb.description = "Invalid XML, no entryID field in fdbControl.\n";
          lfb.rsltType = Fail;
        }
      }
      else
      {
        cur_node = cur_node->next_sibling("entryID");
        if(cur_node != NULL)
        {
          entryID = StringToNumber<int> (cur_node->value());
          lfb.description = llre->disableEntry(entryID);
          lfb.rsltType = Success;
        }
        else
        {
          lfb.description = "Invalid XML, no entryID field in fdbControl.\n";
          lfb.rsltType = Fail;
        }
      }
    }
    else
    {
      lfb.description = "Invalid XML, no type field in fdbControl.\n";
      lfb.rsltType = Fail;
    }
  }
  else if(boost::iequals(nameStr,"query"))
  {
    lfb.description = "query is currently unsupported.\n";
    lfb.rsltType = Fail;
  }
  else
  {
    lfb.description = "Unknown LFB type";
    lfb.rsltType = Fail;
  }

}

/****************************************************************************/
//! Handles a Classifier LFB XML.
/*! Handles a Classifier LFB XML.
 *
 * @param LFB - pointer to the LFB XML.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Classifier_File(const char *LFB)
{
  if (fileExists(LFB))
  {
    vector<char> buffer;
    readFileToBuffer(LFB, buffer);
    return Handle_LFB_Classifier(buffer);
  }
  else
    return "Error opening Classifier file.\n";
}

/****************************************************************************/
//! Handles the converted Classifier LFB XML.
/*! Handles the converted Classifier LFB XML
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Classifier(vector<char> &XML)
{
  Classifier_Handler classifier(this);
  return classifier.parse(XML);
}

/****************************************************************************/
//! Handles the converted Classifier LFB XML.
/*! Handles the converted Classifier LFB XML
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Classifier_FromNode(xml_node<>* curr_node)
{
  Classifier_Handler classifier(this);
  return classifier.parseFromNode(curr_node);
}

/****************************************************************************/
//! Handles a Forwarder LFB XML.
/*! Handles a Forwarder LFB XML
 *
 * @param LFB - pointer to the LFB XML.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Forwarder_File(const char *LFB)
{
//  if (isClassifierLoaded && isMapperLoaded)
//  {
    if (fileExists(LFB))
    {
      vector<char> buffer;
      readFileToBuffer(LFB, buffer);
      return Handle_LFB_Forwarder(buffer);
    }
    else
      return "Error opening Forwarder file.\n";
//  }
//  else
//    return "Classifier and/or Mapper not loaded yet.\n";
}

/****************************************************************************/
//! Handles the converted Forwarder LFB XML.
/*! Handles the converted Forwarder LFB XML, all fields in the following
 *	example must be present:
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Forwarder(vector<char> &XML)
{
  Forwarder_Handler forwarder(this);
  return forwarder.parse(XML);
}

/****************************************************************************/
//! Handles the converted Forwarder LFB XML.
/*! Handles the converted Forwarder LFB XML, all fields in the following
 *      example must be present:
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Forwarder_FromNode(xml_node<>* curr_node)
{
  Forwarder_Handler forwarder(this);
  return forwarder.parseFromNode(curr_node);
}

/****************************************************************************/
//! Handles a Mapper LFB XML.
/*! Handles a Mapper LFB XML
 *
 * @param LFB - pointer to the LFB XML.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Mapper_File(const char *LFB)
{
//  if (isClassifierLoaded)
//  {
    if (fileExists(LFB))
    {
      vector<char> buffer;
      readFileToBuffer(LFB, buffer);
      return Handle_LFB_Mapper(buffer);
    }
    else
      return "Error opening Mapper file.\n";
//  }
//  else
//    return "Classifier not loaded yet.\n";
}

/****************************************************************************/
//! Handles the converted Mapper LFB XML.
/*! Handles the converted Mapper LFB XML
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Mapper(vector<char> &XML)
{
  Mapper_Handler mapper(this);
  return mapper.parse(XML);
}

/****************************************************************************/
//! Handles the converted Mapper LFB XML.
/*! Handles the converted Mapper LFB XML
 *
 * @param XML - reference to the LFB XML char vector.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_LFB_Mapper_FromNode(xml_node<>* curr_node)
{
  Mapper_Handler mapper(this);
  return mapper.parseFromNode(curr_node);
}

/****************************************************************************/
//! Adds an Action to the actionsVect, according to the passed type.
/*! Adds an Action to the actionsVect, according to the passed type. The
 *	checksum field is updated if necessary, when a field affects more than one
 *	checksum, all the inner checksums are updated before performing the actual
 *	action.
 *
 * @param actionsVect - vector of actions to add the new action to.
 * @param outFrame - outgoing frame name.
 * @param outgoing_field_name - outgoing frame field name.
 * @param fk_outgoing - outgoing FWD_KEY pointer.
 * @param fk_incoming - incoming FWD_KEY pointer.
 * @param fieldValue - pointer to the field value.
 * @param maskValue - pointer to the mask value.
 * @param actionType - 0 assign, 1 copy, 2 arithmetic.
 * @param description - action description - contains arithmetic operation type.
 */
/****************************************************************************/
void XML_Handler::addAction(vector<Action*> &actionsVect, string outFrame, string outgoing_field_name,
                            FWD_KEY *fk_outgoing, FWD_KEY *fk_incoming, BYTE *fieldValue, BYTE *maskValue,
                            int actionType, string description)
{
  if (actionType <= Arithmetic)
  {
    vector<Frame_field *> chksumFieldsVect;
    getChecksumKey(chksumFieldsVect, outFrame, outgoing_field_name);
    int arithmeticCode = -1;
    if (actionType == 2)
    {
//      if (description.compare("SUB") == 0)
      if(boost::iequals(description,"SUB"))
        arithmeticCode = _SUB;
//      else if (description.compare("ADD") == 0)
      else if(boost::iequals(description,"ADD"))
        arithmeticCode = _ADD;
//      else if (description.compare("AND") == 0)
      else if(boost::iequals(description,"AND"))
        arithmeticCode = _AND;
//      else if (description.compare("XOR") == 0)
      else if(boost::iequals(description,"XOR"))
        arithmeticCode = _XOR;
    }

    if (chksumFieldsVect.size() > 0)
    {
      //update only the checksum until last checksum is calculated
      for (unsigned int i = 0; i < chksumFieldsVect.size(); ++i)
      {
        if (i != chksumFieldsVect.size() - 1) //not last checksum
        {
          switch (actionType)
          {
            case Assign:
              actionsVect.push_back(new Action_Assign(fk_outgoing->getPosition(), fieldValue,
                                                      chksumFieldsVect.at(i)->getOffset(),
                                                      chksumFieldsVect.at(i)->getLength(), false));
              break;
            case Copy:
              actionsVect.push_back(new Action_Copy(fk_incoming->getPosition(), fk_outgoing->getPosition(),
                                                    chksumFieldsVect.at(i)->getOffset(),
                                                    chksumFieldsVect.at(i)->getLength(), false));
              break;
            case Arithmetic:
              actionsVect.push_back(new Action_Arithmetic(fk_outgoing->getOffset(), fk_outgoing->getLength(),
                                                          arithmeticCode, fieldValue, maskValue,
                                                          chksumFieldsVect.at(i)->getOffset(),
                                                          chksumFieldsVect.at(i)->getLength(), false));
              break;
            default:
              break;
          }
        }
        else
        {
          switch (actionType)
          {
            case Assign:
              actionsVect.push_back(new Action_Assign(fk_outgoing->getPosition(), fieldValue,
                                                      chksumFieldsVect.at(i)->getOffset(),
                                                      chksumFieldsVect.at(i)->getLength(), true));
              break;
            case Copy:
              actionsVect.push_back(new Action_Copy(fk_incoming->getPosition(), fk_outgoing->getPosition(),
                                                    chksumFieldsVect.at(i)->getOffset(),
                                                    chksumFieldsVect.at(i)->getLength(), true));
              break;
            case Arithmetic:
              actionsVect.push_back(new Action_Arithmetic(fk_outgoing->getOffset(), fk_outgoing->getLength(),
                                                          arithmeticCode, fieldValue, maskValue,
                                                          chksumFieldsVect.at(i)->getOffset(),
                                                          chksumFieldsVect.at(i)->getLength(), true));
              break;
            default:
              break;
          }
        }
      }
    }
    else
    {
      switch (actionType)
      {
        case Assign:
          actionsVect.push_back(new Action_Assign(fk_outgoing->getPosition(), fieldValue));
          break;
        case Copy:
          actionsVect.push_back(new Action_Copy(fk_incoming->getPosition(), fk_outgoing->getPosition()));
          break;
        case Arithmetic:
          actionsVect.push_back(new Action_Arithmetic(fk_outgoing->getOffset(), fk_outgoing->getLength(),
                                                      arithmeticCode, fieldValue, maskValue));
          break;
        default:
          break;
      }
    }
  }
  else
  { //add and cut don't affect checksum
    //TODO add multiple adds/cuts
    switch (actionType)
    {

      case Add: //actionsVect.push_back(new Action_Add(fk_outgoing->getOffset(), fk_outgoing->getLength()));
        if (addPosition.getLength() == 0)
          addPosition.setOffset(fk_outgoing->getOffset());

        addPosition.setLength(addPosition.getLength() + (addPosition.getFirstBitOffset()
            + fk_outgoing->getPosition().getNumberOfBits()) / 8);
        addPosition.setFirstBitOffset((addPosition.getFirstBitOffset() + fk_outgoing->getPosition().getNumberOfBits())
            % 8);

        break;
      case Cut: //actionsVect.push_back(new Action_Cut(fk_outgoing->getOffset(), fk_outgoing->getLength()));
        if (cutPosition.getLength() == 0)
          cutPosition.setOffset(fk_incoming->getOffset());

        cutPosition.setLength(cutPosition.getLength() + (cutPosition.getFirstBitOffset()
            + fk_incoming->getPosition().getNumberOfBits()) / 8);
        cutPosition.setFirstBitOffset((cutPosition.getFirstBitOffset() + fk_incoming->getPosition().getNumberOfBits())
            % 8);
        break;
      default:
        break;
    }
  }
}

/****************************************************************************/
//! Converts string to BYTEs.
/*! Converts string to BYTEs.
 *
 * @param s - string.
 * @return BYTE representation of the string.
 */
/****************************************************************************/
BYTE *XML_Handler::stringToBYTEs(const string &s, const Position &pos)
{
  string str(s);
  BYTE *bytes = NULL;

  boost::trim(str); //remove white spaces

  if (isValidString(str, HEXSTRREGEX))
  {
    bytes = hexStringToBYTEs(str, pos.getLength());
  }
  else if (isValidString(str, DECSTRREGEX))
  {
    bytes = decStringToBYTEs(str, pos.getLength());
  }
  else if (isValidString(str, IPREGEX))
  {
    bytes = ipStringToBYTEs(str);
  }

  return bytes;
}

/****************************************************************************/
//! Converts hex string to BYTEs.
/*! Converts hex string to BYTEs.
 *
 * @param s - hex string.
 * @return BYTE representation of the hex string.
 */
/****************************************************************************/
BYTE *XML_Handler::hexStringToBYTEs(const string &s, unsigned int len)
{
  string twoChars;
  string str(s, 2);
  int b = 0;
  int j = 0;
  BYTE *bytes = NULL;

  if ((str.length() / 2) < len)//make an even number of digits
  {
    ostringstream ss;
    ss << setw(len * 2) << setfill('0') << str;
    str = ss.str();
  }

  bytes = new BYTE[len];
  for (unsigned int i = 0; i < str.length(); i += 2)
  {
    twoChars.append(1, str[i]);
    twoChars.append(1, str[i + 1]);
    b = StringToHexNumber<int> (twoChars);
    bytes[j] = (char)b;
    ++j;
    twoChars.clear();
  }
  return bytes;
}

/****************************************************************************/
//! Converts decimal string to BYTEs.
/*! Converts decimal string to BYTEs.
 *
 * @param s - decimal string.
 * @return BYTE representation of the decimal string.
 */
/****************************************************************************/
BYTE *XML_Handler::decStringToBYTEs(const string &s, unsigned int len)
{
  string str(s);
  BYTE *bytes = NULL;

  if (s.length() > 0)
  {
    int number = StringToNumber<int> (str);
    int size = ceil((log(number) / log(2)) / 8);
    bytes = new BYTE[size];
    unsigned int j = size - 1;
    for (int i = 0; i < size; i++)
    {
      bytes[j] = (BYTE)(number >> (i * 8));
      j--;
    }
  }
  return bytes;
}

/****************************************************************************/
//! Converts ip string to BYTEs.
/*! Converts ip string to BYTEs. String is assumed to be correct.
 *
 * @param s - ip string.
 * @return BYTE representation of the ip string.
 */
/****************************************************************************/
BYTE *XML_Handler::ipStringToBYTEs(const string &s)
{
  string str(s);
  string tuple;
  BYTE *bytes = new BYTE[4];

  size_t delim_pos = str.find('.');
  tuple = str.substr(0, delim_pos);
  bytes[0] = StringToNumber<int> (tuple);
  str = str.substr(delim_pos + 1, str.size());

  delim_pos = str.find('.');
  tuple = str.substr(0, delim_pos);
  bytes[1] = StringToNumber<int> (tuple);
  str = str.substr(delim_pos + 1, str.size());

  delim_pos = str.find('.');
  tuple = str.substr(0, delim_pos);
  bytes[2] = StringToNumber<int> (tuple);

  tuple = str.substr(delim_pos + 1, str.size());
  bytes[3] = StringToNumber<int> (tuple);

  return bytes;
}

/****************************************************************************/
//! Reads the passed fileName into vector buffer.
/*! Reads the passed fileName into vector buffer.
 *
 * @param fileName - points to the file to read.
 * @param buffer - buffer to store the file.
 * @return true if successful, false otherwise.
 */
/****************************************************************************/
bool XML_Handler::readFileToBuffer(const char *fileName, vector<char> &buffer)
{
  ifstream file(fileName);
  if (file)
  {
    file >> noskipws;
    copy(istream_iterator<char> (file), istream_iterator<char> (), back_inserter(buffer));
    buffer.push_back('\0');
    return true;
  }
  return false;
}

/****************************************************************************/
//! Handles a port assignment request to handle frameTypeID
/*! Handles a port assignment request to handle frameTypeID
 *
 * @param port - network interface number.
 * @param frameTypeId - frame type id.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Handle_Port_Assignment(unsigned int port, unsigned char frameTypeId)
{
  //TODO implement
  return "DONE.";
}

/****************************************************************************/
//! Reset port configuration.
/*! Reset port configuration.
 *
 * @param port - network interface number.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Reset_Port(unsigned int port)
{
  //TODO implement
  return "DONE.";
}

/****************************************************************************/
//! Deletes the entryID data base entry.
/*! Deletes the entryID data base entry.
 *
 * @param entryID - entry number to delete.
 * @return a string that describes the result of the operation.
 */
/****************************************************************************/
string XML_Handler::Delete_Entry(int entryID)
{
  //TODO implement
  return "DONE.";
}

/****************************************************************************/
//! Returns the LLRE pointer.
/*! Returns the LLRE pointer.
 *
 * @return Returns the LLRE pointer.
 */
/****************************************************************************/
LLRE *XML_Handler::getLlre() const
{
  return llre;
}

/****************************************************************************/
//! Sets the LLRE pointer.
/*! Sets the LLRE pointer.
 *
 * @param llre - the LLRE pointer.
 */
/****************************************************************************/
void XML_Handler::setLlre(LLRE *llre)
{
  this->llre = llre;
}

/****************************************************************************/
//! Prints the Initial_Lookup Table.
/*! Prints the Initial_Lookup Table.
 */
/****************************************************************************/
void XML_Handler::printTables()
{
  cout << "ID_KEYs Table:\n";
  for (umultimap_idKeys::const_iterator it = idKeysMap.begin(); it != idKeysMap.end(); ++it)
  {
    cout << "[" << it->first << "," << it->second << "]\n";
  }
  cout << endl;

  cout << "FWD_KEYs Table:\n";
  for (umultimap_fwdKeys::const_iterator it = fwdKeysMap.begin(); it != fwdKeysMap.end(); ++it)
  {
    cout << "[" << it->first << "," << it->second << "]\n";
  }
  cout << endl;
}

/****************************************************************************/
//! Returns the classifier FWD_KEY matching the field_name and frame_type_id.
/*! Returns the classifier FWD_KEY matching the field_name and frame_type_id.
 *
 * @param frame_type_id - type ID of the frame.
 * @param field_name - name of the field.
 * @return FWD_KEY matching the field_name and frame_type_id.
 */
/****************************************************************************/
FWD_KEY *XML_Handler::getClassifierFields(string &frame_type_id, string &field_name)
{
  umultimap_mapOFfwdKeysMaps::iterator iter = classifierKeys.find(frame_type_id);
  if (iter != classifierKeys.end())
  {
    umultimap_fwdKeys::iterator it = iter->second.find(field_name);
    if (it != iter->second.end())
      return &(it->second);
  }
  return NULL;
}

/****************************************************************************/
//! Returns the classifier FWD_KEY matching the name and frame_type_id.
/*! Returns the classifier FWD_KEY matching the name and frame_type_id.
 *
 * @param frame_type_id - type ID of the frame.
 * @param name - name of the field.
 * @return CHKSUM_KEY matching the field_name and frame_type_id.
 */
/****************************************************************************/
void XML_Handler::getChecksumKey(vector<Frame_field*> &frameFieldsVect, const string &frame_type_id, const string &name)
{
  umultimap_mapOFchksumFieldsNamesMaps::iterator iter = fieldsToChecksumsMaps.find(frame_type_id);
  if (iter != fieldsToChecksumsMaps.end())
  {
    pair<umultimap_chksumFieldsNames::iterator, umultimap_chksumFieldsNames::iterator> range =
        iter->second.equal_range(name);

    for (umultimap_chksumFieldsNames::iterator it = range.first; it != range.second; ++it)
    {
      frameFieldsVect.push_back(&(it->second));
    }
  }
}

/****************************************************************************/
//! Returns a vector of Actions corresponding to the mappingId.
/*! Returns a vector of Actions corresponding to the mappingId.
 *
 * @param mappingId - mappingId string.
 * @param mapActionsVect vector - of Actions corresponding to the mappingId.
 */
/****************************************************************************/
void XML_Handler::getMapperActions(string &mappingId, vector<Action*> &mapActionsVect)
{
  umultimap_mapperActions::iterator iter = mapperActions.find(mappingId);
  if (iter != mapperActions.end())
  {
    for (unsigned int i = 0; i < (iter->second).size(); ++i)
    {
      mapActionsVect.push_back(((iter->second)[i])->clone());
    }
  }
}

/****************************************************************************/
//! Clears the map of Actions corresponding to the mappingIds.
/*! Clears the map of Actions corresponding to the mappingIds.
 */
/****************************************************************************/
void XML_Handler::clearMapperActions()
{
  for (umultimap_mapperActions::iterator iter = mapperActions.begin(); iter != mapperActions.end(); ++iter)
  {
    for (unsigned int i = 0; i < (iter->second).size(); ++i)
    {
      if (((iter->second)[i]) != NULL)
        delete (iter->second)[i];
    }
  }
  mapperActions.clear();
  mappingIdToOutFrameMap.clear();
}

/****************************************************************************/
//! Returns a vector of ID_KEYs that match the Frame Identifier.
/*! Returns a vector of ID_KEYs that match the Frame Identifier.
 *
 * @param frame_type_id - Frame Identifier.
 * @param idKeysVect - a vector of ID_KEYs that match the Frame Identifier.
 */
/****************************************************************************/
void XML_Handler::getID_KEYs(string &frame_type_id, vector<ID_KEY*> &idKeysVect, umultimap_idKeys &kMap)
{
  pair<umultimap_idKeys::iterator, umultimap_idKeys::iterator> range = kMap.equal_range(frame_type_id);

  for (umultimap_idKeys::iterator it = range.first; it != range.second; ++it)
  {
    idKeysVect.push_back(&(it->second));
  }
}

/****************************************************************************/
//! Returns a vector of FWD_KEYs that match the Frame Identifier.
/*! Returns a vector of FWD_KEYs that match the Frame Identifier.
 *
 * @param frame_type_id - Frame Identifier.
 * @param fwdKeysVect - a vector of FWD_KEYs that match the Frame Identifier.
 */
/****************************************************************************/
void XML_Handler::getFWD_KEYs(string &frame_type_id, vector<FWD_KEY*> &fwdKeysVect)
{
  pair<umultimap_fwdKeys::iterator, umultimap_fwdKeys::iterator> range = fwdKeysMap.equal_range(frame_type_id);

  for (umultimap_fwdKeys::iterator it = range.first; it != range.second; ++it)
  {
    fwdKeysVect.push_back(&(it->second));
  }
}

/****************************************************************************/
//! Fills a vector of FWD_KEYs with keys that match the name.
/*! Fills a vector of FWD_KEYs with keys that match the name.
 *
 * @param name - key name.
 * @param fwdKeysVect - a vector of FWD_KEYs.
 */
/****************************************************************************/
void XML_Handler::getResFWD_KEYs(string &name, vector<FWD_KEY*> &fwdKeysVect)
{
  for (umultimap_fwdKeys::iterator it = fwdKeysMap.begin(); it != fwdKeysMap.end(); ++it)
  {
//    if ((it->second).getName().compare(name) == 0)
    if (boost::iequals((it->second).getName(),name))
      fwdKeysVect.push_back(&(it->second));
  }
}

///****************************************************************************/
////! Starts a new queue processing thread.
///*! Starts a new queue processing thread, to serve the XMLs queue.
// */
///****************************************************************************/
//void XML_Handler::startQueueProcessingThread()
//{
//  threadPtr = 0;
//  if (pthread_create(&threadPtr, NULL, processQueue, (void *)this))
//    throw GFEexception("Failed to start new XML_Handler thread.");
//}

///****************************************************************************/
////! Processing thread function, to serve the XMLs queue.
///*! Processing thread function, to serve the XMLs queue. This thread is
// * terminated by adding a NULL frame to queue.
// *
// * @param _xml_handler - an LLRE object.
// */
///****************************************************************************/
//void *XML_Handler::processQueue(void *_xml_handler)
//{
//  XML_Handler *xml_handler = (XML_Handler *)_xml_handler;
//  LFB_structure *lfb = NULL;
//  string desc;
//  bool processFlag = true;
//  while (processFlag)
//  {
//    xml_handler->xmlQ.wait_and_pop(lfb);
//
//    if (lfb != NULL)
//    {
//      vector<char> xmlVect(lfb->xml.begin(), lfb->xml.end());
//      xmlVect.push_back('\0');
//
//      //TODO add port assignment handling
//      //TODO add feedback
//      switch (lfb->lfbType)
//      {
//        case Classifier:
//
//          if (lfb->description != NULL)
//            delete lfb->description;
//          lfb->description = xml_handler->Handle_LFB_Classifier(xmlVect);
//          if (desc.compare("Classifier XML loaded successfully.\n") == 0)
//            lfb->rsltType = Success;
//          else
//            lfb->rsltType = Fail;
//
//          break;
//        case Mapper:
//
//          if (lfb->description != NULL)
//            delete lfb->description;
//          lfb->description = xml_handler->Handle_LFB_Mapper(xmlVect);
//          if (desc.compare("Mapper XML loaded successfully.\n") == 0)
//            lfb->rsltType = Success;
//          else
//            lfb->rsltType = Fail;
//
//          break;
//        case Forwarder:
//          if (lfb->description != NULL)
//            delete lfb->description;
//          lfb->description = xml_handler->Handle_LFB_Forwarder(xmlVect);
//          if (desc.compare("Forwarder XML loaded successfully.\n") == 0)
//            lfb->rsltType = Success;
//          else
//            lfb->rsltType = Fail;
//
//          break;
//        default:
//
//          lfb->description = "Unknown LFB. \n";
//
//          break;
//      }
//      //TODO FIX!!!!
////      xml_handler->gfcpClient->pushLFBToQueue(lfb);
//      //			delete lfb->xml;
//      //			delete lfb;
//    }
//    else
//      processFlag = false;
//  }
//  pthread_exit(0);
//  return 0;
//}

///****************************************************************************/
////! Adds a new LFB to the processing queue.
///*! Adds a new LFB to the processing queue. This queue is used to pass
// * LFB between the protocol agent and the XML Handler, for thread safety.
// *
// * @param LFB - new LFB to process.
// */
///****************************************************************************/
//void XML_Handler::pushLFBToQueue(LFB_structure *LFB)
//{
//  xmlQ.push(LFB);
//}

/****************************************************************************/
//! Returns the xmls load status.
/*! Returns the xmls load status:
 *	0 - No xmls loaded.
 *	1 - Classifier loaded.
 *	2 - Classifier and Forwarder loaded.
 *	3 - Classifier, Forwarder and Mapper loaded.
 *
 * @return - xml load status
 */
/****************************************************************************/
int XML_Handler::xmlLoadStatus()
{
  return 1 * isClassifierLoaded + 1 * isForwarderLoaded + 1 * isMapperLoaded;
}

/**
 * taken from:
 * http://www.techbytes.ca/techbyte103.html
 */
bool XML_Handler::fileExists(string strFilename)
{
  struct stat stFileInfo;
  bool blnReturn;
  int intStat;

  // Attempt to get the file attributes
  intStat = stat(strFilename.c_str(), &stFileInfo);
  if (intStat == 0)
  {
    // We were able to get the file attributes
    // so the file obviously exists.
    blnReturn = true;
  }
  else
  {
    // We were not able to get the file attributes.
    // This may mean that we don't have permission to
    // access the folder which contains this file. If you
    // need to do that level of checking, lookup the
    // return values of stat which will give you
    // more details on why stat failed.
    blnReturn = false;
  }

  return (blnReturn);
}

/****************************************************************************/
//! Checks received string matches the regex define.
/*! Checks received string matches the regex define.
 *
 * @param str a char pointer - points to the received string.
 * @return true if valid false if not.
 */
/****************************************************************************/
bool XML_Handler::isValidString(const string &str, const char *regexDefine)
{
  bool isValid = true;
  regex_t regex;

  if (regcomp(&regex, regexDefine, REG_EXTENDED) != 0)
    isValid = false;

  if (regexec(&regex, str.c_str(), 0, NULL, 0) != 0)
    isValid = false;

  regfree(&regex);

  return isValid;
}
