/****************************************************************************
 *
 * File Name: XML_Handler.h
 *
 * General Description: Defines the XML Handler class.
 *
 * Created By: Alex Creation Date: 01/01/2011
 *
 * Revisions And Updates:
 *
 * Date 		 Updated By 	  Description
 * ========== ================ =========================
 *
 *
 ****************************************************************************/
#ifndef XML_HANDLER_H
#define XML_HANDLER_H
/****************************************************************************
 * File description
 ****************************************************************************/
/*! \file XML_Handler.h
 *\brief Defines the XML Handler class.
 *
 * This module is responsible for converting received LFBs, in a form of XMLs,
 * into the data base entries which it then stores in the lookup tables of
 * the LLRE.
 *
 *\author Alex
 */

#include <sys/stat.h>
#include <regex.h>
#include "RapidXML/rapidxml.hpp"
//#include "RapidXML/rapidxml_iterators.hpp"
#include "RapidXML/rapidxml_print.hpp"
#include "RapidXML/rapidxml_utils.hpp"
#include "../LLRE/LLRE.h"
#include "../Includes/concurrent_queue.h"
#include <boost/algorithm/string.hpp>


#define IPREGEX "^((0|1[0-9]{0,2}|2[0-9]{0,1}|2[0-4][0-9]|25[0-5]|[3-9][0-9]{0,1})\\.){3}(0|1[0-9]{0,2}|2[0-9]{0,1}|2[0-4][0-9]|25[0-5]|[3-9][0-9]{0,1})$"
#define HEXSTRREGEX "^(0[xX][0-9a-fA-F]+)$"
#define DECSTRREGEX "^([0-9]+)$"

#define IDENTIFIER 	"identifier"
#define FORWARDING 	"forwarding"
#define CHECKSUM 	"checksum"
#define QOS	 		"QoS"
#define REGULAR		"regular"
#define METERING	"metering"

class GFCP_WS_Client;

enum TransactionType
{ //transactionType
  Classifier, Forwarder, Mapper, portAssign, toggleFDB, query
};

enum ActionType
{
  Assign, Copy, Arithmetic, Add, Cut
};

enum ResultType
{
  Success, Fail, Not_Set
};

typedef struct LFB_structure_t
{
  std::string src;
  std::string dst;
  std::string transactionID;
  TransactionType lfbType; //lfb type
  std::string xml; //pointer to xml file string
  ResultType rsltType;
  std::string description; //pointer to xml file string
} LFB_structure;

typedef boost::unordered_multimap<std::string, ID_KEY> umultimap_idKeys;
typedef boost::unordered_multimap<std::string, FWD_KEY> umultimap_fwdKeys;
typedef boost::unordered_multimap<std::string, Frame_field> umultimap_chksumFieldsNames;
typedef boost::unordered_multimap<std::string, std::vector<Action*> > umultimap_mapperActions;
typedef boost::unordered_multimap<std::string, umultimap_fwdKeys> umultimap_mapOFfwdKeysMaps;
typedef boost::unordered_multimap<std::string, umultimap_chksumFieldsNames> umultimap_mapOFchksumFieldsNamesMaps;
typedef boost::unordered_map<std::string, std::string> umap_mappingIdToOutFrameID;

class XML_Handler
{
public:

  friend class Classifier_Handler;
  friend class Mapper_Handler;
  friend class Forwarder_Handler;

  XML_Handler();
  XML_Handler(LLRE *_llre);
  XML_Handler(LLRE *_llre, GFCP_WS_Client *_gfcpClient);
  virtual ~XML_Handler();
  void Handle_LFB(LFB_structure &lfb);
  std::string Handle_LFB_Classifier_File(const char *LFB);
  std::string Handle_LFB_Classifier(std::vector<char> &XML);
  std::string Handle_LFB_Classifier_FromNode(rapidxml::xml_node<>* curr_node);
  std::string Handle_LFB_Forwarder_File(const char *LFB);
  std::string Handle_LFB_Forwarder(std::vector<char> &XML);
  std::string Handle_LFB_Forwarder_FromNode(rapidxml::xml_node<>* curr_node);
  std::string Handle_LFB_Mapper_File(const char *LFB);
  std::string Handle_LFB_Mapper(std::vector<char> &XML);
  std::string Handle_LFB_Mapper_FromNode(rapidxml::xml_node<>* curr_node);
  std::string Handle_Port_Assignment(unsigned int port, unsigned char frameTypeId);
  std::string Reset_Port(unsigned int port);
  std::string Delete_Entry(int entryID);
  LLRE *getLlre() const;
  void setLlre(LLRE *llre);
  void printTables();
//  void pushLFBToQueue(LFB_structure *LFB);
  int xmlLoadStatus();

  static bool fileExists(std::string strFilename);
  static BYTE *stringToBYTEs(const std::string &s, const Position &pos);
  static bool isValidString(const std::string &str, const char *regexDefine);

private:
  bool readFileToBuffer(const char *fileName, std::vector<char> &buffer);
  void getID_KEYs(std::string &frame_type_id, std::vector<ID_KEY*> &idKeysVect, umultimap_idKeys &kMap);
  void getFWD_KEYs(std::string &frame_type_id, std::vector<FWD_KEY*> &fwdKeysVect);
  void getResFWD_KEYs(std::string &name, std::vector<FWD_KEY*> &fwdKeysVect);
  void setChecksumFlags(std::string frame_type_id, std::string field_name);
  void addAction(std::vector<Action*> &actionsVect, std::string outFrame, std::string outgoing_field_name,
                 FWD_KEY *fk_outgoing, FWD_KEY *fk_incoming, BYTE *fieldValue, BYTE *maskValue, int actionType,
                 std::string description);
  FWD_KEY *getClassifierFields(std::string &frame_type_id, std::string &field_name);
  void getChecksumKey(std::vector<Frame_field*> &frameFieldsVect, const std::string &frame_type_id,
                      const std::string &name);
  void getMapperActions(std::string &mappingId, std::vector<Action*> &mapActionsVect);
  void clearMapperActions();
  void dispatch_LFB(LFB_structure &lfb, rapidxml::xml_node<>* cur_node);
//  void startQueueProcessingThread();
//  static void *processQueue(void *_xml_handler);
  std::string processClassifierName(std::string frame_type_id, rapidxml::xml_node<>* fields_node, Position &pos);
  static BYTE *hexStringToBYTEs(const std::string &s, unsigned int len);
  static BYTE *decStringToBYTEs(const std::string &s, unsigned int len);
  static BYTE *ipStringToBYTEs(const std::string &s);

protected:
  LLRE * llre; //LLRE pointer
  GFCP_WS_Client * gfcpClient; //
  umultimap_idKeys idKeysMap; //Stores ID_KEYs								//Classifier, Forwarder
  umultimap_idKeys idKeysWithValuesMap; //Stores ID_KEYs with values					//Classifier, Forwarder
  umultimap_fwdKeys fwdKeysMap; //Stores FWD_KEYs								//Classifier, Forwarder
  umultimap_mapOFfwdKeysMaps classifierKeys; //Stores all the frame fields in the classifier //Classifier, Forwarder, Mapper
  umultimap_mapperActions mapperActions; //Actions derived from the Mapper lfb
  umultimap_mapOFchksumFieldsNamesMaps fieldsToChecksumsMaps; //Stores the field_id and field_name to checksum_name
  umap_mappingIdToOutFrameID mappingIdToOutFrameMap; //stores the mapping ID to output frame ID
  pthread_t threadPtr; //queue processing thread id
//  concurrent_queue<LFB_structure *> xmlQ; //thread safe xmls queue
  bool isClassifierLoaded; //indicates whether classifier.xml was loaded
  bool isForwarderLoaded; //indicates whether forwarder.xml was loaded
  bool isMapperLoaded; //indicates whether mapper.xml was loaded
  Position addPosition; //position of the Action_Add
  Position cutPosition; //position of the Action_Cut
};

#endif /* XML_HANDLER_H */
